(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/core.js                                                      //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require('crypto');                                    // 17
var Capnp = Npm.require('capnp');                                      // 18
                                                                       //
var PersistentHandle = Capnp.importSystem('sandstorm/supervisor.capnp').PersistentHandle;
var SandstormCore = Capnp.importSystem('sandstorm/supervisor.capnp').SandstormCore;
var SandstormCoreFactory = Capnp.importSystem('sandstorm/backend.capnp').SandstormCoreFactory;
var PersistentOngoingNotification = Capnp.importSystem('sandstorm/supervisor.capnp').PersistentOngoingNotification;
                                                                       //
var SandstormCoreImpl = (function () {                                 //
  function SandstormCoreImpl(grainId) {                                // 26
    babelHelpers.classCallCheck(this, SandstormCoreImpl);              //
                                                                       //
    this.grainId = grainId;                                            // 27
  }                                                                    //
                                                                       //
  SandstormCoreImpl.prototype.restore = (function () {                 // 25
    function restore(sturdyRef, requiredPermissions) {                 // 30
      var _this = this;                                                // 31
      return inMeteor(function () {                                    // 32
        var hashedSturdyRef = hashSturdyRef(sturdyRef);                // 33
        var token = ApiTokens.findOne(hashedSturdyRef);                // 34
        var requirements = [{                                          // 35
          tokenValid: hashedSturdyRef                                  // 37
        }];                                                            //
                                                                       //
        if (requiredPermissions && token.owner.grain.introducerIdentity) {
          requirements.push({                                          // 42
            permissionsHeld: {                                         // 43
              permissions: requiredPermissions,                        // 44
              identityId: token.owner.grain.introducerIdentity,        // 45
              grainId: _this.grainId                                   // 46
            }                                                          //
          });                                                          //
        }                                                              //
                                                                       //
        return restoreInternal(hashedSturdyRef, { grain: Match.ObjectIncluding({ grainId: _this.grainId }) }, requirements, sturdyRef);
      });                                                              //
    }                                                                  //
                                                                       //
    return restore;                                                    //
  })();                                                                //
                                                                       //
  SandstormCoreImpl.prototype.drop = (function () {                    // 25
    function drop(sturdyRef) {                                         // 57
      var _this = this;                                                // 58
      return inMeteor(function () {                                    // 59
        return dropInternal(sturdyRef, { grain: Match.ObjectIncluding({ grainId: _this.grainId }) });
      });                                                              //
    }                                                                  //
                                                                       //
    return drop;                                                       //
  })();                                                                //
                                                                       //
  SandstormCoreImpl.prototype.makeToken = (function () {               // 25
    function makeToken(ref, owner, requirements) {                     // 64
      var _this = this;                                                // 65
      return inMeteor(function () {                                    // 66
        var sturdyRef = new Buffer(generateSturdyRef());               // 67
        var hashedSturdyRef = hashSturdyRef(sturdyRef);                // 68
        ApiTokens.insert({                                             // 69
          _id: hashedSturdyRef,                                        // 70
          grainId: _this.grainId,                                      // 71
          objectId: ref,                                               // 72
          owner: owner,                                                // 73
          created: new Date(),                                         // 74
          requirements: requirements                                   // 75
        });                                                            //
                                                                       //
        return {                                                       // 78
          token: sturdyRef                                             // 79
        };                                                             //
      });                                                              //
    }                                                                  //
                                                                       //
    return makeToken;                                                  //
  })();                                                                //
                                                                       //
  SandstormCoreImpl.prototype.makeChildToken = (function () {          // 25
    function makeChildToken(parent, owner, requirements) {             // 84
      var _this = this;                                                // 85
      return inMeteor(function () {                                    // 86
        return makeChildTokenInternal(hashSturdyRef(parent), owner, requirements, _this.grainId);
      });                                                              //
    }                                                                  //
                                                                       //
    return makeChildToken;                                             //
  })();                                                                //
                                                                       //
  SandstormCoreImpl.prototype.getOwnerNotificationTarget = (function () {
    function getOwnerNotificationTarget() {                            // 91
      var grainId = this.grainId;                                      // 92
      return {                                                         // 93
        owner: {                                                       // 94
          addOngoing: function (displayInfo, notification) {           // 95
            return inMeteor(function () {                              // 96
              var grain = Grains.findOne({ _id: grainId });            // 97
              if (!grain) {                                            // 98
                throw new Error('Grain not found.');                   // 99
              }                                                        //
                                                                       //
              var castedNotification = notification.castAs(PersistentOngoingNotification);
              var wakelockToken = waitPromise(castedNotification.save()).sturdyRef;
                                                                       //
              // We have to close both the casted cap and the original. Perhaps this should be fixed in
              // node-capnp?                                           //
              castedNotification.close();                              // 107
              notification.close();                                    // 108
              var notificationId = Notifications.insert({              // 109
                ongoing: wakelockToken,                                // 110
                grainId: grainId,                                      // 111
                userId: grain.userId,                                  // 112
                text: displayInfo.caption,                             // 113
                timestamp: new Date(),                                 // 114
                isUnread: true                                         // 115
              });                                                      //
                                                                       //
              return { handle: makeNotificationHandle(notificationId, false) };
            });                                                        //
          }                                                            //
        }                                                              //
      };                                                               //
    }                                                                  //
                                                                       //
    return getOwnerNotificationTarget;                                 //
  })();                                                                //
                                                                       //
  return SandstormCoreImpl;                                            //
})();                                                                  //
                                                                       //
var makeSandstormCore = function (grainId) {                           // 126
  return new Capnp.Capability(new SandstormCoreImpl(grainId), SandstormCore);
};                                                                     //
                                                                       //
var NotificationHandle = (function () {                                //
  function NotificationHandle(notificationId, saved) {                 // 131
    babelHelpers.classCallCheck(this, NotificationHandle);             //
                                                                       //
    this.notificationId = notificationId;                              // 132
    this.saved = saved;                                                // 133
  }                                                                    //
                                                                       //
  NotificationHandle.prototype.close = (function () {                  // 130
    function close() {                                                 // 136
      var _this = this;                                                // 137
      return inMeteor(function () {                                    // 138
        if (!_this.saved) {                                            // 139
          dismissNotification(_this.notificationId);                   // 140
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  NotificationHandle.prototype.save = (function () {                   // 130
    function save(params) {                                            // 145
      return saveFrontendRef({ notificationHandle: this.notificationId }, params.sealFor);
    }                                                                  //
                                                                       //
    return save;                                                       //
  })();                                                                //
                                                                       //
  return NotificationHandle;                                           //
})();                                                                  //
                                                                       //
function makeNotificationHandle(notificationId, saved) {               // 150
  return new Capnp.Capability(new NotificationHandle(notificationId, saved), PersistentHandle);
}                                                                      //
                                                                       //
function dropWakelock(grainId, wakeLockNotificationId) {               // 154
  waitPromise(globalBackend.useGrain(grainId, function (supervisor) {  // 155
    return supervisor.drop({ wakeLockNotification: wakeLockNotificationId });
  }));                                                                 //
}                                                                      //
                                                                       //
function dismissNotification(notificationId, callCancel) {             // 160
  var notification = Notifications.findOne({ _id: notificationId });   // 161
  if (notification) {                                                  // 162
    Notifications.remove({ _id: notificationId });                     // 163
    if (notification.ongoing) {                                        // 164
      // For some reason, Mongo returns an object that looks buffer-like, but isn't a buffer.
      // Only way to fix seems to be to copy it.                       //
      var id = new Buffer(notification.ongoing);                       // 167
                                                                       //
      if (!callCancel) {                                               // 169
        dropInternal(id, { frontend: null });                          // 170
      } else {                                                         //
        var notificationCap = restoreInternal(hashSturdyRef(id), { frontend: null }).cap;
        var castedNotification = notificationCap.castAs(PersistentOngoingNotification);
        dropInternal(id, { frontend: null });                          // 174
        try {                                                          // 175
          waitPromise(castedNotification.cancel());                    // 176
          castedNotification.close();                                  // 177
          notificationCap.close();                                     // 178
        } catch (err) {                                                //
          if (err.kjType !== 'disconnected') {                         // 180
            // ignore disconnected errors, since cancel may shutdown the grain before the supervisor
            // responds.                                               //
            throw err;                                                 // 183
          }                                                            //
        }                                                              //
      }                                                                //
    } else if (notification.appUpdates) {                              //
      _.forEach(notification.appUpdates, function (app, appId) {       // 188
        deletePackage(app.packageId);                                  // 189
      });                                                              //
    }                                                                  //
  }                                                                    //
}                                                                      //
                                                                       //
hashSturdyRef = function (sturdyRef) {                                 // 195
  return Crypto.createHash('sha256').update(sturdyRef).digest('base64');
};                                                                     //
                                                                       //
function generateSturdyRef() {                                         // 199
  return Random.id(22);                                                // 200
}                                                                      //
                                                                       //
Meteor.methods({                                                       // 203
  dismissNotification: function (notificationId) {                     // 204
    // This will remove notifications from the database and from view of the user.
    // For ongoing notifications, it will begin the process of cancelling and dropping them from
    // the app.                                                        //
                                                                       //
    check(notificationId, String);                                     // 209
                                                                       //
    var notification = Notifications.findOne({ _id: notificationId });
    if (!notification) {                                               // 212
      throw new Meteor.Error(404, 'Notification id not found.');       // 213
    } else if (notification.userId !== Meteor.userId()) {              //
      throw new Meteor.Error(403, 'Notification does not belong to current user.');
    } else {                                                           //
      dismissNotification(notificationId, true);                       // 217
    }                                                                  //
  },                                                                   //
                                                                       //
  readAllNotifications: function () {                                  // 221
    // Marks all notifications as read for the current user.           //
    if (!Meteor.userId()) {                                            // 223
      throw new Meteor.Error(403, 'User not logged in.');              // 224
    }                                                                  //
                                                                       //
    Notifications.update({ userId: Meteor.userId() }, { $set: { isUnread: false } }, { multi: true });
  }                                                                    //
});                                                                    //
                                                                       //
saveFrontendRef = function (frontendRef, owner, requirements) {        // 231
  return inMeteor(function () {                                        // 232
    var sturdyRef = new Buffer(generateSturdyRef());                   // 233
    var hashedSturdyRef = hashSturdyRef(sturdyRef);                    // 234
    ApiTokens.insert({                                                 // 235
      _id: hashedSturdyRef,                                            // 236
      frontendRef: frontendRef,                                        // 237
      owner: owner,                                                    // 238
      created: new Date(),                                             // 239
      requirements: requirements                                       // 240
    });                                                                //
    return { sturdyRef: sturdyRef };                                   // 242
  });                                                                  //
};                                                                     //
                                                                       //
checkRequirements = function (requirements) {                          // 246
  if (!requirements) {                                                 // 247
    return true;                                                       // 248
  }                                                                    //
                                                                       //
  for (var i in babelHelpers.sanitizeForInObject(requirements)) {      // 251
    var requirement = requirements[i];                                 // 252
    if (requirement.tokenValid) {                                      // 253
      var token = ApiTokens.findOne({ _id: requirement.tokenValid }, { fields: { requirements: 1 } });
      if (!checkRequirements(token.requirements)) {                    // 255
        return false;                                                  // 256
      }                                                                //
    } else if (requirement.permissionsHeld) {                          //
      var p = requirement.permissionsHeld;                             // 259
      var viewInfo = Grains.findOne(p.grainId, { fields: { cachedViewInfo: 1 } }).cachedViewInfo;
      var currentPermissions = SandstormPermissions.grainPermissions(globalDb, { grain: { _id: p.grainId, identityId: p.identityId } }, viewInfo || {});
      if (!currentPermissions) {                                       // 263
        return false;                                                  // 264
      }                                                                //
                                                                       //
      for (var ii = 0; ii < p.permissions.length; ++ii) {              // 267
        if (p.permissions[ii] && !currentPermissions[ii]) {            // 268
          return false;                                                // 269
        }                                                              //
      }                                                                //
    } else if (requirement.userIsAdmin) {                              //
      if (!isAdminById(requirement.userIsAdmin)) {                     // 273
        return false;                                                  // 274
      }                                                                //
    } else {                                                           //
      throw new Meteor.Error(403, 'Unknown requirement');              // 277
    }                                                                  //
  }                                                                    //
                                                                       //
  return true;                                                         // 281
};                                                                     //
                                                                       //
restoreInternal = function (tokenId, ownerPattern, requirements, parentToken) {
  // Restores `sturdyRef`, checking first that its owner matches `ownerPattern`.
  // parentToken and requirements are optional params that are only used in the case of an objectId
  // token                                                             //
  var token = ApiTokens.findOne(tokenId);                              // 288
  if (!token) {                                                        // 289
    throw new Meteor.Error(403, 'No token found to restore');          // 290
  }                                                                    //
                                                                       //
  if (token.revoked) {                                                 // 293
    throw new Meteor.Error(403, 'Token has been revoked');             // 294
  }                                                                    //
                                                                       //
  check(token.owner, ownerPattern);                                    // 297
  if (!checkRequirements(token.requirements)) {                        // 298
    throw new Meteor.Error(403, 'Requirements not satisfied.');        // 299
  }                                                                    //
                                                                       //
  if (token.expires && token.expires.getTime() <= Date.now()) {        // 302
    throw new Meteor.Error(403, 'Authorization token expired');        // 303
  }                                                                    //
                                                                       //
  if (token.expiresIfUnused) {                                         // 306
    if (token.expiresIfUnused.getTime() <= Date.now()) {               // 307
      throw new Meteor.Error(403, 'Authorization token expired');      // 308
    } else {                                                           //
      // It's getting used now, so clear the expiresIfUnused field.    //
      ApiTokens.update(token._id, { $set: { expiresIfUnused: null } });
    }                                                                  //
  }                                                                    //
                                                                       //
  if (token.frontendRef) {                                             // 315
    if (token.frontendRef.notificationHandle) {                        // 316
      var notificationId = token.frontendRef.notificationHandle;       // 317
      return { cap: makeNotificationHandle(notificationId, true) };    // 318
    } else if (token.frontendRef.ipNetwork) {                          //
      return { cap: makeIpNetwork(tokenId) };                          // 320
    } else if (token.frontendRef.ipInterface) {                        //
      return { cap: makeIpInterface(tokenId) };                        // 322
    } else {                                                           //
      throw new Meteor.Error(500, 'Unknown frontend token type.');     // 324
    }                                                                  //
  } else if (token.objectId) {                                         //
    if (!checkRequirements(requirements)) {                            // 327
      throw new Meteor.Error(403, 'Requirements not satisfied.');      // 328
    }                                                                  //
                                                                       //
    if (token.objectId.appRef) {                                       // 331
      token.objectId.appRef = new Buffer(token.objectId.appRef);       // 332
    }                                                                  //
                                                                       //
    return waitPromise(globalBackend.useGrain(token.grainId, function (supervisor) {
      return supervisor.restore(token.objectId, requirements, parentToken);
    }));                                                               //
  } else if (token.parentToken) {                                      //
    return restoreInternal(token.parentToken, Match.Any, requirements, parentToken);
  } else {                                                             //
    throw new Meteor.Error(500, 'Unknown token type.');                // 341
  }                                                                    //
};                                                                     //
                                                                       //
function dropInternal(sturdyRef, ownerPattern) {                       // 345
  // Drops `sturdyRef`, checking first that its owner matches `ownerPattern`.
                                                                       //
  var hashedSturdyRef = hashSturdyRef(sturdyRef);                      // 348
  var token = ApiTokens.findOne({ _id: hashedSturdyRef });             // 349
  if (!token) {                                                        // 350
    return;                                                            // 351
  }                                                                    //
                                                                       //
  check(token.owner, ownerPattern);                                    // 354
                                                                       //
  if (token.frontendRef) {                                             // 356
    if (token.frontendRef.notificationHandle) {                        // 357
      var notificationId = token.frontendRef.notificationHandle;       // 358
      ApiTokens.remove({ _id: hashedSturdyRef });                      // 359
      var anyToken = ApiTokens.findOne({ 'frontendRef.notificationHandle': notificationId });
      if (!anyToken) {                                                 // 361
        // No other tokens referencing this notification exist, so dismiss the notification
        dismissNotification(notificationId);                           // 363
      }                                                                //
    } else {                                                           //
      throw new Error('Unknown frontend token type.');                 // 366
    }                                                                  //
  } else if (token.objectId) {                                         //
    if (token.objectId.wakeLockNotification) {                         // 369
      dropWakelock(token.grainId, token.objectId.wakeLockNotification);
    } else {                                                           //
      throw new Error('Unknown objectId token type.');                 // 372
    }                                                                  //
  } else {                                                             //
    throw new Error('Unknown token type.');                            // 375
  }                                                                    //
}                                                                      //
                                                                       //
makeChildTokenInternal = function (hashedParent, owner, requirements, grainId) {
  var sturdyRef = new Buffer(generateSturdyRef());                     // 380
  var hashedSturdyRef = hashSturdyRef(sturdyRef);                      // 381
                                                                       //
  requirements = requirements.filter(function (requirement) {          // 383
    return requirement.tokenValid !== hashedParent;                    // 384
  });                                                                  //
                                                                       //
  var tokenInfo = {                                                    // 387
    _id: hashedSturdyRef,                                              // 388
    parentToken: hashedParent,                                         // 389
    owner: owner,                                                      // 390
    created: new Date(),                                               // 391
    requirements: requirements                                         // 392
  };                                                                   //
  if (grainId) {                                                       // 394
    tokenInfo.grainId = grainId;                                       // 395
  }                                                                    //
                                                                       //
  ApiTokens.insert(tokenInfo);                                         // 398
                                                                       //
  return {                                                             // 400
    token: sturdyRef                                                   // 401
  };                                                                   //
};                                                                     //
                                                                       //
function SandstormCoreFactoryImpl() {}                                 // 405
                                                                       //
SandstormCoreFactoryImpl.prototype.getSandstormCore = function (grainId) {
  return { core: makeSandstormCore(grainId) };                         // 409
};                                                                     //
                                                                       //
makeSandstormCoreFactory = function () {                               // 412
  return new Capnp.Capability(new SandstormCoreFactoryImpl(), SandstormCoreFactory);
};                                                                     //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=core.js.map
