(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/drivers/ip.js                                                //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Future = Npm.require('fibers/future');                             // 17
var Net = Npm.require('net');                                          // 18
var Dgram = Npm.require('dgram');                                      // 19
var Promise = Npm.require('es6-promise').Promise;                      // 20
var Capnp = Npm.require('capnp');                                      // 21
                                                                       //
var IpRpc = Capnp.importSystem('sandstorm/ip.capnp');                  // 23
                                                                       //
ByteStreamConnection = (function () {                                  // 25
  function ByteStreamConnection(connection) {                          // 26
    babelHelpers.classCallCheck(this, ByteStreamConnection);           //
                                                                       //
    this.connection = connection;                                      // 27
  }                                                                    //
                                                                       //
  ByteStreamConnection.prototype.done = (function () {                 // 25
    function done() {                                                  // 30
      this.connection.end();                                           // 31
    }                                                                  //
                                                                       //
    return done;                                                       //
  })();                                                                //
                                                                       //
  ByteStreamConnection.prototype.write = (function () {                // 25
    function write(data) {                                             // 34
      this.connection.write(data);                                     // 35
    }                                                                  //
                                                                       //
    return write;                                                      //
  })();                                                                //
                                                                       //
  // expectSize not implemented                                        //
  // expectSize(size) { }                                              //
  return ByteStreamConnection;                                         //
})();                                                                  //
                                                                       //
IpInterfaceImpl = (function () {                                       // 42
  function IpInterfaceImpl(parentToken) {                              // 43
    babelHelpers.classCallCheck(this, IpInterfaceImpl);                //
                                                                       //
    this.parentToken = parentToken;                                    // 44
  }                                                                    //
                                                                       //
  IpInterfaceImpl.prototype.save = (function () {                      // 42
    function save(params) {                                            // 47
      var _this = this;                                                // 48
      return inMeteor(function () {                                    // 49
        var ret = makeChildTokenInternal(_this.parentToken, params.sealFor, []);
        return { sturdyRef: ret.token };                               // 51
      });                                                              //
    }                                                                  //
                                                                       //
    return save;                                                       //
  })();                                                                //
                                                                       //
  IpInterfaceImpl.prototype.listenTcp = (function () {                 // 42
    function listenTcp(portNum, port) {                                // 55
      return new Promise(function (resolve, reject) {                  // 56
        var resolved = false;                                          // 57
        var server = Net.createServer(function (connection) {          // 58
          var wrappedConnection = new ByteStreamConnection(connection);
          var upstream = port.connect(wrappedConnection).upstream;     // 60
                                                                       //
          connection.on('data', function (data) {                      // 62
            upstream.write(data);                                      // 63
          });                                                          //
                                                                       //
          connection.on('close', function (hadError) {                 // 66
            upstream.done();                                           // 67
          });                                                          //
                                                                       //
          connection.on('error', function (err) {                      // 70
            if (resolved) {                                            // 71
              connection.write = errorWrite;                           // 72
            } else {                                                   //
              // upstream hasn't been resolved yet, so it's safe to reject
              reject(err);                                             // 75
            }                                                          //
          });                                                          //
        });                                                            //
                                                                       //
        server.listen(portNum, function () {                           // 80
          resolved = true;                                             // 81
          resolve({ handle: server }); // server has a close method which is all we want from a handle
        });                                                            //
                                                                       //
        server.on('error', function (err) {                            // 85
          if (!resolved) {                                             // 86
            reject(err);                                               // 87
          }                                                            //
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return listenTcp;                                                  //
  })();                                                                //
                                                                       //
  IpInterfaceImpl.prototype.listenUdp = (function () {                 // 42
    function listenUdp(portNum, port) {                                // 93
      return new Promise(function (resolve, reject) {                  // 94
        var portMap = {};                                              // 95
        var resolved = false;                                          // 96
        var server = Dgram.createSocket('udp4'); // TODO(someday): handle ipv6 sockets too
        server.bind(portNum);                                          // 98
                                                                       //
        server.on('listening', function () {                           // 100
          // Although UDP is connectionless, we don't resolve until here so that we can handle bind
          // errors such as invalid host                               //
          resolved = true;                                             // 103
          resolve({ handle: server }); // server has a close method which is all we want from a handle
        });                                                            //
                                                                       //
        server.on('error', function (err) {                            // 107
          // TODO(someday): do something about errors after the promise is resolved
          if (!resolved) {                                             // 109
            reject(err);                                               // 110
          } else {                                                     //
            console.error('error in listenUdp: ' + err);               // 112
          }                                                            //
        });                                                            //
                                                                       //
        var returnMap = {};                                            // 116
        server.on('message', function (msg, rinfo) {                   // 117
          var address = rinfo.address + ']:' + rinfo.port;             // 118
          var returnPort = returnMap[address];                         // 119
                                                                       //
          if (!returnPort) {                                           // 121
            returnMap[address] = returnPort = new BoundUdpPortImpl(server, rinfo.address, rinfo.port);
          }                                                            //
                                                                       //
          port.send(msg, returnPort);                                  // 125
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return listenUdp;                                                  //
  })();                                                                //
                                                                       //
  return IpInterfaceImpl;                                              //
})();                                                                  //
                                                                       //
makeIpInterface = function (parentToken, grainId) {                    // 131
  return new Capnp.Capability(new IpInterfaceImpl(parentToken), IpRpc.PersistentIpInterface);
};                                                                     //
                                                                       //
BoundUdpPortImpl = (function () {                                      // 135
  function BoundUdpPortImpl(server, address, port) {                   // 136
    babelHelpers.classCallCheck(this, BoundUdpPortImpl);               //
                                                                       //
    this.server = server;                                              // 137
    this.address = address;                                            // 138
    this.port = port;                                                  // 139
  }                                                                    //
                                                                       //
  BoundUdpPortImpl.prototype.send = (function () {                     // 135
    function send(message, returnPort) {                               // 142
      // TODO(someday): this whole class is a hack to deal with the fact that we can't compare
      // capabilities or build a map with them. What we should be doing is mapping all ports to
      // their raw physical address/port, and using that here          //
      this.server.send(message, 0, message.length, this.port, this.address);
    }                                                                  //
                                                                       //
    return send;                                                       //
  })();                                                                //
                                                                       //
  return BoundUdpPortImpl;                                             //
})();                                                                  //
                                                                       //
var bits16 = Bignum(1).shiftLeft(16).sub(1);                           // 150
var bits32 = Bignum(1).shiftLeft(32).sub(1);                           // 151
                                                                       //
var intToIpv4 = function (num) {                                       // 153
  var part1 = num & 255;                                               // 154
  var part2 = num >> 8 & 255;                                          // 155
  var part3 = num >> 16 & 255;                                         // 156
  var part4 = num >> 24 & 255;                                         // 157
                                                                       //
  return part4 + '.' + part3 + '.' + part2 + '.' + part1;              // 159
};                                                                     //
                                                                       //
var addressToString = function (address) {                             // 162
  var ipv6num = Bignum(address.upper64).shiftLeft(64).add(Bignum(address.lower64));
                                                                       //
  if (ipv6num.shiftRight(32).eq(bits16)) {                             // 165
    // this is an ipv4 address, we should return it as such            //
    var ipv4num = ipv6num.and(bits32).toNumber();                      // 167
    return intToIpv4(ipv4num);                                         // 168
  }                                                                    //
                                                                       //
  var hex = ipv6num.toString(16);                                      // 171
  var numColons = 0;                                                   // 172
  var out = '';                                                        // 173
                                                                       //
  for (var i = 0; i < hex.length; ++i) {                               // 175
    // start with lower bits of address and build the output in reverse
    // this ensures that we can place a colon every 4 characters       //
    out += hex[hex.length - 1 - i];                                    // 178
    if ((i + 1) % 4 === 0) {                                           // 179
      out += ':';                                                      // 180
      ++numColons;                                                     // 181
    }                                                                  //
  }                                                                    //
                                                                       //
  // Double colon represents all bits being 0                          //
  if (numColons < 7) {                                                 // 186
    out += '::';                                                       // 187
  }                                                                    //
                                                                       //
  return out.split('').reverse().join('');                             // 190
};                                                                     //
                                                                       //
var addressType = function (address) {                                 // 193
  var type = 'udp4';                                                   // 194
                                                                       //
  // Check if it's an ipv6 address                                     //
  // TODO(someday): make this less hacky and change address to explicitly pass this information
  if (address.indexOf(':') != -1) {                                    // 198
    type = 'udp6';                                                     // 199
  }                                                                    //
                                                                       //
  return type;                                                         // 202
};                                                                     //
                                                                       //
IpNetworkImpl = (function () {                                         // 205
  function IpNetworkImpl(parentToken) {                                // 206
    babelHelpers.classCallCheck(this, IpNetworkImpl);                  //
                                                                       //
    this.parentToken = parentToken;                                    // 207
  }                                                                    //
                                                                       //
  IpNetworkImpl.prototype.save = (function () {                        // 205
    function save(params) {                                            // 210
      var _this = this;                                                // 211
      return inMeteor(function () {                                    // 212
        var ret = makeChildTokenInternal(_this.parentToken, params.sealFor, []);
        return { sturdyRef: ret.token };                               // 214
      });                                                              //
    }                                                                  //
                                                                       //
    return save;                                                       //
  })();                                                                //
                                                                       //
  IpNetworkImpl.prototype.getRemoteHost = (function () {               // 205
    function getRemoteHost(address) {                                  // 218
      return { host: new IpRemoteHostImpl(address) };                  // 219
    }                                                                  //
                                                                       //
    return getRemoteHost;                                              //
  })();                                                                //
                                                                       //
  IpNetworkImpl.prototype.getRemoteHostByName = (function () {         // 205
    function getRemoteHostByName(address) {                            // 222
      return { host: new IpRemoteHostImpl(address) };                  // 223
    }                                                                  //
                                                                       //
    return getRemoteHostByName;                                        //
  })();                                                                //
                                                                       //
  return IpNetworkImpl;                                                //
})();                                                                  //
                                                                       //
makeIpNetwork = function (parentToken, grainId) {                      // 227
  return new Capnp.Capability(new IpNetworkImpl(parentToken), IpRpc.PersistentIpNetwork);
};                                                                     //
                                                                       //
IpRemoteHostImpl = (function () {                                      // 231
  function IpRemoteHostImpl(address) {                                 // 232
    babelHelpers.classCallCheck(this, IpRemoteHostImpl);               //
                                                                       //
    if (address.upper64 || address.upper64 === 0) {                    // 233
      // address is an ip.capnp:IpAddress, we need to convert it       //
      this.address = addressToString(address);                         // 235
    } else {                                                           //
      this.address = address;                                          // 237
    }                                                                  //
  }                                                                    //
                                                                       //
  IpRemoteHostImpl.prototype.getTcpPort = (function () {               // 231
    function getTcpPort(portNum) {                                     // 241
      return { port: new TcpPortImpl(this.address, portNum) };         // 242
    }                                                                  //
                                                                       //
    return getTcpPort;                                                 //
  })();                                                                //
                                                                       //
  IpRemoteHostImpl.prototype.getUdpPort = (function () {               // 231
    function getUdpPort(portNum) {                                     // 245
      return { port: new UdpPortImpl(this.address, portNum) };         // 246
    }                                                                  //
                                                                       //
    return getUdpPort;                                                 //
  })();                                                                //
                                                                       //
  return IpRemoteHostImpl;                                             //
})();                                                                  //
                                                                       //
TcpPortImpl = (function () {                                           // 250
  function TcpPortImpl(address, portNum) {                             // 251
    babelHelpers.classCallCheck(this, TcpPortImpl);                    //
                                                                       //
    this.address = address;                                            // 252
    this.port = portNum;                                               // 253
  }                                                                    //
                                                                       //
  TcpPortImpl.prototype.connect = (function () {                       // 250
    function connect(downstream) {                                     // 256
      var _this = this;                                                // 257
      var resolved = false;                                            // 258
      return new Promise(function (resolve, reject) {                  // 259
        var client = Net.connect({ host: _this.address, port: _this.port }, function () {
          resolved = true;                                             // 261
          resolve({ upstream: new ByteStreamConnection(client) });     // 262
        });                                                            //
                                                                       //
        client.on('data', function (data) {                            // 265
          downstream.write(data);                                      // 266
        });                                                            //
                                                                       //
        client.on('close', function (hadError) {                       // 269
          downstream.done();                                           // 270
        });                                                            //
                                                                       //
        client.on('error', function (err) {                            // 273
          if (resolved) {                                              // 274
            client.write = errorWrite;                                 // 275
          } else {                                                     //
            // upstream hasn't been resolved yet, so it's safe to reject
            reject(err);                                               // 278
          }                                                            //
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return connect;                                                    //
  })();                                                                //
                                                                       //
  return TcpPortImpl;                                                  //
})();                                                                  //
                                                                       //
var errorWrite = function (data) {                                     // 285
  throw new Error('error occurred in connection');                     // 286
};                                                                     //
                                                                       //
UdpPortImpl = (function () {                                           // 289
  function UdpPortImpl(address, portNum) {                             // 290
    babelHelpers.classCallCheck(this, UdpPortImpl);                    //
                                                                       //
    this.address = address;                                            // 291
    this.port = portNum;                                               // 292
                                                                       //
    var type = addressType(address);                                   // 294
    this.socket = Dgram.createSocket(type);                            // 295
                                                                       //
    // TODO(someday): close socket after a certain time of inactivity?
    // This may be pointless since grains are killed frequently when not in use anyways
                                                                       //
    // Temporary hack. We only expect clients to pass in a single return port, so we'll store it
    // and only send replies here.                                     //
    // This will be changed to be correct when equality comparisons are added to capabilities.
    this.returnPort = null;                                            // 303
                                                                       //
    var _this = this;                                                  // 305
    this.socket.on('message', function (msg, rinfo) {                  // 306
      if (_this.returnPort) {                                          // 307
        _this.returnPort.send(msg, _this);                             // 308
      }                                                                //
    });                                                                //
  }                                                                    //
                                                                       //
  UdpPortImpl.prototype.send = (function () {                          // 289
    function send(message, returnPort) {                               // 313
      this.returnPort = returnPort;                                    // 314
      this.socket.send(message, 0, message.length, this.port, this.address);
                                                                       //
      // TODO(someday): use callback to catch errors and do something with them
    }                                                                  //
                                                                       //
    return send;                                                       //
  })();                                                                //
                                                                       //
  return UdpPortImpl;                                                  //
})();                                                                  //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=ip.js.map
