(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/hack-session.js                                              //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require('crypto');                                    // 17
var Http = Npm.require('http');                                        // 18
var Https = Npm.require('https');                                      // 19
var Future = Npm.require('fibers/future');                             // 20
var Net = Npm.require('net');                                          // 21
var Dgram = Npm.require('dgram');                                      // 22
var Promise = Npm.require('es6-promise').Promise;                      // 23
var Capnp = Npm.require('capnp');                                      // 24
                                                                       //
var EmailRpc = Capnp.importSystem('sandstorm/email.capnp');            // 26
var HackSessionContext = Capnp.importSystem('sandstorm/hack-session.capnp').HackSessionContext;
var PowerboxCapability = Capnp.importSystem('sandstorm/grain.capnp').PowerboxCapability;
var Supervisor = Capnp.importSystem('sandstorm/supervisor.capnp').Supervisor;
var SystemPersistent = Capnp.importSystem('sandstorm/supervisor.capnp').SystemPersistent;
var IpRpc = Capnp.importSystem('sandstorm/ip.capnp');                  // 31
var EmailSendPort = EmailRpc.EmailSendPort;                            // 32
                                                                       //
var Url = Npm.require('url');                                          // 34
                                                                       //
ROOT_URL = Url.parse(process.env.ROOT_URL);                            // 36
HOSTNAME = ROOT_URL.hostname;                                          // 37
                                                                       //
SessionContextImpl = (function () {                                    // 39
  function SessionContextImpl(grainId, sessionId, identityId) {        // 40
    babelHelpers.classCallCheck(this, SessionContextImpl);             //
                                                                       //
    this.grainId = grainId;                                            // 41
    this.sessionId = sessionId;                                        // 42
    this.identityId = identityId;                                      // 43
  }                                                                    //
                                                                       //
  SessionContextImpl.prototype.offer = (function () {                  // 39
    function offer(cap, requiredPermissions) {                         // 46
      var _this = this;                                                // 47
      return inMeteor(function () {                                    // 48
        if (!_this.identityId) {                                       // 49
          // TODO(soon): allow non logged in users?                    //
          throw new Meteor.Error(400, 'Only logged in users can offer capabilities.');
        }                                                              //
                                                                       //
        var castedCap = cap.castAs(SystemPersistent);                  // 54
        var save = castedCap.save({ webkey: null });                   // 55
        var sturdyRef = waitPromise(save).sturdyRef;                   // 56
                                                                       //
        // TODO(soon): This will eventually use SystemPersistent.addRequirements when membranes
        // are fully implemented for supervisors.                      //
        var requirement = {                                            // 60
          permissionsHeld: {                                           // 61
            grainId: _this.grainId,                                    // 62
            identityId: _this.identityId,                              // 63
            permissions: requiredPermissions                           // 64
          }                                                            //
        };                                                             //
                                                                       //
        if (!checkRequirements([requirement])) {                       // 68
          throw new Meteor.Error(403, 'Permissions not satisfied.');   // 69
        }                                                              //
                                                                       //
        ApiTokens.update({ _id: hashSturdyRef(sturdyRef) }, { $push: { requirements: requirement } });
        Sessions.update({ _id: _this.sessionId }, {                    // 73
          $set: {                                                      // 75
            powerboxView: {                                            // 76
              offer: ROOT_URL.protocol + '//' + makeWildcardHost('api') + '#' + sturdyRef
            }                                                          //
          }                                                            //
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return offer;                                                      //
  })();                                                                //
                                                                       //
  return SessionContextImpl;                                           //
})();                                                                  //
                                                                       //
Meteor.methods({                                                       // 86
  finishPowerboxRequest: function (webkeyUrl, saveLabel, identityId, grainId) {
    check(webkeyUrl, String);                                          // 88
    check(saveLabel, Match.OneOf(undefined, null, String));            // 89
    check(identityId, String);                                         // 90
    check(grainId, String);                                            // 91
                                                                       //
    var userId = Meteor.userId();                                      // 93
    if (!userId || !globalDb.userHasIdentity(userId, identityId)) {    // 94
      throw new Meteor.Error(403, 'Not an identity of the current user: ' + identityId);
    }                                                                  //
                                                                       //
    var parsedWebkey = Url.parse(webkeyUrl.trim());                    // 98
    if (parsedWebkey.host !== makeWildcardHost('api')) {               // 99
      console.log(parsedWebkey.hostname, makeWildcardHost('api'));     // 100
      throw new Meteor.Error(500, 'Hostname does not match this server. External webkeys are not ' + 'supported (yet)');
    }                                                                  //
                                                                       //
    var token = parsedWebkey.hash;                                     // 105
    if (!token) {                                                      // 106
      throw new Meteor.Error(400, 'Invalid webkey. You must pass in the full webkey, ' + 'including domain name and hash fragment');
    }                                                                  //
                                                                       //
    token = token.slice(1);                                            // 111
    var cap = restoreInternal(hashSturdyRef(token), Match.Optional({ webkey: Match.Optional(Match.Any) }), [], new Buffer(token)).cap;
    var castedCap = cap.castAs(SystemPersistent);                      // 115
    var grainOwner = {                                                 // 116
      grainId: grainId,                                                // 117
      introducerIdentity: identityId                                   // 118
    };                                                                 //
    if (saveLabel) {                                                   // 120
      grainOwner.saveLabel = {                                         // 121
        defaultText: saveLabel                                         // 122
      };                                                               //
    }                                                                  //
                                                                       //
    var save = castedCap.save({ grain: grainOwner });                  // 126
    var sturdyRef = waitPromise(save).sturdyRef;                       // 127
                                                                       //
    return sturdyRef.toString();                                       // 129
  },                                                                   //
                                                                       //
  finishPowerboxOffer: function (sessionId) {                          // 132
    check(sessionId, String);                                          // 133
                                                                       //
    Sessions.update({ _id: sessionId }, { $unset: { powerboxView: null } });
  }                                                                    //
});                                                                    //
                                                                       //
HackSessionContextImpl = (function (_SessionContextImpl) {             // 139
  babelHelpers.inherits(HackSessionContextImpl, _SessionContextImpl);  //
                                                                       //
  function HackSessionContextImpl(grainId, sessionId, identityId) {    // 140
    babelHelpers.classCallCheck(this, HackSessionContextImpl);         //
                                                                       //
    _SessionContextImpl.call(this, grainId, sessionId, identityId);    // 141
  }                                                                    //
                                                                       //
  HackSessionContextImpl.prototype._getPublicId = (function () {       // 139
    function _getPublicId() {                                          // 144
      // Get the grain's public ID, assigning a new one if it doesn't yet have one.
      //                                                               //
      // Must be called in a Meteor context.                           //
                                                                       //
      while (!this.publicId) {                                         // 149
        // We haven't looked up the public ID yet.                     //
        var grain = Grains.findOne(this.grainId, { fields: { publicId: 1 } });
        if (!grain) throw new Error('Grain does not exist.');          // 152
                                                                       //
        if (grain.publicId) {                                          // 154
          this.publicId = grain.publicId;                              // 155
        } else {                                                       //
          // The grain doesn't have a public ID yet. Generate one.     //
          var candidate = generateRandomHostname(20);                  // 158
                                                                       //
          if (Grains.findOne({ publicId: candidate })) {               // 160
            // This should never ever happen.                          //
            console.error('CRITICAL PROBLEM: Public ID collision. ' + 'CSPRNG is bad or has insufficient entropy.');
            continue;                                                  // 164
          }                                                            //
                                                                       //
          // Carefully perform an update that becomes a no-op if anyone else has assigned a public ID
          // simultaneously.                                           //
          if (Grains.update({ _id: this.grainId, publicId: { $exists: false } }, { $set: { publicId: candidate } }) > 0) {
            // We won the race.                                        //
            this.publicId = candidate;                                 // 172
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      return this.publicId;                                            // 177
    }                                                                  //
                                                                       //
    return _getPublicId;                                               //
  })();                                                                //
                                                                       //
  HackSessionContextImpl.prototype._getAddress = (function () {        // 139
    function _getAddress() {                                           // 180
      // Get the grain's outgoing e-mail address.                      //
      //                                                               //
      // Must be called in a Meteor context.                           //
                                                                       //
      return this._getPublicId() + '@' + HOSTNAME;                     // 185
    }                                                                  //
                                                                       //
    return _getAddress;                                                //
  })();                                                                //
                                                                       //
  HackSessionContextImpl.prototype._getUserAddress = (function () {    // 139
    function _getUserAddress() {                                       // 188
      // Get the user's e-mail address.                                //
      //                                                               //
      // Must be called in a Meteor context.                           //
                                                                       //
      var grain = Grains.findOne(this.grainId, { fields: { identityId: 1 } });
                                                                       //
      var identity = globalDb.getIdentity(grain.identityId);           // 195
                                                                       //
      var emails = SandstormDb.getVerifiedEmails(identity);            // 197
      var email = emails.length > 0 && emails[0] || identity.unverifiedEmail;
                                                                       //
      var result = {};                                                 // 200
      if (email) {                                                     // 201
        result.address = email;                                        // 202
      }                                                                //
                                                                       //
      if (identity.profile.name) {                                     // 205
        result.name = identity.profile.name;                           // 206
      }                                                                //
                                                                       //
      return result;                                                   // 209
    }                                                                  //
                                                                       //
    return _getUserAddress;                                            //
  })();                                                                //
                                                                       //
  HackSessionContextImpl.prototype.send = (function () {               // 139
    function send(email) {                                             // 212
      return hackSendEmail(this, email);                               // 213
    }                                                                  //
                                                                       //
    return send;                                                       //
  })();                                                                //
                                                                       //
  HackSessionContextImpl.prototype.getPublicId = (function () {        // 139
    function getPublicId() {                                           // 216
      return inMeteor((function () {                                   // 217
        var result = {};                                               // 218
                                                                       //
        result.publicId = this._getPublicId();                         // 220
        result.hostname = HOSTNAME;                                    // 221
        result.autoUrl = ROOT_URL.protocol + '//' + makeWildcardHost(result.publicId);
                                                                       //
        var grain = Grains.findOne(this.grainId, { fields: { userId: 1 } });
        result.isDemoUser = Meteor.users.findOne(grain.userId).expires ? true : false;
                                                                       //
        return result;                                                 // 227
      }).bind(this));                                                  //
    }                                                                  //
                                                                       //
    return getPublicId;                                                //
  })();                                                                //
                                                                       //
  HackSessionContextImpl.prototype.httpGet = (function () {            // 139
    function httpGet(url) {                                            // 231
      var _this = this;                                                // 232
      var session = _this;                                             // 233
                                                                       //
      return new Promise(function (resolve, reject) {                  // 235
        var requestMethod = Http.request;                              // 236
        if (url.indexOf('https://') === 0) {                           // 237
          requestMethod = Https.request;                               // 238
        } else if (url.indexOf('http://') !== 0) {                     //
          err = new Error('Protocol not recognized.');                 // 240
          err.nature = 'precondition';                                 // 241
          reject(err);                                                 // 242
        }                                                              //
                                                                       //
        req = requestMethod(url, function (resp) {                     // 245
          var buffers = [];                                            // 246
          var err = undefined;                                         // 247
                                                                       //
          switch (Math.floor(resp.statusCode / 100)) {                 // 249
            case 2:                                                    // 250
              // 2xx response -- OK.                                   //
              resp.on('data', function (buf) {                         // 251
                buffers.push(buf);                                     // 252
              });                                                      //
                                                                       //
              resp.on('end', function () {                             // 255
                resolve({                                              // 256
                  content: Buffer.concat(buffers),                     // 257
                  mimeType: resp.headers['content-type'] || null       // 258
                });                                                    //
              });                                                      //
              break;                                                   // 261
            case 3:                                                    // 261
              // 3xx response -- redirect.                             //
              resolve(session.httpGet(resp.headers.location));         // 263
              break;                                                   // 264
            case 4:                                                    // 264
              // 4xx response -- client error.                         //
              err = new Error('Status code ' + resp.statusCode + ' received in response.');
              err.nature = 'precondition';                             // 267
              reject(err);                                             // 268
              break;                                                   // 269
            case 5:                                                    // 269
              // 5xx response -- internal server error.                //
              err = new Error('Status code ' + resp.statusCode + ' received in response.');
              err.nature = 'localBug';                                 // 272
              reject(err);                                             // 273
              break;                                                   // 274
            default:                                                   // 274
              // ???                                                   //
              err = new Error('Invalid status code ' + resp.statusCode + ' received in response.');
              err.nature = 'localBug';                                 // 277
              reject(err);                                             // 278
              break;                                                   // 279
          }                                                            // 279
        });                                                            //
                                                                       //
        req.on('error', function (e) {                                 // 283
          e.nature = 'networkFailure';                                 // 284
          reject(e);                                                   // 285
        });                                                            //
                                                                       //
        req.setTimeout(15000, function () {                            // 288
          req.abort();                                                 // 289
          err = new Error('Request timed out.');                       // 290
          err.nature = 'localBug';                                     // 291
          err.durability = 'overloaded';                               // 292
          reject(err);                                                 // 293
        });                                                            //
                                                                       //
        req.end();                                                     // 296
      });                                                              //
    }                                                                  //
                                                                       //
    return httpGet;                                                    //
  })();                                                                //
                                                                       //
  HackSessionContextImpl.prototype.getUserAddress = (function () {     // 139
    function getUserAddress() {                                        // 300
      return inMeteor((function () {                                   // 301
        return this._getUserAddress();                                 // 302
      }).bind(this));                                                  //
    }                                                                  //
                                                                       //
    return getUserAddress;                                             //
  })();                                                                //
                                                                       //
  HackSessionContextImpl.prototype.obsoleteGenerateApiToken = (function () {
    function obsoleteGenerateApiToken(petname, userInfo, expires) {    // 306
      throw new Error('generateApiToken() has been removed. Use offer templates instead.');
    }                                                                  //
                                                                       //
    return obsoleteGenerateApiToken;                                   //
  })();                                                                //
                                                                       //
  HackSessionContextImpl.prototype.obsoleteListApiTokens = (function () {
    function obsoleteListApiTokens() {                                 // 310
      throw new Error('listApiTokens() has been removed. Use offer templates instead.');
    }                                                                  //
                                                                       //
    return obsoleteListApiTokens;                                      //
  })();                                                                //
                                                                       //
  HackSessionContextImpl.prototype.obsoleteRevokeApiToken = (function () {
    function obsoleteRevokeApiToken(tokenId) {                         // 314
      throw new Error('revokeApiToken() has been removed. Use offer templates instead.');
    }                                                                  //
                                                                       //
    return obsoleteRevokeApiToken;                                     //
  })();                                                                //
                                                                       //
  HackSessionContextImpl.prototype.getUiViewForEndpoint = (function () {
    function getUiViewForEndpoint(url) {                               // 318
      var parsedUrl = Url.parse(url);                                  // 319
                                                                       //
      if (parsedUrl.hash) {                                            // 321
        // Assume that anything with a fragment is a webkey            //
        if (parsedUrl.pathname) {                                      // 322
          throw new Error('Webkey urls cannot contain a path.');       // 323
        }                                                              //
                                                                       //
        var apiHost = ROOT_URL.protocol + '//' + makeWildcardHost('api');
        var urlProtoAndHost = parsedUrl.protocol + '//' + parsedUrl.host;
        var token = parsedUrl.hash.slice(1); // Get rid of # which is always the first character
        if (urlProtoAndHost === apiHost) {                             // 329
          return getWrappedUiViewForToken(token);                      // 330
        } else {                                                       //
          return { view: new ExternalUiView(url, this.grainId, token) };
        }                                                              //
      } else {                                                         //
        return { view: new ExternalUiView(url, this.grainId) };        // 335
      }                                                                //
    }                                                                  //
                                                                       //
    return getUiViewForEndpoint;                                       //
  })();                                                                //
                                                                       //
  return HackSessionContextImpl;                                       //
})(SessionContextImpl);                                                //
                                                                       //
makeHackSessionContext = function (grainId, sessionId, identityId) {   // 340
  return new Capnp.Capability(new HackSessionContextImpl(grainId, sessionId, identityId), HackSessionContext);
};                                                                     //
                                                                       //
var HOSTNAME_CHARS = 'abcdefghijklmnopqrstuvwxyz0123456789';           // 345
                                                                       //
generateRandomHostname = function (length) {                           // 347
  // Generate a random unique name suitable for use in a hostname.     //
                                                                       //
  var digits = [];                                                     // 350
  for (var i = 0; i < length; i++) {                                   // 351
    digits[i] = Random.choice(HOSTNAME_CHARS);                         // 352
  }                                                                    //
                                                                       //
  return digits.join('');                                              // 355
};                                                                     //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=hack-session.js.map
