(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// server/proxy.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var Crypto = Npm.require('crypto');                                    // 17
var ChildProcess = Npm.require('child_process');                       // 18
var Fs = Npm.require('fs');                                            // 19
var Path = Npm.require('path');                                        // 20
var Future = Npm.require('fibers/future');                             // 21
var Http = Npm.require('http');                                        // 22
var Url = Npm.require('url');                                          // 23
var Promise = Npm.require('es6-promise').Promise;                      // 24
var Capnp = Npm.require('capnp');                                      // 25
var Net = Npm.require('net');                                          // 26
                                                                       //
var ByteStream = Capnp.importSystem('sandstorm/util.capnp').ByteStream;
var ApiSession = Capnp.importSystem('sandstorm/api-session.capnp').ApiSession;
var WebSession = Capnp.importSystem('sandstorm/web-session.capnp').WebSession;
var HackSession = Capnp.importSystem('sandstorm/hack-session.capnp');  // 31
var Supervisor = Capnp.importSystem('sandstorm/supervisor.capnp').Supervisor;
var Backend = Capnp.importSystem('sandstorm/backend.capnp').Backend;   // 33
                                                                       //
SANDSTORM_ALTHOME = Meteor.settings && Meteor.settings.home;           // 35
SANDSTORM_LOGDIR = (SANDSTORM_ALTHOME || '') + '/var/log';             // 36
SANDSTORM_VARDIR = (SANDSTORM_ALTHOME || '') + '/var/sandstorm';       // 37
                                                                       //
var storeReferralProgramInfoApiTokenCreated = function (db, accountId, identityId, apiTokenAccountId) {
  // From the Referral program's perspective, if Bob's Account has no referredByComplete, then we
  // update Bob's Identity to say it's referredBy Alice's Account (which is apiTokenAccountId).
  check(accountId, String);                                            // 42
  check(identityId, String);                                           // 43
  check(apiTokenAccountId, String);                                    // 44
                                                                       //
  // Bail out early if quota enforcement is disabled.                  //
  if (!Meteor.settings['public'].quotaEnabled) {                       // 47
    return;                                                            // 48
  }                                                                    //
                                                                       //
  var aliceAccountId = apiTokenAccountId;                              // 51
  var bobAccountId = accountId;                                        // 52
  var bobIdentityId = identityId;                                      // 53
                                                                       //
  if (Meteor.users.find({                                              // 55
    _id: bobAccountId,                                                 // 56
    referredByComplete: { $exists: true }                              // 57
  }).count() > 0) {                                                    //
    return;                                                            // 59
  }                                                                    //
                                                                       //
  // Only actually update Bob's Identity ID if there is no referredBy.
  Meteor.users.update({ _id: bobIdentityId, referredBy: { $exists: false } }, { $set: { referredBy: aliceAccountId } });
};                                                                     //
                                                                       //
function referralProgramLogSharingTokenUse(db, bobAccountId) {         // 68
  // Hooray! The sharing token is valid! Someone (let's call them Charlie) is going to get a UiView
  // to this grain!  This means that the user who created this apiToken knows how to use the 'share
  // access' interface. Let's call them Bob.                           //
  //                                                                   //
  // If Bob's Account.referredByComplete is not yet set, then look at Bob's Identities and take the
  // first referredBy we find -- let's call that Alice.                //
  //                                                                   //
  // We copy Alice's account ID to Bob's Account.referredByComplete, and then update Alice's
  // referredIdentityIds to point at Bob's Identity, and then remove the referredBy from Bob's
  // Identity since it has become redundant.                           //
  //                                                                   //
  // Implementation note: this does mean that Alice can get referral credit for Bob by sharing a
  // link with Bob, even if Bob already had an account.                //
                                                                       //
  // Bail out early if quota support is not enabled.                   //
  if (!Meteor.settings['public'].quotaEnabled) {                       // 84
    return;                                                            // 85
  }                                                                    //
                                                                       //
  // Bail out if Bob has a referredByComplete.                         //
  if (Meteor.users.find({ _id: bobAccountId, referredByComplete: { $exists: true } }).count() > 0) {
    return;                                                            // 90
  }                                                                    //
                                                                       //
  // Look for a referredBy on any of Bob's identities.                 //
  var bobIdentityIds = SandstormDb.getUserIdentityIds(Meteor.users.findOne({ _id: bobAccountId }));
  var bobIdentityWithReferredBy = Meteor.users.findOne({               // 95
    _id: { $in: bobIdentityIds },                                      // 96
    referredBy: { $exists: true }                                      // 97
  });                                                                  //
                                                                       //
  if (!bobIdentityWithReferredBy) {                                    // 100
    return;                                                            // 101
  }                                                                    //
                                                                       //
  var aliceAccountId = bobIdentityWithReferredBy.referredBy;           // 104
                                                                       //
  // Store Bob's Account.referralCompletedBy.                          //
  var now = new Date();                                                // 107
  Meteor.users.update({                                                // 108
    _id: bobAccountId,                                                 // 109
    referredByComplete: { $exists: false }                             // 110
  }, {                                                                 //
    $set: {                                                            // 112
      referredByComplete: bobIdentityWithReferredBy.referredBy,        // 113
      referredCompleteDate: now                                        // 114
    }                                                                  //
  });                                                                  //
                                                                       //
  // Update Alice's Account.referredIdentityIds.                       //
  Meteor.users.update({ _id: aliceAccountId }, {                       // 119
    $push: { referredIdentityIds: bobIdentityWithReferredBy._id }      // 120
  });                                                                  //
                                                                       //
  // Remove now-redundant Bob identity referredBy.                     //
  Meteor.users.update({ _id: bobIdentityWithReferredBy._id }, {        // 124
    $unset: { referredBy: true }                                       // 125
  });                                                                  //
}                                                                      //
                                                                       //
// User-agent strings that should be allowed to use http basic authentication.
// These are regex matches, so ensure they are escaped properly with double
// backslashes. For security reasons, we MUST NOT whitelist any user-agents
// that may render html and execute embedded scripts.                  //
BASIC_AUTH_USER_AGENTS = ['git\\/', 'GitHub-Hookshot\\/', 'mirall\\/', 'Mozilla\\/5\\.0 \\([^\\\\]*\\) mirall\\/', 'Mozilla\\/5\\.0 \\(iOS\\) ownCloud-iOS\\/', 'Mozilla\\/5\\.0 \\(Android\\) ownCloud-android\\/', 'litmus\\/'];
BASIC_AUTH_USER_AGENTS_REGEX = new RegExp('^(' + BASIC_AUTH_USER_AGENTS.join('|') + ')', '');
                                                                       //
var SESSION_PROXY_TIMEOUT = 60000;                                     // 144
                                                                       //
var sandstormCoreFactory = makeSandstormCoreFactory();                 // 146
var backendAddress = 'unix:' + (SANDSTORM_ALTHOME || '') + Backend.socketPath;
var sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);
var sandstormBackend = sandstormBackendConnection.restore(null, Backend);
                                                                       //
// TODO(cleanup): This initilization belongs with the rest of our package initialization in
//   db-deprecated.js. We can't put it there now because we need to contruct sandstormCoreFactory first.
globalBackend = new SandstormBackend(globalDb, sandstormBackend);      // 153
Meteor.onConnection(function (connection) {                            // 154
  connection.sandstormBackend = globalBackend;                         // 155
});                                                                    //
                                                                       //
// We've observed a problem in production where occasionally the front-end stops talking to the
// back-end. It happens very rarely -- like once a month -- and we've been unable to reproduce it
// in testing, making it very hard to debug. The problem appears both on Sandstorm and Blackrock.
// Restarting the node process (and only the node process) always fixes the problem.
//                                                                     //
// Here, I've added some code that attempts to detect the problem by doing a health check
// periodically and then remaking the connection if it seems broken. We'll see if this helps!
var backendHealthy = true;                                             // 165
var disconnectCount = 0;                                               // 166
Meteor.setInterval(function () {                                       // 167
  if (!backendHealthy) {                                               // 168
    if (disconnectCount++ > 2) process.abort();                        // 169
    console.error('error: Backend hasn\'t responded in 30 seconds! Reconnecting.');
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(true);
    sandstormBackendConnection.close();                                // 172
    sandstormBackendConnection = Capnp.connect(backendAddress, sandstormCoreFactory);
    sandstormBackend = sandstormBackendConnection.restore(null, Backend);
    globalBackend._backendCap = sandstormBackend;                      // 175
  }                                                                    //
                                                                       //
  var debugLog = !backendHealthy;                                      // 178
  backendHealthy = false;                                              // 179
  var promise = sandstormBackend.ping().then(function () {             // 180
    backendHealthy = true;                                             // 181
    if (Capnp.enableVerboseDebugLogging) Capnp.enableVerboseDebugLogging(false);
  }, function (err) {                                                  //
    console.error('error: Backend ping threw error!', err.stack);      // 184
    // The connection will be remade on the next interval. Note that we do NOT normally observe
    // exceptions being thrown for this problem; we see the connection simply stop responding.
    // So we don't expect this branch to execute in any case.          //
  });                                                                  //
                                                                       //
  if (debugLog) {                                                      // 190
    console.log('capnp.js: outer promise:', promise);                  // 191
  }                                                                    //
}, 30000);                                                             //
                                                                       //
// =======================================================================================
// Meteor context <-> Async Node.js context adapters                   //
// TODO(cleanup):  Move to a different file.                           //
                                                                       //
var inMeteorInternal = Meteor.bindEnvironment(function (callback) {    // 199
  callback();                                                          // 200
});                                                                    //
                                                                       //
inMeteor = function (callback) {                                       // 203
  // Calls the callback in a Meteor context.  Returns a Promise for its result.
  return new Promise(function (resolve, reject) {                      // 205
    inMeteorInternal(function () {                                     // 206
      try {                                                            // 207
        resolve(callback());                                           // 208
      } catch (err) {                                                  //
        reject(err);                                                   // 210
      }                                                                //
    });                                                                //
  });                                                                  //
};                                                                     //
                                                                       //
promiseToFuture = function (promise) {                                 // 216
  var result = new Future();                                           // 217
  promise.then(result['return'].bind(result), result['throw'].bind(result));
  return result;                                                       // 219
};                                                                     //
                                                                       //
waitPromise = function (promise) {                                     // 222
  return promiseToFuture(promise).wait();                              // 223
};                                                                     //
                                                                       //
// =======================================================================================
// API for creating / starting grains from Meteor methods.             //
                                                                       //
var proxiesByHostId = {};                                              // 229
                                                                       //
Meteor.methods({                                                       // 231
  newGrain: function (packageId, command, title, identityId) {         // 232
    // Create and start a new grain.                                   //
                                                                       //
    check(packageId, String);                                          // 235
    check(command, Object); // Manifest.Command from package.capnp.    // 236
    check(title, String);                                              // 237
    check(identityId, String);                                         // 238
                                                                       //
    if (!this.userId) {                                                // 240
      throw new Meteor.Error(403, 'Unauthorized', 'Must be logged in to create grains.');
    }                                                                  //
                                                                       //
    if (!globalDb.userHasIdentity(this.userId, identityId)) {          // 244
      throw new Meteor.Error(403, 'Current user does not own the identity: ' + identityId);
    }                                                                  //
                                                                       //
    if (!isSignedUpOrDemo()) {                                         // 248
      throw new Meteor.Error(403, 'Unauthorized', 'Only invited users or demo users can create grains.');
    }                                                                  //
                                                                       //
    if (isUserOverQuota(Meteor.user())) {                              // 253
      throw new Meteor.Error(402, 'You are out of storage space. Please delete some things and try again.');
    }                                                                  //
                                                                       //
    var pkg = Packages.findOne(packageId);                             // 258
    var isDev = false;                                                 // 259
    if (!pkg) {                                                        // 260
      // Maybe they wanted a dev package.  Check there too.            //
      pkg = DevPackages.findOne(packageId);                            // 262
      isDev = true;                                                    // 263
    }                                                                  //
                                                                       //
    if (!pkg) {                                                        // 266
      throw new Meteor.Error(404, 'Not Found', 'No such package is installed.');
    }                                                                  //
                                                                       //
    var appId = pkg.appId;                                             // 270
    var manifest = pkg.manifest;                                       // 271
    var grainId = Random.id(22); // 128 bits of entropy                // 272
    Grains.insert({                                                    // 273
      _id: grainId,                                                    // 274
      packageId: packageId,                                            // 275
      appId: appId,                                                    // 276
      appVersion: manifest.appVersion,                                 // 277
      userId: this.userId,                                             // 278
      identityId: identityId,                                          // 279
      title: title,                                                    // 280
      'private': true                                                  // 281
    });                                                                //
                                                                       //
    globalBackend.startGrainInternal(packageId, grainId, this.userId, command, true, isDev);
    globalBackend.updateLastActive(grainId, this.userId, identityId);  // 285
    return grainId;                                                    // 286
  },                                                                   //
                                                                       //
  openSession: function (grainId, identityId, cachedSalt) {            // 289
    // Open a new UI session on an existing grain.  Starts the grain if it is not already
    // running.                                                        //
                                                                       //
    check(grainId, String);                                            // 293
    check(identityId, Match.OneOf(undefined, null, String));           // 294
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 295
                                                                       //
    if (this.userId && identityId && !globalDb.userHasIdentity(this.userId, identityId)) {
      throw new Meteor.Error(403, 'Current user does not own the identity: ' + identityId);
    }                                                                  //
                                                                       //
    var db = this.connection.sandstormDb;                              // 301
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 302
    if (!SandstormPermissions.mayOpenGrain(db, { grain: { _id: grainId, identityId: identityId } })) {
      throw new Meteor.Error(403, 'Unauthorized', 'User is not authorized to open this grain.');
    }                                                                  //
                                                                       //
    return globalBackend.openSessionInternal(grainId, this.userId, identityId, null, null, cachedSalt);
  },                                                                   //
                                                                       //
  openSessionFromApiToken: function (params, identityId, cachedSalt) {
    // Given an API token, either opens a new WebSession to the underlying grain or returns a
    // path to which the client should redirect in order to open such a session.
                                                                       //
    check(params, {                                                    // 315
      token: String,                                                   // 316
      incognito: Boolean                                               // 317
    });                                                                //
    check(identityId, Match.OneOf(undefined, null, String));           // 319
    check(cachedSalt, Match.OneOf(undefined, null, String));           // 320
                                                                       //
    if (this.userId && identityId && !globalDb.userHasIdentity(this.userId, identityId)) {
      throw new Meteor.Error(403, 'Current user does not own the identity: ' + identityId);
    }                                                                  //
                                                                       //
    var token = params.token;                                          // 326
    var incognito = params.incognito;                                  // 327
    var hashedToken = Crypto.createHash('sha256').update(token).digest('base64');
    var apiToken = ApiTokens.findOne(hashedToken);                     // 329
    validateWebkey(apiToken);                                          // 330
    var grain = Grains.findOne({ _id: apiToken.grainId });             // 331
    if (!grain) {                                                      // 332
      throw new Meteor.Error(404, 'Grain not found', 'Grain ID: ' + apiToken.grainId);
    }                                                                  //
                                                                       //
    if (apiToken.accountId) {                                          // 336
      referralProgramLogSharingTokenUse(globalDb, apiToken.accountId);
    }                                                                  //
                                                                       //
    var title = undefined;                                             // 340
    if (grain.userId === apiToken.accountId) {                         // 341
      title = grain.title;                                             // 342
    } else {                                                           //
      var sharerToken = apiToken.identityId && ApiTokens.findOne({     // 344
        grainId: apiToken.grainId,                                     // 346
        'owner.user.identityId': apiToken.identityId                   // 347
      }, {                                                             //
        sort: {                                                        // 349
          'owner.user.lastUsed': -1                                    // 350
        }                                                              //
      });                                                              //
      if (sharerToken) {                                               // 353
        title = sharerToken.owner.user.title;                          // 354
      } else {                                                         //
        title = 'shared grain';                                        // 356
      }                                                                //
    }                                                                  //
                                                                       //
    if (this.userId && !incognito) {                                   // 360
      if (identityId != apiToken.identityId && identityId != grain.identityId && !ApiTokens.findOne({ 'owner.user.identityId': identityId, parentToken: hashedToken })) {
        var owner = { user: { identityId: identityId, title: title } };
                                                                       //
        // Create a new API token for the identity redeeming this token.
        var result = SandstormPermissions.createNewApiToken(globalDb, { rawParentToken: token }, apiToken.grainId, apiToken.petname, { allAccess: null }, owner);
        globalDb.addContact(apiToken.accountId, identityId);           // 368
                                                                       //
        // If the parent API token is forSharing and it has an accountId, then the logged-in user (call
        // them Bob) is about to access a grain owned by someone (call them Alice) and save a reference
        // to it as a new ApiToken. (For share-by-link, this occurs when viewing the grain. For
        // share-by-identity, this happens immediately.)               //
        if (result.parentApiToken) {                                   // 374
          var parentApiToken = result.parentApiToken;                  // 375
          if (parentApiToken.forSharing && parentApiToken.accountId) {
            storeReferralProgramInfoApiTokenCreated(globalDb, this.userId, owner.user.identityId, parentApiToken.accountId);
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      return { redirectToGrain: apiToken.grainId };                    // 383
    } else {                                                           //
      if (!SandstormPermissions.mayOpenGrain(globalDb, { token: apiToken })) {
        throw new Meteor.Error(403, 'Unauthorized', 'User is not authorized to open this grain.');
      }                                                                //
                                                                       //
      return globalBackend.openSessionInternal(apiToken.grainId, null, null, title, apiToken, cachedSalt);
    }                                                                  //
  },                                                                   //
                                                                       //
  keepSessionAlive: function (sessionId) {                             // 394
    // TODO(security):  Prevent draining someone else's quota by holding open several grains shared
    //   by them.                                                      //
    check(sessionId, String);                                          // 397
                                                                       //
    var session = Sessions.findAndModify({                             // 399
      query: { _id: sessionId },                                       // 400
      update: { $set: { timestamp: new Date().getTime() } },           // 401
      fields: { grainId: 1 }                                           // 402
    });                                                                //
                                                                       //
    if (session) {                                                     // 405
      // Session still present in database, so send keep-alive to backend.
      try {                                                            // 407
        var grainId = session.grainId;                                 // 408
        waitPromise(globalBackend.openGrain(grainId, false).supervisor.keepAlive());
        globalBackend.updateLastActive(grainId, this.userId, session.identityId);
      } catch (err) {                                                  //
        // Ignore disconnects, which imply that the grain shut down already. It'll start back up on
        // the next request, so whatever.                              //
        if (err.kjType !== 'disconnected') {                           // 414
          throw err;                                                   // 415
        }                                                              //
      }                                                                //
                                                                       //
      return true;                                                     // 419
    } else {                                                           //
      return false;                                                    // 421
    }                                                                  //
  },                                                                   //
                                                                       //
  shutdownGrain: function (grainId) {                                  // 425
    check(grainId, String);                                            // 426
    var grain = Grains.findOne(grainId);                               // 427
    if (!grain || !this.userId || grain.userId !== this.userId) {      // 428
      throw new Meteor.Error(403, 'Unauthorized', 'User is not the owner of this grain');
    }                                                                  //
                                                                       //
    waitPromise(globalBackend.shutdownGrain(grainId, grain.userId, true));
  }                                                                    //
});                                                                    //
                                                                       //
var validateWebkey = function (apiToken, refreshedExpiration) {        // 436
  // Validates that `apiToken` is a valid UiView webkey, throwing an exception if it is not. If
  // `refreshedExpiration` is set and if the token has an `expiresIfUnused` field, then the
  // `expiresIfUnused` field is reset to `refreshedExpiration`.        //
                                                                       //
  if (!apiToken) {                                                     // 441
    throw new Meteor.Error(403, 'Invalid authorization token');        // 442
  }                                                                    //
                                                                       //
  if (apiToken.revoked) {                                              // 445
    throw new Meteor.Error(403, 'Authorization token has been revoked');
  }                                                                    //
                                                                       //
  if (apiToken.owner && !('webkey' in apiToken.owner)) {               // 449
    throw new Meteor.Error(403, 'Unauthorized to open non-webkey token.');
  }                                                                    //
                                                                       //
  if (apiToken.expires && apiToken.expires.getTime() <= Date.now()) {  // 453
    throw new Meteor.Error(403, 'Authorization token expired');        // 454
  }                                                                    //
                                                                       //
  if (apiToken.expiresIfUnused) {                                      // 457
    if (apiToken.expiresIfUnused.getTime() <= Date.now()) {            // 458
      throw new Meteor.Error(403, 'Authorization token expired');      // 459
    } else if (refreshedExpiration) {                                  //
      ApiTokens.update(apiToken._id, { $set: { expiresIfUnused: refreshedExpiration } });
    } else {                                                           //
      // It's getting used now, so clear the expiresIfUnused field.    //
      ApiTokens.update(apiToken._id, { $set: { expiresIfUnused: null } });
    }                                                                  //
  }                                                                    //
                                                                       //
  if (apiToken.objectId || apiToken.frontendRef) {                     // 468
    throw new Meteor.Error(403, 'ApiToken refers to a non-webview Capability.');
  }                                                                    //
};                                                                     //
                                                                       //
// Used by shared/grain.js (which sounds like a broken dependency, since this code is only available
// on the server)                                                      //
getGrainSize = function (supervisor, oldSize) {                        // 475
  var promise = undefined;                                             // 476
  if (oldSize === undefined) {                                         // 477
    promise = supervisor.getGrainSize();                               // 478
  } else {                                                             //
    promise = supervisor.getGrainSizeWhenDifferent(oldSize);           // 480
  }                                                                    //
                                                                       //
  var promise2 = promise.then(function (result) {                      // 483
    return parseInt(result.size);                                      // 483
  });                                                                  //
  promise2.cancel = function () {                                      // 484
    promise.cancel();                                                  // 484
  };                                                                   //
                                                                       //
  return promise2;                                                     // 486
};                                                                     //
                                                                       //
Meteor.startup(function () {                                           // 489
  var shutdownApp = function (appId) {                                 // 490
    Grains.find({ appId: appId }).forEach(function (grain) {           // 491
      waitPromise(globalBackend.shutdownGrain(grain._id, grain.userId));
    });                                                                //
  };                                                                   //
                                                                       //
  DevPackages.find().observe({                                         // 496
    removed: function (devPackage) {                                   // 497
      shutdownApp(devPackage.appId);                                   // 497
    },                                                                 //
                                                                       //
    changed: function (oldDevPackage, newDevPackage) {                 // 499
      shutdownApp(oldDevPackage.appId);                                // 500
      if (oldDevPackage.appId !== newDevPackage.appId) {               // 501
        shutdownApp(newDevPackage.appId);                              // 502
      }                                                                //
    },                                                                 //
                                                                       //
    added: function (devPackage) {                                     // 506
      shutdownApp(devPackage.appId);                                   // 506
    }                                                                  //
  });                                                                  //
                                                                       //
  Sessions.find().observe({                                            // 509
    removed: function (session) {                                      // 510
      var proxy = proxiesByHostId[session.hostId];                     // 511
      delete proxiesByHostId[session.hostId];                          // 512
      if (proxy) {                                                     // 513
        proxy.close();                                                 // 514
      }                                                                //
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
// Kill off sessions idle for >~3 minutes.                             //
var TIMEOUT_MS = 180000;                                               // 521
var gcSessions = function () {                                         // 522
  var now = new Date().getTime();                                      // 523
  Sessions.remove({ timestamp: { $lt: now - TIMEOUT_MS } });           // 524
};                                                                     //
                                                                       //
SandstormDb.periodicCleanup(TIMEOUT_MS, gcSessions);                   // 527
                                                                       //
var getProxyForHostId = function (hostId, isAlreadyOpened) {           // 529
  // Get the Proxy corresponding to the given grain session host, possibly (re)creating it if it
  // doesn't already exist. The first request on the session host will always create a new proxy.
  // Later requests may create a proxy if they go to a different front-end replica or if the
  // front-end was restarted.                                          //
  check(hostId, String);                                               // 534
                                                                       //
  return Promise.resolve(undefined).then(function () {                 // 536
    var proxy = proxiesByHostId[hostId];                               // 537
    if (proxy) {                                                       // 538
      return proxy;                                                    // 539
    } else {                                                           //
      // Set table entry to null for now so that we can detect if it is concurrently deleted.
      proxiesByHostId[hostId] = null;                                  // 542
                                                                       //
      return inMeteor(function () {                                    // 544
        var session = Sessions.findOne({ hostId: hostId });            // 545
        if (!session) {                                                // 546
          if (isAlreadyOpened) {                                       // 547
            return new Promise(function (resolve, reject) {            // 548
              var observer = undefined;                                // 549
              var task = Meteor.setTimeout(function () {               // 550
                observer.stop();                                       // 551
                reject(new Meteor.Error(504, 'Requested session that no longer exists, and ' + 'timed out waiting for client to restore it. This can happen if you have ' + 'opened an app\'s content in a new window and then closed it in the ' + 'UI. If you see this error *inside* the Sandstorm UI, please report a ' + 'bug and describe the circumstances of the error.'));
              }, SESSION_PROXY_TIMEOUT);                               //
              observer = Sessions.find({ hostId: hostId }).observe({   // 558
                added: function () {                                   // 559
                  observer.stop();                                     // 560
                  Meteor.clearTimeout(task);                           // 561
                  resolve(getProxyForHostId(hostId, false));           // 562
                }                                                      //
              });                                                      //
            });                                                        //
          } else {                                                     //
            // Does not appear to be a valid session host.             //
            return undefined;                                          // 568
          }                                                            //
        }                                                              //
                                                                       //
        var apiToken = undefined;                                      // 572
        if (session.hashedToken) {                                     // 573
          apiToken = ApiTokens.findOne({ _id: session.hashedToken });  // 574
          // We don't have to fully validate the API token here because if it changed the session
          // would have been deleted.                                  //
          if (!apiToken) {                                             // 577
            throw new Meteor.Error(410, 'ApiToken has been deleted');  // 578
          }                                                            //
        }                                                              //
                                                                       //
        var grain = Grains.findOne(session.grainId);                   // 582
        if (!grain) {                                                  // 583
          // Grain was deleted, I guess.                               //
          throw new Meteor.Error(410, 'Resource has been deleted');    // 585
        }                                                              //
                                                                       //
        // Note that we don't need to call mayOpenGrain() because the existence of a session
        // implies this check was already performed.                   //
                                                                       //
        var proxy = new Proxy(grain._id, grain.userId, session._id, hostId, session.identityId, false);
        if (apiToken) proxy.apiToken = apiToken;                       // 592
                                                                       //
        // Only add the proxy to the table if it was not concurrently deleted (which could happen
        // e.g. if the user's access was revoked).                     //
        if (hostId in proxiesByHostId) {                               // 596
          proxiesByHostId[hostId] = proxy;                             // 597
        } else {                                                       //
          throw new Meteor.Error(403, 'Session was concurrently closed.');
        }                                                              //
                                                                       //
        return proxy;                                                  // 602
      });                                                              //
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
// =======================================================================================
// API tokens                                                          //
                                                                       //
var proxiesByApiToken = {};                                            // 611
                                                                       //
Meteor.startup(function () {                                           // 613
  var clearSessionsAndProxies = function (token) {                     // 614
    // Clears all sessions and API proxies associated with `token` or any token that is downstream
    // in the sharing graph.                                           //
    // TODO(soon): Only clear sessions and proxies for which the permissions have changed.
    var downstream = SandstormPermissions.downstreamTokens(globalDb, { token: token });
    downstream.push(token);                                            // 619
    var identityIds = [];                                              // 620
    var tokenIds = [];                                                 // 621
                                                                       //
    downstream.forEach(function (token) {                              // 623
      var proxy = proxiesByApiToken[token._id];                        // 624
      if (proxy) {                                                     // 625
        proxy.close();                                                 // 626
      }                                                                //
                                                                       //
      delete proxiesByApiToken[token._id];                             // 629
      tokenIds.push(token._id);                                        // 630
      if (token.owner && token.owner.user) {                           // 631
        identityIds.push(token.owner.user.identityId);                 // 632
      }                                                                //
    });                                                                //
                                                                       //
    Sessions.find({                                                    // 636
      grainId: token.grainId,                                          // 637
      $or: [{ identityId: { $in: identityIds } }, { hashedToken: { $in: tokenIds } }]
    }, {                                                               //
      fields: { hostId: 1 }                                            // 642
    }).forEach(function (session) {                                    //
      var proxy = proxiesByHostId[session.hostId];                     // 644
      if (proxy) {                                                     // 645
        proxy.close();                                                 // 646
      }                                                                //
                                                                       //
      delete proxiesByHostId[session.hostId];                          // 649
    });                                                                //
                                                                       //
    Sessions.remove({                                                  // 652
      grainId: token.grainId,                                          // 653
      $or: [{ identityId: { $in: identityIds } }, { hashedToken: { $in: tokenIds } }]
    });                                                                //
  };                                                                   //
                                                                       //
  Grains.find().observe({                                              // 661
    changed: function (newGrain, oldGrain) {                           // 662
      if (oldGrain['private'] != newGrain['private']) {                // 663
        Sessions.remove({ grainId: oldGrain._id, identityId: { $ne: oldGrain.identityId } });
        ApiTokens.find({ grainId: oldGrain._id }).forEach(function (apiToken) {
          delete proxiesByApiToken[apiToken._id];                      // 666
        });                                                            //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  ApiTokens.find({ grainId: { $exists: true }, objectId: { $exists: false } }).observe({
    added: function (newApiToken) {                                    // 673
      // TODO(soon): Unfortunately, added() gets called for all existing role assignments when the
      //   front-end restarts, meaning clearing sessions here will cause people's views to refresh
      //   on server upgrade, which is not a nice user experience. It's also sad to force-refresh
      //   people when they gained new permissions since they might be in the middle of something,
      //   and it's not strictly necessary for security. OTOH, it's sad to be non-reactive. Maybe
      //   we should notify people that they have new permissions and let them click a thing to
      //   refresh?                                                    //
      //clearSessions(roleAssignment.grainId, roleAssignment.recipient);
      //clearApiProxies(roleAssignment.grainId);                       //
    },                                                                 //
                                                                       //
    changed: function (newApiToken, oldApiToken) {                     // 685
      if (!_.isEqual(newApiToken.roleAssignment, oldApiToken.roleAssignment) || !_.isEqual(newApiToken.revoked, oldApiToken.revoked)) {
        clearSessionsAndProxies(newApiToken);                          // 688
      }                                                                //
    },                                                                 //
                                                                       //
    removed: function (oldApiToken) {                                  // 692
      clearSessionsAndProxies(oldApiToken);                            // 693
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
// Used by server/drivers/external-ui/view.js                          //
getProxyForApiToken = function (token) {                               // 699
  check(token, String);                                                // 700
  var hashedToken = Crypto.createHash('sha256').update(token).digest('base64');
  return Promise.resolve(undefined).then(function () {                 // 702
    var proxy = proxiesByApiToken[hashedToken];                        // 703
    if (proxy) {                                                       // 704
      if (proxy.expires && proxy.expires.getTime() <= Date.now()) {    // 705
        throw new Meteor.Error(403, 'Authorization token expired');    // 706
      }                                                                //
                                                                       //
      return proxy;                                                    // 709
    } else {                                                           //
      // Set table entry to null for now so that we can detect if it is concurrently deleted.
      proxiesByApiToken[hashedToken] = null;                           // 712
                                                                       //
      return inMeteor(function () {                                    // 714
        var tokenInfo = ApiTokens.findOne(hashedToken);                // 715
        validateWebkey(tokenInfo);                                     // 716
                                                                       //
        var grain = Grains.findOne(tokenInfo.grainId);                 // 718
        if (!grain) {                                                  // 719
          // Grain was deleted, I guess.                               //
          throw new Meteor.Error(410, 'Resource has been deleted');    // 721
        }                                                              //
                                                                       //
        var proxy = undefined;                                         // 724
        if (tokenInfo.userInfo) {                                      // 725
          throw new Error('API tokens created with arbitrary userInfo no longer supported');
        } else {                                                       //
          var identityId = null;                                       // 728
          if (tokenInfo.identityId && !tokenInfo.forSharing) {         // 729
            identityId = tokenInfo.identityId;                         // 730
          }                                                            //
                                                                       //
          proxy = new Proxy(tokenInfo.grainId, grain.userId, null, null, identityId, true);
          proxy.apiToken = tokenInfo;                                  // 734
        }                                                              //
                                                                       //
        if (!SandstormPermissions.mayOpenGrain(globalDb, { token: tokenInfo })) {
          // Note that only public grains may be opened without a user ID.
          throw new Meteor.Error(403, 'Unauthorized.');                // 739
        }                                                              //
                                                                       //
        if (tokenInfo.expires) {                                       // 742
          proxy.expires = tokenInfo.expires;                           // 743
        }                                                              //
                                                                       //
        // Only add the proxy to the table if it was not concurrently deleted (which could happen
        // e.g. if the token was revoked).                             //
        if (hashedToken in proxiesByApiToken) {                        // 748
          proxiesByApiToken[hashedToken] = proxy;                      // 749
        } else {                                                       //
          throw new Meteor.Error(403, 'Token was concurrently revoked.');
        }                                                              //
                                                                       //
        return proxy;                                                  // 754
      });                                                              //
    }                                                                  //
  });                                                                  //
};                                                                     //
                                                                       //
var apiUseBasicAuth = function (req) {                                 // 760
  // For clients with no convenient way to add an 'Authorization: Bearer' header, we allow the token
  // to be transmitted as a basic auth password.                       //
  var agent = req.headers['user-agent'];                               // 763
  return agent.match(BASIC_AUTH_USER_AGENTS_REGEX);                    // 764
};                                                                     //
                                                                       //
var apiTokenForRequest = function (req) {                              // 767
  var auth = req.headers.authorization;                                // 768
  if (auth && auth.slice(0, 7).toLowerCase() === 'bearer ') {          // 769
    return auth.slice(7).trim();                                       // 770
  } else if (auth && auth.slice(0, 6).toLowerCase() === 'basic ' && apiUseBasicAuth(req)) {
    return new Buffer(auth.slice(6).trim(), 'base64').toString().split(':')[1];
  } else {                                                             //
    return undefined;                                                  // 774
  }                                                                    //
};                                                                     //
                                                                       //
// =======================================================================================
// Routing to proxies.                                                 //
//                                                                     //
                                                                       //
// pre-meteor.js calls this                                            //
tryProxyUpgrade = function (hostId, req, socket, head) {               // 783
  // Attempt to handle a WebSocket upgrade by dispatching it to a grain. Returns a promise that
  // resolves true if an appropriate grain is found, false if there was no match (but the caller
  // should consider other host types, like static web publishing), or throws an error if the
  // request is definitely invalid.                                    //
                                                                       //
  if (hostId === 'api') {                                              // 789
    var token = apiTokenForRequest(req);                               // 790
    if (token) {                                                       // 791
      return getProxyForApiToken(token).then(function (proxy) {        // 792
        // Meteor sets the timeout to five seconds. Change that back to two
        // minutes, which is the default value.                        //
        socket.setTimeout(120000);                                     // 795
                                                                       //
        proxy.upgradeHandler(req, socket, head);                       // 797
        return true;                                                   // 798
      });                                                              //
    } else {                                                           //
      return Promise.resolve(false);                                   // 801
    }                                                                  //
  } else {                                                             //
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf('sandstorm-sid=') !== -1;
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {
      if (proxy) {                                                     // 806
        // Cross-origin requests are not allowed on UI session hosts.  //
        var origin = req.headers.origin;                               // 808
        if (origin !== PROTOCOL + '//' + req.headers.host) {           // 809
          throw new Meteor.Error(403, 'Detected illegal cross-origin WebSocket from: ' + origin);
        }                                                              //
                                                                       //
        // Meteor sets the timeout to five seconds. Change that back to two
        // minutes, which is the default value.                        //
        socket.setTimeout(120000);                                     // 815
                                                                       //
        proxy.upgradeHandler(req, socket, head);                       // 817
        return true;                                                   // 818
      } else {                                                         //
        return false;                                                  // 820
      }                                                                //
    });                                                                //
  }                                                                    //
};                                                                     //
                                                                       //
// pre-meteor.js calls this                                            //
tryProxyRequest = function (hostId, req, res) {                        // 827
  // Attempt to handle an HTTP request by dispatching it to a grain. Returns a promise that
  // resolves true if an appropriate grain is found, false if there was no match (but the caller
  // should consider other host types, like static web publishing), or throws an error if the
  // request is definitely invalid.                                    //
                                                                       //
  if (hostId === 'api') {                                              // 833
    var _ret = (function () {                                          //
      // This is a request for the API host.                           //
                                                                       //
      res.setHeader('Access-Control-Allow-Origin', '*');               // 836
                                                                       //
      if (req.method === 'OPTIONS') {                                  // 838
        // Reply to CORS preflight request.                            //
                                                                       //
        // All we want to do is permit APIs to be accessed from arbitrary origins. Since clients
        // must send a valid Authorization header, and since cookies are not used for
        // authorization, this is perfectly safe. In a sane world, we would only need to send back
        // 'Access-Control-Allow-Origin: *' and be done with it.       //
        //                                                             //
        // However, CORS demands that we explicitly whitelist individual methods and headers for
        // use cross-origin, as if this is somehow useful for implementing any practical security
        // policy (it isn't). To make matters worse, we are REQUIRED to enumerate each one
        // individually. We cannot just write '*' for these lists. WTF, CORS?
        //                                                             //
        // Luckily, the request tells us exactly what method and headers are being requested, so we
        // only need to copy those over, rather than create an exhaustive list. But this is still
        // overly complicated.                                         //
                                                                       //
        var accessControlHeaders = {                                   // 855
          'Access-Control-Allow-Methods': 'GET, HEAD, POST, PUT, PATCH, DELETE',
          'Access-Control-Max-Age': '3600'                             // 857
        };                                                             //
                                                                       //
        // Copy all requested headers to the allowed headers list.     //
        var requestedHeaders = req.headers['access-control-request-headers'];
        if (requestedHeaders) {                                        // 862
          accessControlHeaders['Access-Control-Allow-Headers'] = requestedHeaders;
        }                                                              //
                                                                       //
        // Add the requested method to the allowed methods list, if it's not there already.
        var requestedMethod = req.headers['access-control-request-method'];
        if (requestedMethod && !_.contains(['GET', 'HEAD', 'POST', 'PUT', 'PATCH', 'DELETE'], requestedMethod)) {
          accessControlHeaders['Access-Control-Allow-Methods'] += ', ' + requestedMethod;
        }                                                              //
                                                                       //
        for (header in babelHelpers.sanitizeForInObject(accessControlHeaders)) {
          res.setHeader(header, accessControlHeaders[header]);         // 874
        }                                                              //
      }                                                                //
                                                                       //
      var errorHandler = function (err) {                              // 878
        if (err instanceof Meteor.Error) {                             // 879
          console.log('error: ' + err);                                // 880
          res.writeHead(err.error, err.reason, { 'Content-Type': 'text/plain' });
        } else {                                                       //
          res.writeHead(500, 'Internal Server Error', { 'Content-Type': 'text/plain' });
        }                                                              //
                                                                       //
        res.end(err.stack);                                            // 886
      };                                                               //
                                                                       //
      var token = apiTokenForRequest(req);                             // 889
      if (token && req.headers['x-sandstorm-token-keepalive']) {       // 890
        inMeteor(function () {                                         // 891
          var keepaliveDuration = parseInt(req.headers['x-sandstorm-token-keepalive']);
          check(keepaliveDuration, Match.Integer);                     // 893
          var hashedToken = Crypto.createHash('sha256').update(token).digest('base64');
          validateWebkey(ApiTokens.findOne(hashedToken), new Date(Date.now() + keepaliveDuration));
        }).then(function () {                                          //
          res.writeHead(200, {});                                      // 897
          res.end();                                                   // 898
        }, errorHandler);                                              //
      } else if (token) {                                              //
        getProxyForApiToken(token).then(function (proxy) {             // 901
          proxy.requestHandler(req, res);                              // 902
        }, errorHandler);                                              //
      } else if (req.method === 'OPTIONS') {                           //
        // OPTIONS request with no authorization token. Fine to end here.
        res.writeHead(200, {});                                        // 906
        res.end();                                                     // 907
      } else {                                                         //
        if (apiUseBasicAuth(req)) {                                    // 909
          res.writeHead(401, {                                         // 910
            'Content-Type': 'text/plain',                              // 911
            'WWW-Authenticate': 'Basic realm=\'Sandstorm API\''        // 912
          });                                                          //
        } else {                                                       //
          // TODO(someday): Display some sort of nifty API browser.    //
          res.writeHead(403, { 'Content-Type': 'text/plain' });        // 916
        }                                                              //
                                                                       //
        res.end('Missing or invalid authorization header.\n\n' + 'This address serves APIs, which allow external apps (such as a phone app) to\n' + 'access data on your Sandstorm server. This address is not meant to be opened\n' + 'in a regular browser.');
      }                                                                //
                                                                       //
      return {                                                         // 925
        v: Promise.resolve(true)                                       //
      };                                                               //
    })();                                                              //
                                                                       //
    if (typeof _ret === 'object') return _ret.v;                       //
  } else {                                                             //
    var isAlreadyOpened = req.headers.cookie && req.headers.cookie.indexOf('sandstorm-sid=') !== -1;
    return getProxyForHostId(hostId, isAlreadyOpened).then(function (proxy) {
      if (proxy) {                                                     // 929
        proxy.requestHandler(req, res);                                // 930
        return true;                                                   // 931
      } else {                                                         //
        return false;                                                  // 933
      }                                                                //
    });                                                                //
  }                                                                    //
};                                                                     //
                                                                       //
// =======================================================================================
// Proxy class                                                         //
//                                                                     //
// Connects to a grain and exports it on a wildcard host.              //
//                                                                     //
                                                                       //
Proxy = (function () {                                                 // 945
  function Proxy(grainId, ownerId, sessionId, hostId, identityId, isApi, supervisor) {
    var _this2 = this;                                                 //
                                                                       //
    babelHelpers.classCallCheck(this, Proxy);                          //
                                                                       //
    this.grainId = grainId;                                            // 947
    this.ownerId = ownerId;                                            // 948
    this.identityId = identityId;                                      // 949
    this.supervisor = supervisor; // note: optional parameter; we can reconnect
    this.sessionId = sessionId;                                        // 951
    this.isApi = isApi;                                                // 952
    this.hasLoaded = false;                                            // 953
    this.websockets = [];                                              // 954
    if (sessionId) {                                                   // 955
      if (!hostId) throw new Error('sessionId must come with hostId');
      if (isApi) throw new Error('API proxy shouldn\'t have sessionId');
      this.hostId = hostId;                                            // 958
    } else {                                                           //
      if (!isApi) throw new Error('non-API proxy requires sessionId');
      if (hostId) throw new Error('API proxy sholudn\'t have hostId');
    }                                                                  //
                                                                       //
    if (this.identityId) {                                             // 964
      var identity = globalDb.getIdentity(this.identityId);            // 965
      if (!identity) {                                                 // 966
        throw new Error('identity not found: ' + this.identityId);     // 967
      }                                                                //
                                                                       //
      this.userInfo = {                                                // 970
        displayName: { defaultText: identity.profile.name },           // 971
        preferredHandle: identity.profile.handle,                      // 972
        identityId: new Buffer(identity._id, 'hex')                    // 973
      };                                                               //
      if (identity.profile.pictureUrl) this.userInfo.pictureUrl = identity.profile.pictureUrl;
      if (identity.profile.pronoun) this.userInfo.pronouns = identity.profile.pronoun;
    } else {                                                           //
      this.userInfo = {                                                // 978
        displayName: { defaultText: 'Anonymous User' },                // 979
        preferredHandle: 'anonymous'                                   // 980
      };                                                               //
    }                                                                  //
                                                                       //
    var _this = this;                                                  // 984
                                                                       //
    this.requestHandler = function (request, response) {               // 986
      if (_this2.sessionId) {                                          // 987
        // Implement /_sandstorm-init for setting the session cookie.  //
        var url = Url.parse(request.url, true);                        // 989
        if (url.pathname === '/_sandstorm-init' && url.query.sessionid === _this.sessionId) {
          _this.doSessionInit(request, response, url.query.path);      // 991
          return;                                                      // 992
        }                                                              //
      }                                                                //
                                                                       //
      Promise.resolve(undefined).then(function () {                    // 996
        var contentLength = request.headers['content-length'];         // 997
        if ((request.method === 'POST' || request.method === 'PUT') && (contentLength === undefined || contentLength > 1024 * 1024)) {
          // The input is either very long, or we don't know how long it is, so use streaming mode.
          return _this.handleRequestStreaming(request, response, contentLength, 0);
        } else {                                                       //
          return readAll(request).then(function (data) {               // 1003
            return _this.handleRequest(request, data, response, 0);    // 1004
          });                                                          //
        }                                                              //
      })['catch'](function (err) {                                     //
        _this.setHasLoaded();                                          // 1008
                                                                       //
        var body = err.stack;                                          // 1010
        if (err.cppFile) {                                             // 1011
          body += '\nC++ location:' + err.cppFile + ':' + (err.line || '??');
        }                                                              //
                                                                       //
        if (err.kjType) {                                              // 1015
          body += '\ntype: ' + err.kjType;                             // 1016
        }                                                              //
                                                                       //
        if (response.headersSent) {                                    // 1019
          // Unfortunately, it's too late to tell the client what happened.
          console.error('HTTP request failed after response already sent:', body);
          response.end();                                              // 1022
        } else {                                                       //
          if (err instanceof Meteor.Error) {                           // 1024
            response.writeHead(err.error, err.reason, { 'Content-Type': 'text/plain' });
          } else {                                                     //
            response.writeHead(500, 'Internal Server Error', { 'Content-Type': 'text/plain' });
          }                                                            //
                                                                       //
          response.end(body);                                          // 1030
        }                                                              //
      });                                                              //
    };                                                                 //
                                                                       //
    this.upgradeHandler = function (request, socket, head) {           // 1035
      _this.handleWebSocket(request, socket, head, 0)['catch'](function (err) {
        console.error('WebSocket setup failed:', err.stack);           // 1037
        // TODO(cleanup):  Manually send back a 500 response?          //
        socket.destroy();                                              // 1039
      });                                                              //
    };                                                                 //
  }                                                                    //
                                                                       //
  Proxy.prototype.close = (function () {                               // 945
    function close() {                                                 // 1044
      this.websockets.forEach(function (socket) {                      // 1045
        socket.destroy();                                              // 1046
      });                                                              //
                                                                       //
      this.websockets = [];                                            // 1049
                                                                       //
      if (this.session) {                                              // 1051
        this.session.close();                                          // 1052
        delete this.session;                                           // 1053
      }                                                                //
                                                                       //
      if (this.uiView) {                                               // 1056
        this.uiView.close();                                           // 1057
        delete this.uiView;                                            // 1058
      }                                                                //
                                                                       //
      if (this.supervisor) {                                           // 1061
        this.supervisor.close();                                       // 1062
        delete this.supervisor;                                        // 1063
      }                                                                //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  Proxy.prototype.getConnection = (function () {                       // 945
    function getConnection() {                                         // 1067
      if (!this.supervisor) {                                          // 1068
        this.supervisor = globalBackend.cap().getGrain(this.ownerId, this.grainId).supervisor;
        this.uiView = null;                                            // 1070
      }                                                                //
                                                                       //
      if (!this.uiView) {                                              // 1073
        this.uiView = this.supervisor.getMainView().view;              // 1074
      }                                                                //
    }                                                                  //
                                                                       //
    return getConnection;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewWebSession = (function () {                  // 945
    function _callNewWebSession(request, userInfo) {                   // 1078
      var params = Capnp.serialize(WebSession.Params, {                // 1079
        basePath: PROTOCOL + '//' + request.headers.host,              // 1080
        userAgent: 'user-agent' in request.headers ? request.headers['user-agent'] : 'UnknownAgent/0.0',
        acceptableLanguages: 'accept-language' in request.headers ? request.headers['accept-language'].split(',').map(function (s) {
          return s.trim();                                             // 1085
        }) : ['en-US', 'en']                                           //
      });                                                              //
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId), WebSession.typeId, params).session;
    }                                                                  //
                                                                       //
    return _callNewWebSession;                                         //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewApiSession = (function () {                  // 945
    function _callNewApiSession(request, userInfo) {                   // 1093
      var _this = this;                                                // 1094
      var params = {};                                                 // 1095
                                                                       //
      if ('x-sandstorm-passthrough' in request.headers) {              // 1097
        var optIns = request.headers['x-sandstorm-passthrough'].split(',').map(function (s) {
          return s.trim();                                             // 1100
        });                                                            //
        // The only currently supported passthrough value is 'address', but others could be useful in
        // the future                                                  //
        if (optIns.indexOf('address') !== -1) {                        // 1103
          // Sadly, we can't use request.socket.remoteFamily because it's not available in the (rather-old)
          // version of node that comes in the Meteor bundle we're using.  Hence this hackery.
          var addressToPass = request.socket.remoteAddress;            // 1106
          if (isRfc1918OrLocal(addressToPass) && 'x-real-ip' in request.headers) {
            // Allow overriding the socket's remote address with X-Real-IP header if the request comes
            // from either localhost or an RFC1918 address.  These are not useful for geolocation
            // anyway.                                                 //
            addressToPass = request.headers['x-real-ip'];              // 1111
          }                                                            //
                                                                       //
          if (Net.isIPv4(addressToPass)) {                             // 1114
            // Map IPv4 addresses in IPv6.                             //
            // This conveniently comes out to a 48-bit number, which is precisely representable in a
            // double (which has 53 mantissa bits). Thus we can avoid using Bignum/strings, which we
            // might otherwise need to precisely represent 64-bit fields.
            var v4Int = 0xFFFF00000000 + addressToPass.split('.').map(function (x) {
              return parseInt(x, 10);                                  // 1120
            }).reduce(function (a, b) {                                //
              return 256 * a + b;                                      // 1121
            });                                                        //
            params.remoteAddress = {                                   // 1122
              lower64: v4Int,                                          // 1123
              upper64: 0                                               // 1124
            };                                                         //
          } else if (Net.isIPv6(addressToPass)) {                      //
            // TODO(test): Unit test this                              //
            // Parse a valid v6 address.                               //
            // Split into groups, then insert an appropriate number of 0's if :: was used.
            var groups = addressToPass.split(':');                     // 1130
                                                                       //
            // Strip extra empty group in the case of a leading or trailing '::'.
            if (groups[0] === '') {                                    // 1133
              groups.shift();                                          // 1134
            }                                                          //
                                                                       //
            if (groups[groups.length - 1] === '') {                    // 1137
              groups.pop();                                            // 1138
            }                                                          //
                                                                       //
            var lastGroup = groups[groups.length - 1];                 // 1141
            // Handle IPv4-mapped IPv6 addresses.  These end in a dotted-quad IPv4 address, which we
            // should expand into two groups of 4-character hex strings, like the rest of the address.
            if (Net.isIPv4(lastGroup)) {                               // 1144
              groups.pop();                                            // 1145
              var _quad = lastGroup.split('.').map(function (x) {      // 1146
                return parseInt(x, 10);                                // 1146
              });                                                      //
              groups.push((_quad[0] * 256 + _quad[1]).toString(16));   // 1147
              groups.push((_quad[2] * 256 + _quad[3]).toString(16));   // 1148
            }                                                          //
                                                                       //
            var groupsToAdd = 8 - groups.length;                       // 1151
            var emptyGroupIndex = groups.indexOf('');                  // 1152
            var cleanGroups = undefined;                               // 1153
            if (emptyGroupIndex !== -1) {                              // 1154
              var head = groups.slice(0, emptyGroupIndex);             // 1155
              // groupsToAdd + 1 because we sliced out the empty element
              var mid = Array(groupsToAdd + 1);                        // 1157
              for (var i = 0; i < groupsToAdd + 1; i++) {              // 1158
                mid[i] = '0';                                          // 1159
              }                                                        //
                                                                       //
              var tail = groups.slice(emptyGroupIndex + 1, groups.length);
              cleanGroups = [].concat(head, mid, tail);                // 1163
            } else {                                                   //
              cleanGroups = groups;                                    // 1165
            }                                                          //
                                                                       //
            var ints = cleanGroups.map(function (x) {                  // 1168
              return parseInt(x, 16);                                  // 1168
            });                                                        //
            // We use strings because we'd lose data from loss of precision casting the 64-bit uints
            // into 53-bit-mantissa doubles.                           //
            params.remoteAddress = {                                   // 1171
              upper64: quadToIntString(ints.slice(0, 4)),              // 1172
              lower64: quadToIntString(ints.slice(4, 8))               // 1173
            };                                                         //
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      var serializedParams = Capnp.serialize(ApiSession.Params, params);
                                                                       //
      // TODO(someday): We are currently falling back to WebSession if we get any kind of error upon
      // calling newSession with an ApiSession._id.                    //
      // Eventually we'll remove this logic once we're sure apps have updated.
      return this.uiView.newSession(userInfo, makeHackSessionContext(this.grainId, this.sessionId, this.identityId), ApiSession.typeId, serializedParams).then(function (session) {
        return session.session;                                        // 1188
      }, function (err) {                                              //
        return _this._callNewWebSession(request, userInfo);            // 1190
      });                                                              //
    }                                                                  //
                                                                       //
    return _callNewApiSession;                                         //
  })();                                                                //
                                                                       //
  Proxy.prototype._callNewSession = (function () {                     // 945
    function _callNewSession(request, viewInfo) {                      // 1195
      var userInfo = _.clone(this.userInfo);                           // 1196
      var _this = this;                                                // 1197
      var promise = inMeteor(function () {                             // 1198
        var vertex = undefined;                                        // 1199
        if (_this.apiToken) {                                          // 1200
          vertex = { token: _this.apiToken };                          // 1201
        } else {                                                       //
          // (_this.identityId might be null; this is fine)            //
          vertex = { grain: { _id: _this.grainId, identityId: _this.identityId } };
        }                                                              //
                                                                       //
        var permissions = SandstormPermissions.grainPermissions(globalDb, vertex, viewInfo);
        if (!permissions) {                                            // 1208
          throw new Meteor.Error(403, 'Unauthorized', 'User is not authorized to open this grain.');
        }                                                              //
                                                                       //
        Sessions.update({                                              // 1212
          _id: _this.sessionId                                         // 1213
        }, {                                                           //
          $set: {                                                      // 1215
            viewInfo: viewInfo,                                        // 1216
            permissions: permissions                                   // 1217
          }                                                            //
        });                                                            //
                                                                       //
        return permissions;                                            // 1221
      });                                                              //
                                                                       //
      return promise.then(function (permissions) {                     // 1224
        userInfo.permissions = permissions;                            // 1225
                                                                       //
        var numBytes = Math.ceil(permissions.length / 8);              // 1227
        var buf = new Buffer(numBytes);                                // 1228
        for (var ii = 0; ii < numBytes; ++ii) {                        // 1229
          buf[ii] = 0;                                                 // 1230
        }                                                              //
                                                                       //
        for (var ii = 0; ii < permissions.length; ++ii) {              // 1233
          var byteNum = Math.floor(ii / 8);                            // 1234
          var bitNum = ii % 8;                                         // 1235
          if (permissions[ii]) {                                       // 1236
            buf[byteNum] = buf[byteNum] | 1 << bitNum;                 // 1237
          }                                                            //
        }                                                              //
                                                                       //
        userInfo.deprecatedPermissionsBlob = buf;                      // 1241
                                                                       //
        if (_this.isApi) {                                             // 1243
          return _this._callNewApiSession(request, userInfo);          // 1244
        } else {                                                       //
          return _this._callNewWebSession(request, userInfo);          // 1246
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return _callNewSession;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.getSession = (function () {                          // 945
    function getSession(request) {                                     // 1251
      var _this3 = this;                                               //
                                                                       //
      if (!this.session) {                                             // 1252
        (function () {                                                 //
          _this3.getConnection(); // make sure we're connected         // 1253
          var _this = _this3;                                          // 1254
          var promise = _this3.uiView.getViewInfo().then(function (viewInfo) {
            return inMeteor(function () {                              // 1256
              Grains.update(_this.grainId, { $set: { cachedViewInfo: viewInfo } });
            }).then(function () {                                      //
              return _this._callNewSession(request, viewInfo);         // 1259
            });                                                        //
          }, function (error) {                                        //
            if (error.kjType === 'failed' || error.kjType === 'unimplemented') {
              // Method not implemented.                               //
              // TODO(apibump): Don't treat 'failed' as 'unimplemented'. Unfortunately, old apps built
              //   with old versions of Cap'n Proto don't throw 'unimplemented' exceptions, so we have
              //   to accept 'failed' here at least until the next API bump.
              return _this._callNewSession(request, {});               // 1267
            } else {                                                   //
              return Promise.reject(error);                            // 1269
            }                                                          //
          });                                                          //
          _this3.session = new Capnp.Capability(promise, WebSession);  // 1272
        })();                                                          //
      }                                                                //
                                                                       //
      return this.session;                                             // 1275
    }                                                                  //
                                                                       //
    return getSession;                                                 //
  })();                                                                //
                                                                       //
  Proxy.prototype.keepAlive = (function () {                           // 945
    function keepAlive() {                                             // 1278
      this.getConnection();                                            // 1279
      return this.supervisor.keepAlive();                              // 1280
    }                                                                  //
                                                                       //
    return keepAlive;                                                  //
  })();                                                                //
                                                                       //
  Proxy.prototype.resetConnection = (function () {                     // 945
    function resetConnection() {                                       // 1283
      if (this.session) {                                              // 1284
        this.session.close();                                          // 1285
        delete this.session;                                           // 1286
      }                                                                //
                                                                       //
      if (this.uiView) {                                               // 1289
        this.uiView.close();                                           // 1290
        delete this.uiView;                                            // 1291
      }                                                                //
                                                                       //
      if (this.supervisor) {                                           // 1294
        this.supervisor.close();                                       // 1295
        delete globalBackend.runningGrains[this.grainId];              // 1296
        delete this.supervisor;                                        // 1297
      }                                                                //
    }                                                                  //
                                                                       //
    return resetConnection;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.maybeRetryAfterError = (function () {                // 945
    function maybeRetryAfterError(error, retryCount) {                 // 1301
      // If the error may be caused by the grain dying or a network failure, try to restart it,
      // returning a promise that resolves once restarted. Otherwise, just rethrow the error.
      // `retryCount` should be incremented for every successful retry as part of the same request;
      // we only want to retry once.                                   //
      var _this = this;                                                // 1306
      if (SandstormBackend.shouldRestartGrain(error, retryCount)) {    // 1307
        this.resetConnection();                                        // 1308
        return inMeteor(function () {                                  // 1309
          _this.supervisor = globalBackend.continueGrain(_this.grainId).supervisor;
        });                                                            //
      } else {                                                         //
        throw error;                                                   // 1313
      }                                                                //
    }                                                                  //
                                                                       //
    return maybeRetryAfterError;                                       //
  })();                                                                //
                                                                       //
  Proxy.prototype.doSessionInit = (function () {                       // 945
    function doSessionInit(request, response, requestPath) {           // 1317
      // jscs:disable disallowQuotedKeysInObjects                      //
      var path = requestPath || '/';                                   // 1319
                                                                       //
      // Check that the path is relative (ie. starts with a /).        //
      // Also ensure that it doesn't start with 2 /, because that is interpreted as non-relative
      if (path.lastIndexOf('/', 0) !== 0 || path.lastIndexOf('//', 0) === 0) {
        response.writeHead(400, 'Invalid path supplied', { 'Content-Type': 'text/plain' });
        response.end('Invalid path supplied.');                        // 1325
        return;                                                        // 1326
      }                                                                //
                                                                       //
      // Set the session ID.                                           //
      response.setHeader('Set-Cookie', ['sandstorm-sid=', this.sessionId, '; Max-Age=31536000; HttpOnly'].join(''));
                                                                       //
      response.setHeader('Cache-Control', 'no-cache, private');        // 1332
                                                                       //
      // Redirect to the app's root URL.                               //
      // Note:  All browsers support relative locations and the next update to HTTP/1.1 will officially
      //   make them valid.  http://tools.ietf.org/html/draft-ietf-httpbis-p2-semantics-26#page-67
      response.writeHead(303, 'See Other', { 'Location': encodeURI(path) });
      response.end();                                                  // 1338
    }                                                                  //
                                                                       //
    return doSessionInit;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype.makeContext = (function () {                         // 945
    function makeContext(request, response) {                          // 1341
      // Parses the cookies from the request, checks that the session ID is present and valid, then
      // returns the request context which contains the other cookies.  Throws an exception if the
      // session ID is missing or invalid.                             //
                                                                       //
      var context = {};                                                // 1346
                                                                       //
      if (this.hostId) {                                               // 1348
        var parseResult = parseCookies(request);                       // 1349
        if (!parseResult.sessionId || parseResult.sessionId !== this.sessionId) {
          throw new Meteor.Error(403, 'Unauthorized');                 // 1351
        }                                                              //
                                                                       //
        if (parseResult.cookies.length > 0) {                          // 1354
          context.cookies = parseResult.cookies;                       // 1355
        }                                                              //
      } else {                                                         //
        // jscs:disable disallowEmptyBlocks                            //
        // This is an API request. Cookies are not supported.          //
      }                                                                //
                                                                       //
      context.accept = parseAcceptHeader(request);                     // 1362
                                                                       //
      context.eTagPrecondition = parsePreconditionHeader(request);     // 1364
                                                                       //
      context.additionalHeaders = [];                                  // 1366
      WebSession.Context.headerWhitelist.forEach(function (headerName) {
        if (request.headers[headerName]) {                             // 1368
          context.additionalHeaders.push({                             // 1369
            name: headerName,                                          // 1370
            value: request.headers[headerName]                         // 1371
          });                                                          //
        };                                                             //
      });                                                              //
                                                                       //
      var promise = new Promise(function (resolve, reject) {           // 1376
        response.resolveResponseStream = resolve;                      // 1377
        response.rejectResponseStream = reject;                        // 1378
      });                                                              //
                                                                       //
      context.responseStream = new Capnp.Capability(promise, ByteStream);
                                                                       //
      return context;                                                  // 1383
    }                                                                  //
                                                                       //
    return makeContext;                                                //
  })();                                                                //
                                                                       //
  Proxy.prototype.translateResponse = (function () {                   // 945
    function translateResponse(rpcResponse, response, request) {       // 1386
      if (this.hostId) {                                               // 1387
        if (rpcResponse.setCookies && rpcResponse.setCookies.length > 0) {
          response.setHeader('Set-Cookie', rpcResponse.setCookies.map(makeSetCookieHeader));
        }                                                              //
                                                                       //
        // TODO(security): Add a Content-Security-Policy header which:
        // (1) Prevents the app from initiating HTTP requests to third parties.
        // (2) Prevents the app from navigating the parent frame.      //
        // (3) Prevents the app from opening popups.                   //
        // (4) Prohibits anyone other than the Sandstorm shell from framing the app (as a backup
        //   defense vs. clickjacking, though unguessable hostnames already mostly prevent this).
      } else {                                                         //
          // jscs:disable validateQuoteMarks                           //
          // This is an API request. Cookies are not supported.        //
                                                                       //
          // We need to make sure caches know that different bearer tokens get totally different results.
          response.setHeader('Vary', 'Authorization');                 // 1403
                                                                       //
          // APIs can be called from any origin. Because we ignore cookies, there is no security problem.
          response.setHeader('Access-Control-Allow-Origin', '*');      // 1406
                                                                       //
          // Add a Content-Security-Policy as a backup in case someone finds a way to load this resource
          // in a browser context. This policy should thoroughly neuter it.
          response.setHeader('Content-Security-Policy', "default-src 'none'; sandbox");
        }                                                              //
                                                                       //
      // On first response, update the session to have hasLoaded=true  //
      this.setHasLoaded();                                             // 1414
                                                                       //
      // TODO(security): Set X-Content-Type-Options: nosniff?          //
                                                                       //
      if ('content' in rpcResponse) {                                  // 1418
        var content = rpcResponse.content;                             // 1419
        var code = successCodes[content.statusCode];                   // 1420
        if (!code) {                                                   // 1421
          throw new Error('Unknown status code: ', content.statusCode);
        }                                                              //
                                                                       //
        if (content.mimeType) {                                        // 1425
          response.setHeader('Content-Type', content.mimeType);        // 1426
        }                                                              //
                                                                       //
        if (content.encoding) {                                        // 1429
          response.setHeader('Content-Encoding', content.encoding);    // 1430
        }                                                              //
                                                                       //
        if (content.language) {                                        // 1433
          response.setHeader('Content-Language', content.language);    // 1434
        }                                                              //
                                                                       //
        if (content.eTag) {                                            // 1437
          response.setHeader('ETag', composeETag(content.eTag));       // 1438
        }                                                              //
                                                                       //
        if ('disposition' in content && 'download' in content.disposition) {
          response.setHeader('Content-Disposition', 'attachment; filename="' + content.disposition.download.replace(/([\\"\n])/g, "\\$1") + '"');
        }                                                              //
                                                                       //
        if ('stream' in content.body) {                                // 1446
          if (request.method === 'HEAD') {                             // 1447
            content.body.stream.close();                               // 1448
            response.rejectResponseStream(new Error('HEAD request; content doesn\'t matter.'));
          } else {                                                     //
            var _ret3 = (function () {                                 //
              var streamHandle = content.body.stream;                  // 1451
              response.writeHead(code.id, code.title);                 // 1452
              var promise = new Promise(function (resolve, reject) {   // 1453
                response.resolveResponseStream(new Capnp.Capability(new ResponseStream(response, streamHandle, resolve, reject), ByteStream));
              });                                                      //
              promise.streamHandle = streamHandle;                     // 1457
              return {                                                 // 1458
                v: promise                                             //
              };                                                       //
            })();                                                      //
                                                                       //
            if (typeof _ret3 === 'object') return _ret3.v;             //
          }                                                            //
        } else {                                                       //
          response.rejectResponseStream(new Error('Response content body was not a stream.'));
                                                                       //
          if ('bytes' in content.body) {                               // 1464
            response.setHeader('Content-Length', content.body.bytes.length);
          } else {                                                     //
            throw new Error('Unknown content body type.');             // 1467
          }                                                            //
        }                                                              //
                                                                       //
        response.writeHead(code.id, code.title);                       // 1471
                                                                       //
        if ('bytes' in content.body && request.method !== 'HEAD') {    // 1473
          response.write(content.body.bytes);                          // 1474
        }                                                              //
                                                                       //
        response.end();                                                // 1477
      } else if ('noContent' in rpcResponse) {                         //
        var noContent = rpcResponse.noContent;                         // 1479
        var noContentCode = noContentSuccessCodes[noContent.shouldResetForm * 1];
        response.writeHead(noContentCode.id, noContentCode.title);     // 1481
        response.end();                                                // 1482
      } else if ('preconditionFailed' in rpcResponse) {                //
        var preconditionFailed = rpcResponse.preconditionFailed;       // 1484
        if (request.method === 'GET' && 'if-none-match' in request.headers) {
          if (preconditionFailed.matchingETag) {                       // 1486
            response.setHeader('ETag', composeETag(preconditionFailed.matchingETag));
          }                                                            //
                                                                       //
          response.writeHead(304, 'Not Modified');                     // 1490
        } else {                                                       //
          response.writeHead(412, 'Precondition Failed');              // 1492
        }                                                              //
                                                                       //
        response.end();                                                // 1495
      } else if ('redirect' in rpcResponse) {                          //
        var redirect = rpcResponse.redirect;                           // 1497
        var redirectCode = redirectCodes[redirect.switchToGet * 2 + redirect.isPermanent];
        response.writeHead(redirectCode.id, redirectCode.title, {      // 1499
          'Location': redirect.location                                // 1500
        });                                                            //
        response.end();                                                // 1502
      } else if ('clientError' in rpcResponse) {                       //
        var clientError = rpcResponse.clientError;                     // 1504
        var errorCode = errorCodes[clientError.statusCode];            // 1505
        if (!errorCode) {                                              // 1506
          throw new Error('Unknown status code: ', clientError.statusCode);
        }                                                              //
                                                                       //
        response.writeHead(errorCode.id, errorCode.title, {            // 1510
          'Content-Type': 'text/html'                                  // 1511
        });                                                            //
                                                                       //
        if (request.method !== 'HEAD') {                               // 1514
          if (clientError.descriptionHtml) {                           // 1515
            response.write(clientError.descriptionHtml);               // 1516
          } else {                                                     //
            // TODO(someday):  Better default error page.              //
            response.write('<html><body><h1>' + errorCode.id + ': ' + errorCode.title + '</h1></body></html>');
          }                                                            //
        }                                                              //
                                                                       //
        response.end();                                                // 1524
      } else if ('serverError' in rpcResponse) {                       //
        response.writeHead(500, 'Internal Server Error', {             // 1526
          'Content-Type': 'text/html'                                  // 1527
        });                                                            //
                                                                       //
        if (request.method !== 'HEAD') {                               // 1530
          if (rpcResponse.serverError.descriptionHtml) {               // 1531
            response.write(rpcResponse.serverError.descriptionHtml);   // 1532
          } else {                                                     //
            // TODO(someday):  Better default error page.              //
            response.write('<html><body><h1>500: Internal Server Error</h1></body></html>');
          }                                                            //
        }                                                              //
                                                                       //
        response.end();                                                // 1539
      } else {                                                         //
        throw new Error('Unknown HTTP response type:\n' + JSON.stringify(rpcResponse));
      }                                                                //
                                                                       //
      return Promise.resolve(undefined);                               // 1544
    }                                                                  //
                                                                       //
    return translateResponse;                                          //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleRequest = (function () {                       // 945
    function handleRequest(request, data, response, retryCount) {      // 1547
      var _this = this;                                                // 1548
      return Promise.resolve(undefined).then(function () {             // 1549
        return _this.makeContext(request, response);                   // 1550
      }).then(function (context) {                                     //
        // jscs:disable requireDotNotation                             //
        // Send the RPC.                                               //
        var path = request.url.slice(1); // remove leading '/'         // 1554
        var session = _this.getSession(request);                       // 1555
                                                                       //
        var requestContent = function () {                             // 1557
          return {                                                     // 1558
            content: data,                                             // 1559
            encoding: request.headers['content-encoding'],             // 1560
            mimeType: request.headers['content-type']                  // 1561
          };                                                           //
        };                                                             //
                                                                       //
        var xmlContent = function () {                                 // 1565
          var type = request.headers['content-type'] || 'application/xml;charset=utf-8';
          var match = type.match(/[^/]*\/xml(; *charset *= *([^ ;]*))?/);
          if (!match) {                                                // 1568
            response.writeHead(415, 'Unsupported media type.', {       // 1569
              'Content-Type': 'text/plain'                             // 1570
            });                                                        //
            response.end('expected XML request body');                 // 1572
            throw new Error('expected XML request body');              // 1573
          }                                                            //
                                                                       //
          var charset = match[2] || 'ISO-8859-1';                      // 1576
                                                                       //
          var encoding = request.headers['content-encoding'];          // 1578
          if (encoding && encoding !== 'identity') {                   // 1579
            if (encoding !== 'gzip') throw new Error('unknown Content-Encoding: ' + encoding);
            data = gunzipSync(data);                                   // 1581
          }                                                            //
                                                                       //
          return data.toString(charset.toLowerCase() === 'utf-8' ? 'utf8' : 'binary');
        };                                                             //
                                                                       //
        var propfindDepth = function () {                              // 1587
          var depth = request.headers['depth'];                        // 1588
          return depth === '0' ? 'zero' : depth === '1' ? 'one' : 'infinity';
        };                                                             //
                                                                       //
        var shallow = function () {                                    // 1594
          return request.headers['depth'] === '0';                     // 1595
        };                                                             //
                                                                       //
        var noOverwrite = function () {                                // 1598
          return (request.headers['overwrite'] || '').toLowerCase() === 'f';
        };                                                             //
                                                                       //
        var destination = function () {                                // 1602
          var result = request.headers['destination'];                 // 1603
          if (!result) throw new Error('missing destination');         // 1604
          return Url.parse(result).path.slice(1); // remove leading '/'
        };                                                             //
                                                                       //
        if (request.method === 'GET' || request.method === 'HEAD') {   // 1608
          return session.get(path, context, request.method === 'HEAD');
        } else if (request.method === 'POST') {                        //
          return session.post(path, requestContent(), context);        // 1611
        } else if (request.method === 'PUT') {                         //
          return session.put(path, requestContent(), context);         // 1613
        } else if (request.method === 'PATCH') {                       //
          return session.patch(path, requestContent(), context);       // 1615
        } else if (request.method === 'DELETE') {                      //
          return session['delete'](path, context);                     // 1617
        } else if (request.method === 'PROPFIND') {                    //
          return session.propfind(path, xmlContent(), propfindDepth(), context);
        } else if (request.method === 'PROPPATCH') {                   //
          return session.proppatch(path, xmlContent(), context);       // 1621
        } else if (request.method === 'MKCOL') {                       //
          return session.mkcol(path, requestContent(), context);       // 1623
        } else if (request.method === 'COPY') {                        //
          return session.copy(path, destination(), noOverwrite(), shallow(), context);
        } else if (request.method === 'MOVE') {                        //
          return session.move(path, destination(), noOverwrite(), context);
        } else if (request.method === 'LOCK') {                        //
          return session.lock(path, xmlContent(), shallow(), context);
        } else if (request.method === 'UNLOCK') {                      //
          return session.unlock(path, request.headers['lock-token'], context);
        } else if (request.method === 'ACL') {                         //
          return session.acl(path, xmlContent(), context);             // 1633
        } else if (request.method === 'REPORT') {                      //
          return session.report(path, requestContent(), context);      // 1635
        } else if (request.method === 'OPTIONS') {                     //
          return session.options(path, context).then(function (options) {
            var dav = [];                                              // 1638
            if (options.davClass1) dav.push('1');                      // 1639
            if (options.davClass2) dav.push('2');                      // 1640
            if (options.davClass3) dav.push('3');                      // 1641
            if (dav.length > 0) response.setHeader('DAV', dav.join(', '));
            response.end();                                            // 1643
            // Return no response; we already handled everything.      //
          }, function (err) {                                          //
            if (err.kjType !== 'unimplemented') throw err;             // 1646
            response.end();                                            // 1647
            // Return no response; we already handled everything.      //
          });                                                          //
        } else {                                                       //
            throw new Error('Sandstorm only supports the following methods: GET, POST, PUT, PATCH, DELETE, HEAD, PROPFIND, PROPPATCH, MKCOL, COPY, MOVE, LOCK, UNLOCK, ACL, REPORT, and OPTIONS.');
          }                                                            //
      }).then(function (rpcResponse) {                                 //
        if (rpcResponse !== undefined) {                               // 1654
          // Will be undefined for OPTIONS request.                    //
          return _this.translateResponse(rpcResponse, response, request);
        }                                                              //
      })['catch'](function (error) {                                   //
        return _this.maybeRetryAfterError(error, retryCount).then(function () {
          return _this.handleRequest(request, data, response, retryCount + 1);
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleRequest;                                              //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleRequestStreaming = (function () {              // 945
    function handleRequestStreaming(request, response, contentLength, retryCount) {
      var _this = this;                                                // 1665
      var context = this.makeContext(request, response);               // 1666
      var path = request.url.slice(1); // remove leading '/'           // 1667
      var session = this.getSession(request);                          // 1668
                                                                       //
      var mimeType = request.headers['content-type'] || 'application/octet-stream';
      var encoding = request.headers['content-encoding'];              // 1671
                                                                       //
      var requestStreamPromise = undefined;                            // 1673
      if (request.method === 'POST') {                                 // 1674
        requestStreamPromise = session.postStreaming(path, mimeType, context, encoding);
      } else if (request.method === 'PUT') {                           //
        requestStreamPromise = session.putStreaming(path, mimeType, context, encoding);
      } else {                                                         //
        throw new Error('Sandstorm only supports streaming POST and PUT requests.');
      }                                                                //
                                                                       //
      // TODO(perf): We ought to be pipelining the body, but we can't currently, because we have to
      //   handle the case where the app doesn't actually support streaming. We could pipeline while
      //   also buffering the data on the side in case we need it again later, but that's kind of
      //   complicated. We should fix the whole protocol to make streaming the standard.
      return requestStreamPromise.then(function (requestStreamResult) {
        var requestStream = requestStreamResult.stream;                // 1687
                                                                       //
        // Initialized when getResponse() returns, if the response is streaming.
        var downloadStreamHandle = undefined;                          // 1690
                                                                       //
        // Initialized if an upload-stream method throws.              //
        var uploadStreamError = undefined;                             // 1693
                                                                       //
        // We call `getResponse()` immediately so that the app can start streaming data down even while
        // data is still being streamed up. This theoretically allows apps to perform bidirectional
        // streaming, though probably very few actually do that.       //
        //                                                             //
        // Note that we need to be able to cancel `responsePromise` below, so it's important that it is
        // the raw Cap'n Proto promise. Hence `translateResponsePromise` is a separate variable.
        var responsePromise = requestStream.getResponse();             // 1701
                                                                       //
        var reportUploadStreamError = function (err) {                 // 1703
          // Called when an upload-stream method throws.               //
                                                                       //
          if (!uploadStreamError) {                                    // 1706
            uploadStreamError = err;                                   // 1707
                                                                       //
            // If we're still waiting on any response stuff, cancel it.
            responsePromise.cancel();                                  // 1710
            requestStream.close();                                     // 1711
            if (downloadStreamHandle) {                                // 1712
              downloadStreamHandle.close();                            // 1713
            }                                                          //
          }                                                            //
        };                                                             //
                                                                       //
        // If we have a Content-Length, pass it along to the app by calling `expectSize()`.
        if (contentLength !== undefined) {                             // 1719
          requestStream.expectSize(contentLength)['catch'](function (err) {
            // expectSize() is allowed to be unimplemented.            //
            if (err.kjType !== 'unimplemented') {                      // 1722
              reportUploadStreamError(err);                            // 1723
            }                                                          //
          });                                                          //
        }                                                              //
                                                                       //
        // Pipe the input stream to the app.                           //
        request.on('data', function (buf) {                            // 1729
          // TODO(soon): Only allow a small number of write()s to be in-flight at once,
          //   pausing the input stream if we hit that limit, so that we block the TCP socket all the
          //   way back to the source. May want to also coalesce small writes for this purpose.
          // TODO(security): The above problem may allow a DoS attack on the front-end.
          if (!uploadStreamError) requestStream.write(buf)['catch'](reportUploadStreamError);
        });                                                            //
                                                                       //
        request.on('end', function () {                                // 1737
          if (!uploadStreamError) requestStream.done()['catch'](reportUploadStreamError);
                                                                       //
          // We're all done making calls to requestStream.             //
          requestStream.close();                                       // 1741
        });                                                            //
                                                                       //
        request.on('close', function () {                              // 1744
          reportUploadStreamError(new Error('HTTP connection unexpectedly closed during request.'));
        });                                                            //
                                                                       //
        request.on('error', function (err) {                           // 1748
          reportUploadStreamError(err);                                // 1749
        });                                                            //
                                                                       //
        return responsePromise.then(function (rpcResponse) {           // 1752
          // Stop here if the upload stream has already failed.        //
          if (uploadStreamError) throw uploadStreamError;              // 1754
          var promise = _this.translateResponse(rpcResponse, response, request);
          downloadStreamHandle = promise.streamHandle;                 // 1756
          return promise;                                              // 1757
        });                                                            //
      }, function (err) {                                              //
        if (err.kjType === 'failed' && err.message.indexOf('not implemented') !== -1) {
          // Hack to work around old apps using an old version of Cap'n Proto, before the
          // 'unimplemented' exception type was introduced. :(         //
          // TODO(cleanup): When we transition to API version 2, we can move this into the
          //   compatibility layer.                                    //
          err.kjType = 'unimplemented';                                // 1765
        }                                                              //
                                                                       //
        if (SandstormBackend.shouldRestartGrain(err, 0)) {             // 1768
          // This is the kind of error that indicates we should retry. Note that we passed 0 for the
          // retry count above because we were just checking if this is a retriable error (vs. possibly
          // a method-not-implemented error); maybeRetryAfterError() will check again with the proper
          // retry count.                                              //
          return _this.maybeRetryAfterError(err, retryCount).then(function () {
            return _this.handleRequestStreaming(request, response, contentLength, retryCount + 1);
          });                                                          //
        } else if (err.kjType === 'unimplemented') {                   //
          // Streaming is not implemented. Fall back to non-streaming version.
          return readAll(request).then(function (data) {               // 1778
            return _this.handleRequest(request, data, response, 0);    // 1779
          });                                                          //
        } else {                                                       //
          throw err;                                                   // 1782
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleRequestStreaming;                                     //
  })();                                                                //
                                                                       //
  Proxy.prototype.handleWebSocket = (function () {                     // 945
    function handleWebSocket(request, socket, head, retryCount) {      // 1787
      var _this = this;                                                // 1788
                                                                       //
      return Promise.resolve(undefined).then(function () {             // 1790
        return _this.makeContext(request);                             // 1791
      }).then(function (context) {                                     //
        var path = request.url.slice(1); // remove leading '/'         // 1793
        var session = _this.getSession(request);                       // 1794
                                                                       //
        if (!('sec-websocket-key' in request.headers)) {               // 1796
          throw new Error('Missing Sec-WebSocket-Accept header.');     // 1797
        }                                                              //
                                                                       //
        var magic = request.headers['sec-websocket-key'] + '258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
        var acceptKey = Crypto.createHash('sha1').update(magic).digest('base64');
                                                                       //
        var protocols = [];                                            // 1803
        if ('sec-websocket-protocol' in request.headers) {             // 1804
          protocols = request.headers['sec-websocket-protocol'].split(',').map(function (s) {
            return s.trim();                                           // 1806
          });                                                          //
        }                                                              //
                                                                       //
        var receiver = new WebSocketReceiver(socket);                  // 1809
        // TODO(someday): do we want to make these be weak references somehow?
        _this.websockets.push(socket);                                 // 1811
                                                                       //
        var promise = session.openWebSocket(path, context, protocols, receiver);
                                                                       //
        if (head.length > 0) {                                         // 1815
          promise.serverStream.sendBytes(head);                        // 1816
        }                                                              //
                                                                       //
        pumpWebSocket(socket, promise.serverStream);                   // 1819
                                                                       //
        return promise.then(function (response) {                      // 1821
          var headers = ['HTTP/1.1 101 Switching Protocols', 'Upgrade: websocket', 'Connection: Upgrade', 'Sec-WebSocket-Accept: ' + acceptKey];
          if (response.protocol && response.protocol.length > 0) {     // 1828
            headers.push('Sec-WebSocket-Protocol: ' + response.protocol.join(', '));
          }                                                            //
                                                                       //
          headers.push('');                                            // 1832
          headers.push('');                                            // 1833
                                                                       //
          socket.write(headers.join('\r\n'));                          // 1835
          receiver.go();                                               // 1836
                                                                       //
          // Note:  At this point errors are out of our hands.         //
        });                                                            //
      })['catch'](function (error) {                                   //
        return _this.maybeRetryAfterError(error, retryCount).then(function () {
          return _this.handleWebSocket(request, socket, head, retryCount + 1);
        });                                                            //
      });                                                              //
    }                                                                  //
                                                                       //
    return handleWebSocket;                                            //
  })();                                                                //
                                                                       //
  Proxy.prototype.setHasLoaded = (function () {                        // 945
    function setHasLoaded() {                                          // 1847
      var _this4 = this;                                               //
                                                                       //
      if (!this.hasLoaded) {                                           // 1848
        (function () {                                                 //
          _this4.hasLoaded = true;                                     // 1849
          var sessionId = _this4.sessionId;                            // 1850
          inMeteor(function () {                                       // 1851
            Sessions.update({ _id: sessionId }, { $set: { hasLoaded: true } });
          });                                                          //
        })();                                                          //
      }                                                                //
    }                                                                  //
                                                                       //
    return setHasLoaded;                                               //
  })();                                                                //
                                                                       //
  return Proxy;                                                        //
})();                                                                  //
                                                                       //
var PROTOCOL = Url.parse(process.env.ROOT_URL).protocol;               // 1858
                                                                       //
var isRfc1918OrLocal = function (address) {                            // 1860
  if (Net.isIPv4(address)) {                                           // 1861
    quad = address.split('.').map(function (x) {                       // 1862
      return parseInt(x, 10);                                          // 1862
    });                                                                //
    return quad[0] === 127 || quad[0] === 10 || quad[0] === 192 && quad[1] === 168 || quad[0] === 172 && quad[1] >= 16 && quad[1] < 32;
  } else if (Net.isIPv6(address)) {                                    //
    // IPv6 specifies ::1 as localhost and fd:: as reserved for private networks
    return address === '::1' || address.lastIndexOf('fd', 0) === 0;    // 1868
  } else {                                                             //
    // Ignore things that are neither IPv4 nor IPv6                    //
    return false;                                                      // 1871
  }                                                                    //
};                                                                     //
                                                                       //
var quadToIntString = function (quad) {                                // 1875
  var num = Bignum(quad[0]).shiftLeft(16).add(quad[1]).shiftLeft(16).add(quad[2]).shiftLeft(16).add(quad[3]);
  return num.toString();                                               // 1880
};                                                                     //
                                                                       //
// -----------------------------------------------------------------------------
// Session cookie management                                           //
                                                                       //
var parseCookies = function (request) {                                // 1886
  // jscs:disable requireDotNotation                                   //
  var header = request.headers['cookie'];                              // 1888
                                                                       //
  var result = { cookies: [] };                                        // 1890
  if (header) {                                                        // 1891
    var reqCookies = header.split(';');                                // 1892
    for (var i in babelHelpers.sanitizeForInObject(reqCookies)) {      // 1893
      var reqCookie = reqCookies[i];                                   // 1894
      var equalsPos = reqCookie.indexOf('=');                          // 1895
      var cookie = undefined;                                          // 1896
      if (equalsPos === -1) {                                          // 1897
        cookie = { key: reqCookie.trim(), value: '' };                 // 1898
      } else {                                                         //
        cookie = { key: reqCookie.slice(0, equalsPos).trim(), value: reqCookie.slice(equalsPos + 1) };
      }                                                                //
                                                                       //
      if (cookie.key === 'sandstorm-sid') {                            // 1903
        if (result.sessionId) {                                        // 1904
          throw new Error('Multiple sandstorm session IDs?');          // 1905
        }                                                              //
                                                                       //
        result.sessionId = cookie.value;                               // 1908
      } else {                                                         //
        result.cookies.push(cookie);                                   // 1910
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  return result;                                                       // 1915
};                                                                     //
                                                                       //
var parsePreconditionHeader = function (request) {                     // 1918
  if (request.headers['if-match']) {                                   // 1919
    if (request.headers['if-match'].trim() === '*') {                  // 1920
      return { exists: null };                                         // 1921
    }                                                                  //
                                                                       //
    var matches = parseETagList(request.headers['if-match']);          // 1924
    if (matches.length > 0) {                                          // 1925
      return { matchesOneOf: matches };                                // 1926
    }                                                                  //
  }                                                                    //
                                                                       //
  if (request.headers['if-none-match']) {                              // 1930
    var noneMatches = parseETagList(request.headers['if-none-match']);
    if (noneMatches.length > 0) {                                      // 1932
      return { matchesNoneOf: noneMatches };                           // 1933
    }                                                                  //
  }                                                                    //
                                                                       //
  return { none: null };                                               // 1937
};                                                                     //
                                                                       //
var parseETagList = function (input) {                                 // 1940
  // An ETag is a quoted, \-escaped string, possibly prefixed with W/ (outside the quotes) to
  // indicate that it is weak. We are parsing a list of comma-delimited etags.
                                                                       //
  input = input.trim();                                                // 1944
  var results = [];                                                    // 1945
                                                                       //
  while (input.length > 0) {                                           // 1947
    var match = input.match(/^\s*(W\/)?"(([^"\\]|\\.)*)"\s*($|,)/);    // 1948
    if (!match) throw new Meteor.Error(400, 'invalid etag');           // 1949
                                                                       //
    input = input.slice(match[0].length).trim();                       // 1951
    results.push({ weak: !!match[1], value: match[2].replace(/\\(.)/g, '$1') });
  }                                                                    //
                                                                       //
  return results;                                                      // 1955
};                                                                     //
                                                                       //
var composeETag = function (tag) {                                     // 1958
  var result = '"' + (tag.value || '').replace(/([\\"])/g, '\\$1') + '"';
  if (tag.weak) result = 'W/' + result;                                // 1960
  return result;                                                       // 1961
};                                                                     //
                                                                       //
var parseAcceptHeader = function (request) {                           // 1964
  // jscs:disable requireDotNotation                                   //
  var header = request.headers['accept'];                              // 1966
                                                                       //
  var result = [];                                                     // 1968
  if (header) {                                                        // 1969
    var acceptList = header.split(',');                                // 1970
    for (var i in babelHelpers.sanitizeForInObject(acceptList)) {      // 1971
      var acceptStr = acceptList[i];                                   // 1972
      var tokensList = acceptStr.split(';');                           // 1973
                                                                       //
      var temp = { mimeType: tokensList[0].trim() };                   // 1975
                                                                       //
      var tokensListRest = tokensList.slice(1);                        // 1977
      for (var j in babelHelpers.sanitizeForInObject(tokensListRest)) {
        var token = tokensListRest[j];                                 // 1979
        var equalsPos = token.indexOf('=');                            // 1980
        if (equalsPos) {                                               // 1981
          var key = token.slice(0, equalsPos).trim();                  // 1982
          var value = token.slice(equalsPos + 1).trim();               // 1983
                                                                       //
          if (key === 'q') {                                           // 1985
            temp.qValue = +value;                                      // 1986
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      result.push(temp);                                               // 1991
    }                                                                  //
  }                                                                    //
                                                                       //
  return result;                                                       // 1995
};                                                                     //
                                                                       //
// -----------------------------------------------------------------------------
// Regular HTTP request handling                                       //
                                                                       //
var readAll = function (stream) {                                      // 2001
  return new Promise(function (resolve, reject) {                      // 2002
    var buffers = [];                                                  // 2003
    stream.on('data', function (buf) {                                 // 2004
      buffers.push(buf);                                               // 2005
    });                                                                //
                                                                       //
    stream.on('end', function () {                                     // 2008
      resolve(Buffer.concat(buffers));                                 // 2009
    });                                                                //
                                                                       //
    stream.on('error', reject);                                        // 2012
  });                                                                  //
};                                                                     //
                                                                       //
var makeSetCookieHeader = function (cookie) {                          // 2016
  var result = [cookie.name, '=', cookie.value];                       // 2017
                                                                       //
  if ('absolute' in cookie.expires) {                                  // 2019
    result.push('; Expires=');                                         // 2020
    result.push(new Date(cookie.expires.absolute * 1000).toUTCString());
  } else if ('relative' in cookie.expires) {                           //
    result.push('; Max-Age=' + cookie.expires.relative);               // 2023
  }                                                                    //
                                                                       //
  if (cookie.path) {                                                   // 2026
    result.push('; Path=' + cookie.path);                              // 2027
  }                                                                    //
                                                                       //
  if (cookie.httpOnly) {                                               // 2030
    result.push('; HttpOnly');                                         // 2031
  }                                                                    //
                                                                       //
  return result.join('');                                              // 2034
};                                                                     //
                                                                       //
// TODO(cleanup):  Auto-generate based on annotations in web-session.capnp.
var successCodes = {                                                   // 2038
  ok: { id: 200, title: 'OK' },                                        // 2039
  created: { id: 201, title: 'Created' },                              // 2040
  accepted: { id: 202, title: 'Accepted' },                            // 2041
  multiStatus: { id: 207, title: 'Multi-Status' }                      // 2042
};                                                                     //
var noContentSuccessCodes = [                                          // 2044
// Indexed by shouldResetForm * 1                                      //
{ id: 204, title: 'No Content' }, { id: 205, title: 'Reset Content' }];
var redirectCodes = [                                                  // 2049
// Indexed by switchToGet * 2 + isPermanent                            //
{ id: 307, title: 'Temporary Redirect' }, { id: 308, title: 'Permanent Redirect' }, { id: 303, title: 'See Other' }, { id: 301, title: 'Moved Permanently' }];
var errorCodes = {                                                     // 2056
  badRequest: { id: 400, title: 'Bad Request' },                       // 2057
  forbidden: { id: 403, title: 'Forbidden' },                          // 2058
  notFound: { id: 404, title: 'Not Found' },                           // 2059
  methodNotAllowed: { id: 405, title: 'Method Not Allowed' },          // 2060
  notAcceptable: { id: 406, title: 'Not Acceptable' },                 // 2061
  conflict: { id: 409, title: 'Conflict' },                            // 2062
  gone: { id: 410, title: 'Gone' },                                    // 2063
  requestEntityTooLarge: { id: 413, title: 'Request Entity Too Large' },
  requestUriTooLong: { id: 414, title: 'Request-URI Too Long' },       // 2065
  unsupportedMediaType: { id: 415, title: 'Unsupported Media Type' },  // 2066
  imATeapot: { id: 418, title: 'I\'m a teapot' },                      // 2067
  unprocessableEntity: { id: 422, title: 'Unprocessable Entity' }      // 2068
};                                                                     //
                                                                       //
ResponseStream = (function () {                                        // 2071
  function ResponseStream(response, streamHandle, resolve, reject) {   // 2072
    babelHelpers.classCallCheck(this, ResponseStream);                 //
                                                                       //
    this.response = response;                                          // 2073
    this.streamHandle = streamHandle;                                  // 2074
    this.resolve = resolve;                                            // 2075
    this.reject = reject;                                              // 2076
    this.ended = false;                                                // 2077
  }                                                                    //
                                                                       //
  ResponseStream.prototype.write = (function () {                      // 2071
    function write(data) {                                             // 2080
      this.response.write(data);                                       // 2081
    }                                                                  //
                                                                       //
    return write;                                                      //
  })();                                                                //
                                                                       //
  ResponseStream.prototype.done = (function () {                       // 2071
    function done() {                                                  // 2084
      this.response.end();                                             // 2085
      this.streamHandle.close();                                       // 2086
      this.ended = true;                                               // 2087
    }                                                                  //
                                                                       //
    return done;                                                       //
  })();                                                                //
                                                                       //
  ResponseStream.prototype.close = (function () {                      // 2071
    function close() {                                                 // 2090
      if (this.ended) {                                                // 2091
        this.resolve();                                                // 2092
      } else {                                                         //
        this.streamHandle.close();                                     // 2094
        this.reject(new Error('done() was never called on outbound stream.'));
      }                                                                //
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  return ResponseStream;                                               //
})();                                                                  //
                                                                       //
// TODO(cleanup): Node 0.12 has a `gunzipSync` but 0.10 (which Meteor still uses) does not.
var Zlib = Npm.require('zlib');                                        // 2101
var gunzipSync = Meteor.wrapAsync(Zlib.gunzip, Zlib);                  // 2102
                                                                       //
// -----------------------------------------------------------------------------
// WebSocket handling                                                  //
                                                                       //
WebSocketReceiver = (function () {                                     // 2107
  function WebSocketReceiver(socket) {                                 // 2108
    babelHelpers.classCallCheck(this, WebSocketReceiver);              //
                                                                       //
    this.socket = socket;                                              // 2109
    this.queue = [];                                                   // 2110
  }                                                                    //
                                                                       //
  WebSocketReceiver.prototype.go = (function () {                      // 2107
    function go() {                                                    // 2113
      for (var i in babelHelpers.sanitizeForInObject(this.queue)) {    // 2114
        this.socket.write(this.queue[i]);                              // 2115
      }                                                                //
                                                                       //
      this.queue = null;                                               // 2118
    }                                                                  //
                                                                       //
    return go;                                                         //
  })();                                                                //
                                                                       //
  WebSocketReceiver.prototype.sendBytes = (function () {               // 2107
    function sendBytes(message) {                                      // 2121
      // TODO(someday):  Flow control of some sort?                    //
      if (this.queue === null) {                                       // 2123
        this.socket.write(message);                                    // 2124
      } else {                                                         //
        this.queue.push(message);                                      // 2126
      }                                                                //
    }                                                                  //
                                                                       //
    return sendBytes;                                                  //
  })();                                                                //
                                                                       //
  WebSocketReceiver.prototype.close = (function () {                   // 2107
    function close() {                                                 // 2130
      this.socket.end();                                               // 2131
    }                                                                  //
                                                                       //
    return close;                                                      //
  })();                                                                //
                                                                       //
  return WebSocketReceiver;                                            //
})();                                                                  //
                                                                       //
pumpWebSocket = function (socket, rpcStream) {                         // 2135
  socket.on('data', function (chunk) {                                 // 2136
    rpcStream.sendBytes(chunk)['catch'](function (err) {               // 2137
      if (err.kjType !== 'disconnected') {                             // 2138
        console.error('WebSocket sendBytes failed: ' + err.stack);     // 2139
      }                                                                //
                                                                       //
      socket.destroy();                                                // 2142
    });                                                                //
  });                                                                  //
                                                                       //
  socket.on('end', function (chunk) {                                  // 2146
    rpcStream.close();                                                 // 2147
  });                                                                  //
};                                                                     //
                                                                       //
// =======================================================================================
// Debug log access                                                    //
                                                                       //
Meteor.publish('grainLog', function (grainId) {                        // 2154
  var _this5 = this;                                                   //
                                                                       //
  check(grainId, String);                                              // 2155
  var id = 0;                                                          // 2156
  var grain = Grains.findOne(grainId);                                 // 2157
  if (!grain || !this.userId || grain.userId !== this.userId) {        // 2158
    this.added('grainLog', id++, { text: 'Only the grain owner can view the debug log.' });
    this.ready();                                                      // 2160
    return;                                                            // 2161
  }                                                                    //
                                                                       //
  var connected = false;                                               // 2164
  var _this = this;                                                    // 2165
                                                                       //
  var receiver = {                                                     // 2167
    write: function (data) {                                           // 2168
      connected = true;                                                // 2169
      _this.added('grainLog', id++, { text: data.toString('utf8') });  // 2170
    },                                                                 //
                                                                       //
    close: function () {                                               // 2173
      if (connected) {                                                 // 2174
        _this.added('grainLog', id++, {                                // 2175
          text: '*** lost connection to grain (probably because it shut down) ***'
        });                                                            //
      }                                                                //
    }                                                                  //
  };                                                                   //
                                                                       //
  try {                                                                // 2182
    (function () {                                                     //
      var handle = waitPromise(globalBackend.useGrain(grainId, function (supervisor) {
        return supervisor.watchLog(8192, receiver);                    // 2184
      })).handle;                                                      //
      connected = true;                                                // 2186
      _this5.onStop(function () {                                      // 2187
        handle.close();                                                // 2188
      });                                                              //
    })();                                                              //
  } catch (err) {                                                      //
    if (!connected) {                                                  // 2191
      this.added('grainLog', id++, {                                   // 2192
        text: '*** couldn\'t connect to grain (' + err + ') ***'       // 2193
      });                                                              //
    }                                                                  //
  }                                                                    //
                                                                       //
  // Notify ready.                                                     //
  this.ready();                                                        // 2199
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=proxy.js.map
