(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// shared/admin.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var ADMIN_TOKEN_EXPIRATION_TIME = 15 * 60 * 1000;                      // 17
var publicAdminSettings = ["google", "github", "emailToken", "splashUrl", "signupDialog", "adminAlert", "adminAlertTime", "adminAlertUrl", "termsUrl", "privacyUrl", "appMarketUrl", "appIndexUrl", "appUpdatesEnabled", "serverTitle", "returnAddress"];
                                                                       //
DEFAULT_SIGNUP_DIALOG = "You've been invited to join this Sandstorm server!";
                                                                       //
var adminRoute = RouteController.extend({                              // 25
  template: "admin",                                                   // 26
  waitOn: function () {                                                // 27
    var subs = [Meteor.subscribe("admin", this.params._token), Meteor.subscribe("adminServiceConfiguration", this.params._token), Meteor.subscribe("allUsers", this.params._token)];
    if (this.params._token) {                                          // 33
      subs.push(Meteor.subscribe("adminToken", this.params._token));   // 34
    }                                                                  //
    return subs;                                                       // 36
  },                                                                   //
                                                                       //
  data: function () {                                                  // 39
    var adminToken = AdminToken.findOne();                             // 40
    return {                                                           // 41
      settings: Settings.find(),                                       // 42
      token: this.params._token,                                       // 43
      isUserPermitted: isAdmin() || adminToken && adminToken.tokenIsValid
    };                                                                 //
  },                                                                   //
                                                                       //
  action: function () {                                                // 48
    var state = this.state;                                            // 49
    Meteor.call("getSmtpUrl", this.params._token, function (error, result) {
      state.set("smtpUrl", result);                                    // 51
    });                                                                //
    var user = Meteor.user();                                          // 53
    if (user && user.loginIdentities) {                                // 54
      if (this.params._token) {                                        // 55
        if (!user.signupKey || !user.isAdmin) {                        // 56
          Meteor.call("signUpAsAdmin", this.params._token);            // 57
        } else if (user.isAdmin) {                                     //
          // We don't need the token. Redirect to the current route, minus the token parameter.
          Router.go(this.route.getName(), {}, _.pick(this.params, "query", "hash"));
        }                                                              //
      }                                                                //
    }                                                                  //
    resetResult(state);                                                // 64
    state.set("configurationServiceName", null);                       // 65
    state.set("token", this.params._token);                            // 66
    this.render();                                                     // 67
  }                                                                    //
});                                                                    //
                                                                       //
Router.map(function () {                                               // 71
  this.route("adminSettings", {                                        // 72
    path: "/admin/settings/:_token?",                                  // 73
    controller: adminRoute                                             // 74
  });                                                                  //
  this.route("adminUsers", {                                           // 76
    path: "/admin/users/:_token?",                                     // 77
    controller: adminRoute                                             // 78
  });                                                                  //
  this.route("adminStats", {                                           // 80
    path: "/admin/stats/:_token?",                                     // 81
    controller: adminRoute                                             // 82
  });                                                                  //
  this.route("adminLog", {                                             // 84
    path: "/admin/log/:_token?",                                       // 85
    controller: adminRoute                                             // 86
  });                                                                  //
  this.route("adminInvites", {                                         // 88
    path: "/admin/invites/:_token?",                                   // 89
    controller: adminRoute                                             // 90
  });                                                                  //
  this.route("adminCaps", {                                            // 92
    path: "/admin/capabilities/:_token?",                              // 93
    controller: adminRoute                                             // 94
  });                                                                  //
  this.route("adminAdvanced", {                                        // 96
    path: "/admin/advanced/:_token?",                                  // 97
    controller: adminRoute                                             // 98
  });                                                                  //
  this.route("adminOld", {                                             // 100
    path: "/admin/:_token?",                                           // 101
    action: function () {                                              // 102
      this.redirect("adminSettings", this.params);                     // 103
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
if (Meteor.isClient) {                                                 // 108
  AdminToken = new Mongo.Collection("adminToken"); // see Meteor.publish("adminToken")
  AdminLog = new Meteor.Collection("adminLog");                        // 110
  Meteor.subscribe("publicAdminSettings");                             // 111
                                                                       //
  var resetResult = function (state) {                                 // 113
    state = state || Iron.controller().state;                          // 114
    state.set("numSettings", 1);                                       // 115
    state.set("successes", 0);                                         // 116
    state.set("failures", 0);                                          // 117
    state.set("errors", []);                                           // 118
    state.set("fadeAlert", false);                                     // 119
    state.set("successMessage", "Your settings have been saved.");     // 120
    state.set("powerboxOfferUrl", null);                               // 121
  };                                                                   //
                                                                       //
  var getToken = function () {                                         // 124
    var state = Iron.controller().state;                               // 125
    var token = state.get("token");                                    // 126
    if (!token) {                                                      // 127
      return;                                                          // 128
    } else {                                                           //
      return { _token: token };                                        // 130
    }                                                                  //
  };                                                                   //
                                                                       //
  Template.admin.helpers({                                             // 134
    adminTab: function () {                                            // 135
      return Router.current().route.getName();                         // 136
    },                                                                 //
    success: function () {                                             // 138
      var state = Iron.controller().state;                             // 139
      return state.get("successes") == state.get("numSettings");       // 140
    },                                                                 //
    failure: function () {                                             // 142
      return Iron.controller().state.get("failures");                  // 143
    },                                                                 //
    errors: function () {                                              // 145
      return Iron.controller().state.get("errors");                    // 146
    },                                                                 //
    fadeAlert: function () {                                           // 148
      return Iron.controller().state.get("fadeAlert");                 // 149
    },                                                                 //
    successMessage: function () {                                      // 151
      return Iron.controller().state.get("successMessage");            // 152
    },                                                                 //
    settingsActive: function () {                                      // 154
      return Router.current().route.getName() == "adminSettings";      // 155
    },                                                                 //
    usersActive: function () {                                         // 157
      return Router.current().route.getName() == "adminUsers";         // 158
    },                                                                 //
    invitesActive: function () {                                       // 160
      return Router.current().route.getName() == "adminInvites";       // 161
    },                                                                 //
    statsActive: function () {                                         // 163
      return Router.current().route.getName() == "adminStats";         // 164
    },                                                                 //
    logActive: function () {                                           // 166
      return Router.current().route.getName() == "adminLog";           // 167
    },                                                                 //
    capsActive: function () {                                          // 169
      return Router.current().route.getName() == "adminCaps";          // 170
    },                                                                 //
    advancedActive: function () {                                      // 172
      return Router.current().route.getName() == "adminAdvanced";      // 173
    },                                                                 //
    getToken: getToken                                                 // 175
  });                                                                  //
                                                                       //
  var handleError = function (err) {                                   // 178
    var state = this;                                                  // 179
    Meteor.setTimeout(function () {                                    // 180
      state.set("fadeAlert", true);                                    // 181
    }, 3000);                                                          //
    if (err) {                                                         // 183
      this.set("failures", this.get("failures") + 1);                  // 184
      console.error(err);                                              // 185
      var errors = this.get("errors");                                 // 186
      errors.push(err);                                                // 187
      this.set("errors", errors);                                      // 188
    } else {                                                           //
      this.set("successes", this.get("successes") + 1);                // 190
    }                                                                  //
  };                                                                   //
                                                                       //
  Template.adminSettings.events({                                      // 194
    "click .oauth-checkbox": function (event) {                        // 195
      var state = Iron.controller().state;                             // 196
      var serviceName = event.target.getAttribute("data-servicename");
      var config = Package["service-configuration"].ServiceConfiguration.configurations.findOne({ service: serviceName });
                                                                       //
      var setting = Settings.findOne({ _id: serviceName });            // 200
      if (event.target.checked && (!config || setting && setting.automaticallyReset)) {
        state.set("configurationServiceName", serviceName);            // 202
      }                                                                //
    },                                                                 //
    "click .configure-oauth": function (event) {                       // 205
      var state = Iron.controller().state;                             // 206
      state.set("configurationServiceName", event.target.getAttribute("data-servicename"));
    },                                                                 //
    "click .reset-login-tokens": function (event) {                    // 209
      var state = Iron.controller().state;                             // 210
      resetResult(state);                                              // 211
      var handleErrorBound = handleError.bind(state);                  // 212
      Meteor.call("clearResumeTokensForService", this.token, event.target.getAttribute("data-servicename"), handleErrorBound);
    },                                                                 //
    "click #admin-settings-send-toggle": function (event) {            // 216
      var state = Iron.controller().state;                             // 217
      state.set("isEmailTestActive", !state.get("isEmailTestActive"));
      return false; // prevent form from submitting                    // 219
    },                                                                 //
    "click #admin-settings-send-test": function (event) {              // 221
      var state = Iron.controller().state;                             // 222
      resetResult(state);                                              // 223
      var handleErrorBound = handleError.bind(state);                  // 224
      state.set("successMessage", "Email has been sent.");             // 225
      Meteor.call("testSend", this.token, document.getElementById("smptUrl").value, document.getElementById("email-test-to").value, handleErrorBound);
      return false; // prevent form from submitting                    // 228
    },                                                                 //
    "submit #admin-settings-form": function (event) {                  // 230
      var state = Iron.controller().state;                             // 231
      var token = this.token;                                          // 232
      resetResult(state);                                              // 233
      state.set("numSettings", 4);                                     // 234
                                                                       //
      var handleErrorBound = handleError.bind(state);                  // 236
      if (event.target.emailTokenLogin.checked && !event.target.smtpUrl.value) {
        handleErrorBound(new Meteor.Error(400, "You must configure an SMTP server to use email login."));
        return false;                                                  // 240
      }                                                                //
      Meteor.call("setAccountSetting", token, "google", event.target.googleLogin.checked, handleErrorBound);
      Meteor.call("setAccountSetting", token, "github", event.target.githubLogin.checked, handleErrorBound);
      Meteor.call("setAccountSetting", token, "emailToken", event.target.emailTokenLogin.checked, handleErrorBound);
      Meteor.call("setSetting", token, "smtpUrl", event.target.smtpUrl.value, handleErrorBound);
      return false;                                                    // 246
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.adminSettings.helpers({                                     // 250
    setDocumentTitle: function () {                                    // 251
      document.title = "Settings · Admin · " + globalDb.getServerTitle();
    },                                                                 //
    googleSetting: function () {                                       // 254
      return Settings.findOne({ _id: "google" });                      // 255
    },                                                                 //
    githubSetting: function () {                                       // 257
      return Settings.findOne({ _id: "github" });                      // 258
    },                                                                 //
    emailTokenEnabled: function () {                                   // 260
      var setting = Settings.findOne({ _id: "emailToken" });           // 261
      if (setting) {                                                   // 262
        return setting.value;                                          // 263
      } else {                                                         //
        return false;                                                  // 265
      }                                                                //
    },                                                                 //
    smtpUrl: function () {                                             // 268
      return Iron.controller().state.get("smtpUrl");                   // 269
    },                                                                 //
    isEmailTestActive: function () {                                   // 271
      return Iron.controller().state.get("isEmailTestActive");         // 272
    },                                                                 //
    getToken: getToken                                                 // 274
  });                                                                  //
                                                                       //
  var updateUser = function (options) {                                // 277
    var state = Iron.controller().state;                               // 278
    var token = state.get("token");                                    // 279
    resetResult(state);                                                // 280
    state.set("successMessage", "User has been updated.");             // 281
    var handleErrorBound = handleError.bind(state);                    // 282
    Meteor.call("adminUpdateUser", token, options, handleErrorBound);  // 283
  };                                                                   //
  Template.adminUsers.events({                                         // 285
    "change select.user-class": function (event) {                     // 286
      var value = event.target.selectedOptions[0].value;               // 287
                                                                       //
      if (value == "admin") {                                          // 289
        updateUser({ userId: this._id, signupKey: true, isAdmin: true });
      } else if (value == "invited") {                                 //
        updateUser({ userId: this._id, signupKey: true, isAdmin: false });
      } else if (value == "guest") {                                   //
        updateUser({ userId: this._id, signupKey: false, isAdmin: false });
      } else {                                                         //
        console.error("unrecognized user class");                      // 296
      }                                                                //
    },                                                                 //
    "change .is-signedup-checkbox": function (event) {                 // 299
      // The userid is stored on the the <tr>, which is always 2 nodes up
      var userId = event.target.parentElement.parentElement.getAttribute("data-userid");
    }                                                                  //
  });                                                                  //
  Template.adminUsers.helpers({                                        // 304
    setDocumentTitle: function () {                                    // 305
      document.title = "Users · Admin · " + globalDb.getServerTitle();
    },                                                                 //
    users: function () {                                               // 308
      return Meteor.users.find({ loginIdentities: { $exists: 1 } }, { sort: { createdAt: 1 } });
    },                                                                 //
    userIdentity: function () {                                        // 311
      var identityId = SandstormDb.getUserIdentityIds(this)[0];        // 312
      var identity = Meteor.users.findOne({ _id: identityId });        // 313
      if (identity) {                                                  // 314
        SandstormDb.fillInProfileDefaults(identity);                   // 315
        SandstormDb.fillInIntrinsicName(identity);                     // 316
        return identity;                                               // 317
      }                                                                //
    },                                                                 //
    userSignupNote: function () {                                      // 320
      if (this.signupEmail) {                                          // 321
        return this.signupEmail;                                       // 322
      } else if (this.signupNote) {                                    //
        return this.signupNote;                                        // 324
      } else {                                                         //
        return "";                                                     // 326
      }                                                                //
    },                                                                 //
    userIsAdmin: function () {                                         // 329
      return !!this.isAdmin;                                           // 330
    },                                                                 //
    userIsInvited: function () {                                       // 332
      return !this.isAdmin && !!this.signupKey;                        // 333
    },                                                                 //
    userIsGuest: function () {                                         // 335
      return !this.isAdmin && !this.signupKey;                         // 336
    },                                                                 //
    userStorageUsage: function () {                                    // 338
      return typeof this.storageUsage === "number" ? prettySize(this.storageUsage) : "";
    }                                                                  //
  });                                                                  //
                                                                       //
  var configureLoginServiceDialogTemplateForService = function (serviceName) {
    return Template['configureLoginServiceDialogFor' + capitalize(serviceName)];
  };                                                                   //
                                                                       //
  var configurationFields = function (serviceName) {                   // 347
    var template = configureLoginServiceDialogTemplateForService(serviceName);
    return template ? template.fields() : [];                          // 349
  };                                                                   //
                                                                       //
  Template._adminConfigureLoginServiceDialog.helpers({                 // 352
    configurationFields: function () {                                 // 353
      var serviceName = Iron.controller().state.get("configurationServiceName");
      var configurations = Package['service-configuration'].ServiceConfiguration.configurations;
      var configuration = configurations.findOne({ service: serviceName });
      var fields = configurationFields(serviceName);                   // 357
      if (configuration) {                                             // 358
        return _.map(fields, function (field) {                        // 359
          field.value = configuration[field.property];                 // 360
          return field;                                                // 361
        });                                                            //
      } else {                                                         //
        return fields;                                                 // 364
      }                                                                //
    },                                                                 //
    visible: function () {                                             // 367
      return Iron.controller().state.get("configurationServiceName") !== null;
    },                                                                 //
    configurationSteps: function () {                                  // 370
      // renders the appropriate template                              //
      return configureLoginServiceDialogTemplateForService(Iron.controller().state.get("configurationServiceName"));
    }                                                                  //
  });                                                                  //
                                                                       //
  var capitalize = function (str) {                                    // 377
    return str.charAt(0).toUpperCase() + str.slice(1);                 // 378
  };                                                                   //
                                                                       //
  Template._adminConfigureLoginServiceDialog.events({                  // 381
    'click .configure-login-service-dismiss-button': function () {     // 382
      Iron.controller().state.set("configurationServiceName", null);   // 383
    },                                                                 //
    'click #configure-login-service-dialog-save-configuration': function () {
      var state = Iron.controller().state;                             // 386
      resetResult(state);                                              // 387
                                                                       //
      // This is a bit of a hack, but we set the number high so that a success message is never displayed
      state.set("numSettings", 100);                                   // 390
                                                                       //
      var handleErrorBound = handleError.bind(state);                  // 392
      var serviceName = state.get("configurationServiceName");         // 393
      var token = this.token;                                          // 394
      var configuration = {                                            // 395
        service: serviceName                                           // 396
      };                                                               //
                                                                       //
      // Fetch the value of each input field                           //
      _.each(configurationFields(serviceName), function (field) {      // 400
        configuration[field.property] = document.getElementById('configure-login-service-dialog-' + field.property).value.trim();
      });                                                              //
                                                                       //
      configuration.loginStyle = "redirect";                           // 405
                                                                       //
      Meteor.call("adminConfigureLoginService", token, configuration, function (err) {
        handleErrorBound(err);                                         // 408
        state.set("configurationServiceName", null);                   // 409
      });                                                              //
      Meteor.call("setAccountSetting", token, serviceName, true, handleErrorBound);
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.adminInvites.events({                                       // 415
    "click #send": function (event) {                                  // 416
      var state = Iron.controller().state;                             // 417
      var from = document.getElementById("invite-from").value;         // 418
      var list = document.getElementById("invite-emails").value;       // 419
      var subject = document.getElementById("invite-subject").value;   // 420
      var message = document.getElementById("invite-message").value;   // 421
      var quotaInput = document.getElementById("invite-quota");        // 422
      var quota;                                                       // 423
      if (quotaInput && quotaInput.value.trim() !== "") {              // 424
        quota = parseInt(quotaInput.value);                            // 425
      }                                                                //
                                                                       //
      var sendButton = event.currentTarget;                            // 428
      sendButton.disabled = true;                                      // 429
      var oldContent = sendButton.textContent;                         // 430
      sendButton.textContent = "Sending...";                           // 431
                                                                       //
      Meteor.call("sendInvites", state.get("token"), getOrigin(), from, list, subject, message, quota, function (error, results) {
        sendButton.disabled = false;                                   // 435
        sendButton.textContent = oldContent;                           // 436
        if (error) {                                                   // 437
          state.set("inviteMessage", { error: error.toString() });     // 438
        } else {                                                       //
          state.set("inviteMessage", results);                         // 440
        }                                                              //
      });                                                              //
    },                                                                 //
                                                                       //
    "click #create": function (event) {                                // 445
      var state = Iron.controller().state;                             // 446
      var note = document.getElementById("key-note").value;            // 447
      var quotaInput = document.getElementById("key-quota");           // 448
      var quota;                                                       // 449
      if (quotaInput && quotaInput.value.trim() !== "") {              // 450
        quota = parseInt(quotaInput.value);                            // 451
      }                                                                //
                                                                       //
      Meteor.call("createSignupKey", state.get("token"), note, quota, function (error, key) {
        if (error) {                                                   // 455
          state.set("inviteMessage", { error: error.toString() });     // 456
        } else {                                                       //
          state.set("inviteMessage", {                                 // 458
            url: getOrigin() + Router.routes.signup.path({ key: key })
          });                                                          //
        }                                                              //
      });                                                              //
    },                                                                 //
                                                                       //
    "click #set-quota-submit": function (event) {                      // 465
      var state = Iron.controller().state;                             // 466
      var list = document.getElementById("set-quota-emails").value;    // 467
      var quotaInput = document.getElementById("set-quota-quota");     // 468
      var quota;                                                       // 469
      if (quotaInput && quotaInput.value.trim() !== "") {              // 470
        quota = parseInt(quotaInput.value);                            // 471
      }                                                                //
                                                                       //
      var updateButton = event.currentTarget;                          // 474
      updateButton.disabled = true;                                    // 475
      var oldContent = updateButton.textContent;                       // 476
      updateButton.textContent = "Updating...";                        // 477
                                                                       //
      Meteor.call("updateQuotas", state.get("token"), list, quota, function (error, results) {
        updateButton.disabled = false;                                 // 480
        updateButton.textContent = oldContent;                         // 481
        if (error) {                                                   // 482
          state.set("inviteMessage", { error: error.toString() });     // 483
        } else {                                                       //
          document.getElementById("set-quota-emails").value = "";      // 485
        }                                                              //
      });                                                              //
    },                                                                 //
                                                                       //
    "click .autoSelect": function (event) {                            // 490
      event.currentTarget.select();                                    // 491
    },                                                                 //
                                                                       //
    "click #retry": function (event) {                                 // 494
      Iron.controller().state.set("inviteMessage", undefined);         // 495
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.adminInvites.helpers({                                      // 499
    setDocumentTitle: function () {                                    // 500
      document.title = "Invites · Admin · " + globalDb.getServerTitle();
    },                                                                 //
    error: function () {                                               // 503
      var res = Iron.controller().state.get("inviteMessage");          // 504
      return res && res.error;                                         // 505
    },                                                                 //
    email: function () {                                               // 507
      var me = Meteor.user();                                          // 508
      var email = me.services && me.services.google && me.services.google.email || me.services && me.services.github && me.services.github.email;
      if (email && me.profile.name) {                                  // 511
        email = me.profile.name + " <" + email + ">";                  // 512
      }                                                                //
      email = email || "";                                             // 514
      return email;                                                    // 515
    },                                                                 //
    url: function () {                                                 // 517
      var res = Iron.controller().state.get("inviteMessage");          // 518
      return res && res.url;                                           // 519
    },                                                                 //
    sent: function () {                                                // 521
      var res = Iron.controller().state.get("inviteMessage");          // 522
      return res && res.sent;                                          // 523
    }                                                                  //
  });                                                                  //
  var maybeScrollLog = function () {                                   // 526
    var elem = document.getElementById("adminLog");                    // 527
    if (elem) {                                                        // 528
      // The log already exists. It's about to be updated. Check if it's scrolled to the bottom
      // before the update.                                            //
      if (elem.scrollHeight - elem.scrollTop === elem.clientHeight) {  // 531
        // Indeed, so we want to scroll it back to the bottom after the update.
        Tracker.afterFlush(function () {                               // 533
          scrollLogToBottom(elem);                                     // 533
        });                                                            //
      }                                                                //
    } else {                                                           //
      // No element exists yet, but it's probably about to be created, in which case we definitely
      // want to scroll it.                                            //
      Tracker.afterFlush(function () {                                 // 538
        var elem2 = document.getElementById("adminLog");               // 539
        if (elem2) scrollLogToBottom(elem2);                           // 540
      });                                                              //
    }                                                                  //
  };                                                                   //
                                                                       //
  var scrollLogToBottom = function (elem) {                            // 545
    elem.scrollTop = elem.scrollHeight;                                // 546
  };                                                                   //
                                                                       //
  Template.adminLog.onCreated(function () {                            // 549
    var state = Iron.controller().state;                               // 550
    var token = state.get("token");                                    // 551
    this.subscribe("adminLog", token);                                 // 552
  });                                                                  //
  Template.adminLog.helpers({                                          // 554
    setDocumentTitle: function () {                                    // 555
      document.title = "Log · Admin · " + globalDb.getServerTitle();   // 556
    },                                                                 //
    html: function () {                                                // 558
      return AnsiUp.ansi_to_html(AdminLog.find({}, { $sort: { _id: 1 } }).map(function (entry) {
        return entry.text;                                             // 560
      }).join(""), { use_classes: true });                             //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.adminCaps.onCreated(function () {                           // 565
    var state = Iron.controller().state;                               // 566
    var token = state.get("token");                                    // 567
    this.subscribe("adminApiTokens", token);                           // 568
  });                                                                  //
                                                                       //
  Template.adminCaps.helpers({                                         // 571
    setDocumentTitle: function () {                                    // 572
      document.title = "Capabilities · Admin · " + globalDb.getServerTitle();
    },                                                                 //
    powerboxOfferUrl: function () {                                    // 575
      var state = Iron.controller().state;                             // 576
      return state.get("powerboxOfferUrl");                            // 577
    },                                                                 //
    caps: function () {                                                // 579
      return ApiTokens.find({ $or: [{ "frontendRef.ipNetwork": { $exists: true } }, { "frontendRef.ipInterface": { $exists: true } }] });
    },                                                                 //
    userName: function () {                                            // 583
      var userId = findAdminUserForToken(this);                        // 584
      var user = Meteor.users.findOne({ _id: userId });                // 585
      if (!user) {                                                     // 586
        return "no user";                                              // 587
      }                                                                //
      var identityId = SandstormDb.getUserIdentityIds(user)[0];        // 589
      var identity = Meteor.users.findOne({ _id: identityId });        // 590
      if (identity) {                                                  // 591
        SandstormDb.fillInProfileDefaults(identity);                   // 592
        return identity.profile.name;                                  // 593
      }                                                                //
    },                                                                 //
    isDisabled: function () {                                          // 596
      return !this.userId;                                             // 597
    },                                                                 //
    disabled: function () {                                            // 599
      return this.revoked;                                             // 600
    }                                                                  //
  });                                                                  //
                                                                       //
  var updateCap = function (capId, value) {                            // 604
    var state = Iron.controller().state;                               // 605
    resetResult(state);                                                // 606
    if (!value) {                                                      // 607
      state.set("successMessage", "Capability has been re-enabled.");  // 608
    } else {                                                           //
      state.set("successMessage", "Capability has been disabled.");    // 610
    }                                                                  //
    var handleErrorBound = handleError.bind(state);                    // 612
    Meteor.call("adminToggleDisableCap", state.get("token"), capId, value, handleErrorBound);
  };                                                                   //
                                                                       //
  Template.adminCaps.events({                                          // 616
    "click #offer-ipnetwork": function (event) {                       // 617
      var state = Iron.controller().state;                             // 618
      resetResult(state);                                              // 619
      state.set("successMessage", "IpNetwork webkey created. Look for it in the top bar.");
      Meteor.call("offerIpNetwork", this.token, function (err, webkey) {
        state.set("powerboxOfferUrl", webkey);                         // 622
        handleError.call(state, err);                                  // 623
      });                                                              //
      return false; // prevent form from submitting                    // 625
    },                                                                 //
    "click #offer-ipinterface": function (event) {                     // 627
      var state = Iron.controller().state;                             // 628
      resetResult(state);                                              // 629
      state.set("successMessage", "IpInterface webkey created. Look for it in the top bar.");
      Meteor.call("offerIpInterface", this.token, function (err, webkey) {
        state.set("powerboxOfferUrl", webkey);                         // 632
        handleError.call(state, err);                                  // 633
      });                                                              //
      return false; // prevent form from submitting                    // 635
    },                                                                 //
    "click #powerbox-offer-popup-closer": function (event) {           // 637
      var state = Iron.controller().state;                             // 638
      return state.set("powerboxOfferUrl", null);                      // 639
    },                                                                 //
    "click .disable-cap": function (event) {                           // 641
      var capId = event.target.getAttribute("data-id");                // 642
      var token = ApiTokens.findOne({ _id: capId });                   // 643
                                                                       //
      updateCap(capId, !token.revoked);                                // 645
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.adminAdvanced.events({                                      // 649
    "submit #admin-settings-form": function (event) {                  // 650
      var state = Iron.controller().state;                             // 651
      var token = this.token;                                          // 652
      resetResult(state);                                              // 653
      state.set("numSettings", 12);                                    // 654
                                                                       //
      var handleErrorBound = handleError.bind(state);                  // 656
      Meteor.call("setSetting", token, "serverTitle", event.target.serverTitle.value, handleErrorBound);
      Meteor.call("setSetting", token, "returnAddress", event.target.returnAddress.value, handleErrorBound);
      Meteor.call("setSetting", token, "splashUrl", event.target.splashUrl.value, handleErrorBound);
      Meteor.call("setSetting", token, "signupDialog", event.target.signupDialog.value, handleErrorBound);
      Meteor.call("setSetting", token, "termsUrl", event.target.termsUrl.value, handleErrorBound);
      Meteor.call("setSetting", token, "privacyUrl", event.target.privacyUrl.value, handleErrorBound);
      Meteor.call("setSetting", token, "adminAlert", event.target.adminAlert.value, handleErrorBound);
      Meteor.call("setSetting", token, "appMarketUrl", event.target.appMarketUrl.value, handleErrorBound);
      Meteor.call("setSetting", token, "appIndexUrl", event.target.appIndexUrl.value, handleErrorBound);
      Meteor.call("setSetting", token, "appUpdatesEnabled", event.target.appUpdatesEnabled.checked, handleErrorBound);
      var alertTimeString = event.target.alertTime.value.trim();       // 669
      if (alertTimeString) {                                           // 670
        var alertTime = new Date(alertTimeString);                     // 671
        if (isNaN(alertTime.getTime())) {                              // 672
          // Assume only time and not date was set.                    //
          alertTime = new Date(new Date().toLocaleDateString() + " " + alertTimeString);
        }                                                              //
        if (isNaN(alertTime.getTime())) {                              // 676
          handleErrorBound(new Meteor.Error(400, "Couldn't parse alert time, please be more precise."));
        } else {                                                       //
          Meteor.call("setSetting", token, "adminAlertTime", alertTime, handleErrorBound);
        }                                                              //
      } else {                                                         //
        Meteor.call("setSetting", token, "adminAlertTime", null, handleErrorBound);
      }                                                                //
      Meteor.call("setSetting", token, "adminAlertUrl", event.target.alertUrl.value, handleErrorBound);
      return false;                                                    // 686
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.adminAdvanced.helpers({                                     // 690
    setDocumentTitle: function () {                                    // 691
      document.title = "Advanced · Admin · " + globalDb.getServerTitle();
    },                                                                 //
    serverTitle: function () {                                         // 694
      var setting = Settings.findOne({ _id: "serverTitle" });          // 695
      return setting && setting.value || "";                           // 696
    },                                                                 //
    returnAddress: function () {                                       // 698
      var setting = Settings.findOne({ _id: "returnAddress" });        // 699
      return setting && setting.value || "";                           // 700
    },                                                                 //
    splashUrl: function () {                                           // 702
      var setting = Settings.findOne({ _id: "splashUrl" });            // 703
      return setting && setting.value || "";                           // 704
    },                                                                 //
    origin: function () {                                              // 706
      return getOrigin();                                              // 706
    },                                                                 //
    signupDialog: function () {                                        // 707
      var setting = Settings.findOne({ _id: "signupDialog" });         // 708
      return setting && setting.value || DEFAULT_SIGNUP_DIALOG;        // 709
    },                                                                 //
    termsUrl: function () {                                            // 711
      var setting = Settings.findOne({ _id: "termsUrl" });             // 712
      return setting && setting.value;                                 // 713
    },                                                                 //
    privacyUrl: function () {                                          // 715
      var setting = Settings.findOne({ _id: "privacyUrl" });           // 716
      return setting && setting.value;                                 // 717
    },                                                                 //
    adminAlert: function () {                                          // 719
      var setting = Settings.findOne({ _id: "adminAlert" });           // 720
      return setting && setting.value;                                 // 721
    },                                                                 //
    alertTime: function () {                                           // 723
      var setting = Settings.findOne({ _id: "adminAlertTime" });       // 724
      if (setting && setting.value) {                                  // 725
        return setting.value.toLocaleDateString() + " " + setting.value.toLocaleTimeString();
      } else {                                                         //
        return "";                                                     // 728
      }                                                                //
    },                                                                 //
    alertUrl: function () {                                            // 731
      var setting = Settings.findOne({ _id: "adminAlertUrl" });        // 732
      return setting && setting.value;                                 // 733
    },                                                                 //
    appMarketUrl: function () {                                        // 735
      var setting = Settings.findOne({ _id: "appMarketUrl" });         // 736
      return setting && setting.value;                                 // 737
    },                                                                 //
    appIndexUrl: function () {                                         // 739
      var setting = Settings.findOne({ _id: "appIndexUrl" });          // 740
      return setting && setting.value;                                 // 741
    },                                                                 //
    appUpdatesEnabled: function () {                                   // 743
      var setting = Settings.findOne({ _id: "appUpdatesEnabled" });    // 744
      return setting && setting.value;                                 // 745
    }                                                                  //
  });                                                                  //
}                                                                      //
                                                                       //
if (Meteor.isServer) {                                                 // 750
  var Fs = Npm.require("fs");                                          // 751
  var SANDSTORM_ADMIN_TOKEN = SANDSTORM_VARDIR + "/adminToken";        // 752
                                                                       //
  var getSmtpUrl = function () {                                       // 754
    var setting = Settings.findOne({ _id: "smtpUrl" });                // 755
    if (setting) {                                                     // 756
      return setting.value;                                            // 757
    } else {                                                           //
      return process.env.MAIL_URL;                                     // 759
    }                                                                  //
  };                                                                   //
                                                                       //
  var tokenIsValid = function (token) {                                // 763
    if (token && Fs.existsSync(SANDSTORM_ADMIN_TOKEN)) {               // 764
      var stats = Fs.statSync(SANDSTORM_ADMIN_TOKEN);                  // 765
      var expireTime = new Date(Date.now() - ADMIN_TOKEN_EXPIRATION_TIME);
      if (stats.mtime < expireTime) {                                  // 767
        return false;                                                  // 768
      } else {                                                         //
        return Fs.readFileSync(SANDSTORM_ADMIN_TOKEN, { encoding: "utf8" }) === token;
      }                                                                //
    } else {                                                           //
      return false;                                                    // 773
    }                                                                  //
  };                                                                   //
                                                                       //
  var checkAuth = function (token) {                                   // 777
    check(token, Match.OneOf(undefined, null, String));                // 778
    if (!isAdmin() && !tokenIsValid(token)) {                          // 779
      throw new Meteor.Error(403, "User must be admin or provide a valid token");
    }                                                                  //
  };                                                                   //
  function clearAdminToken(token) {                                    // 783
    if (tokenIsValid(token)) {                                         // 784
      Fs.unlinkSync(SANDSTORM_ADMIN_TOKEN);                            // 785
      console.log("Admin token deleted.");                             // 786
    }                                                                  //
  }                                                                    //
  Meteor.methods({                                                     // 789
    setAccountSetting: function (token, serviceName, value) {          // 790
      checkAuth(token);                                                // 791
      check(serviceName, String);                                      // 792
      check(value, Boolean);                                           // 793
                                                                       //
      // TODO(someday): currently this relies on the fact that an account is tied to a single
      // identity, and thus has only that entry in "services". This will need to be looked at when
      // multiple login methods/identities are allowed for a single account.
      if (!value && !tokenIsValid(token) && serviceName in Meteor.user().services) {
        throw new Meteor.Error(403, "You can not disable the login service that your account uses.");
      }                                                                //
                                                                       //
      // Only check configurations for OAuth services.                 //
      var oauthServices = ["google", "github"];                        // 804
      if (value && oauthServices.indexOf(serviceName) != -1) {         // 805
        var ServiceConfiguration = Package["service-configuration"].ServiceConfiguration;
        var config = ServiceConfiguration.configurations.findOne({ service: serviceName });
        if (!config) {                                                 // 808
          throw new Meteor.Error(403, "You must configure the " + serviceName + " service before you can enable it. Click the \"configure\" link.");
        }                                                              //
        if (!config.clientId || !config.secret) {                      // 812
          throw new Meteor.Error(403, "You must provide a non-empty clientId and secret for the " + serviceName + " service before you can enable it. Click the \"configure\" link.");
        }                                                              //
      }                                                                //
      Settings.upsert({ _id: serviceName }, { $set: { value: value } });
      if (value) {                                                     // 818
        Settings.update({ _id: serviceName }, { $unset: { automaticallyReset: 1 } });
      }                                                                //
    },                                                                 //
    setSetting: function (token, name, value) {                        // 822
      checkAuth(token);                                                // 823
      check(name, String);                                             // 824
      check(value, Match.OneOf(null, String, Date, Boolean));          // 825
                                                                       //
      Settings.upsert({ _id: name }, { $set: { value: value } });      // 827
    },                                                                 //
    getSmtpUrl: function (token) {                                     // 829
      checkAuth(token);                                                // 830
                                                                       //
      return getSmtpUrl();                                             // 832
    },                                                                 //
    "adminConfigureLoginService": function (token, options) {          // 834
      checkAuth(token);                                                // 835
      check(options, Match.ObjectIncluding({ service: String }));      // 836
                                                                       //
      var ServiceConfiguration = Package["service-configuration"].ServiceConfiguration;
                                                                       //
      ServiceConfiguration.configurations.upsert({ service: options.service }, options);
    },                                                                 //
    clearResumeTokensForService: function (token, serviceName) {       // 842
      checkAuth(token);                                                // 843
      check(serviceName, String);                                      // 844
                                                                       //
      var query = {};                                                  // 846
      query["services." + serviceName] = { $exists: true };            // 847
      Meteor.users.find(query).forEach(function (identity) {           // 848
        if (identity.services.resume && identity.services.resume.loginTokens && identity.services.resume.loginTokens.length > 0) {
          Meteor.users.update({ _id: identity._id }, { $set: { "services.resume.loginTokens": [] } });
        }                                                              //
        Meteor.users.update({ "loginIdentities.id": identity._id }, { $set: { "services.resume.loginTokens": [] } });
      });                                                              //
    },                                                                 //
    adminUpdateUser: function (token, userInfo) {                      // 857
      checkAuth(token);                                                // 858
      check(userInfo, {                                                // 859
        userId: String,                                                // 860
        signupKey: Boolean,                                            // 861
        isAdmin: Boolean                                               // 862
      });                                                              //
                                                                       //
      var userId = userInfo.userId;                                    // 865
      if (userId === Meteor.userId() && !userInfo.isAdmin) {           // 866
        throw new Meteor.Error(403, "User cannot remove admin permissions from itself.");
      }                                                                //
                                                                       //
      Meteor.users.update({ _id: userId }, { $set: _.omit(userInfo, ["_id", "userId"]) });
    },                                                                 //
    testSend: function (token, smtpUrl, to) {                          // 872
      checkAuth(token);                                                // 873
      check(smtpUrl, String);                                          // 874
      check(to, String);                                               // 875
                                                                       //
      SandstormEmail.send({                                            // 877
        to: to,                                                        // 878
        from: globalDb.getServerTitle() + " <" + globalDb.getReturnAddress() + ">",
        subject: "Testing your Sandstorm's SMTP setting",              // 880
        text: "Success! Your outgoing SMTP is working.",               // 881
        smtpUrl: smtpUrl                                               // 882
      });                                                              //
    },                                                                 //
    createSignupKey: function (token, note, quota) {                   // 885
      checkAuth(token);                                                // 886
      check(note, String);                                             // 887
      check(quota, Match.OneOf(undefined, null, Number));              // 888
                                                                       //
      var key = Random.id();                                           // 890
      var content = { _id: key, used: false, note: note };             // 891
      if (typeof quota === "number") content.quota = quota;            // 892
      SignupKeys.insert(content);                                      // 893
      return key;                                                      // 894
    },                                                                 //
    sendInvites: function (token, origin, from, list, subject, message, quota) {
      checkAuth(token);                                                // 897
      check([origin, from, list, subject, message], [String]);         // 898
      check(quota, Match.OneOf(undefined, null, Number));              // 899
                                                                       //
      if (!from.trim()) {                                              // 901
        throw new Meteor.Error(403, "Must enter 'from' address.");     // 902
      }                                                                //
                                                                       //
      if (!list.trim()) {                                              // 905
        throw new Meteor.Error(403, "Must enter 'to' addresses.");     // 906
      }                                                                //
                                                                       //
      this.unblock();                                                  // 909
                                                                       //
      list = list.split("\n");                                         // 911
      for (var i in babelHelpers.sanitizeForInObject(list)) {          // 912
        var email = list[i].trim();                                    // 913
                                                                       //
        if (email) {                                                   // 915
          var key = Random.id();                                       // 916
                                                                       //
          var content = { _id: key, used: false, note: "E-mail invite to " + email,
            email: email, definitelySent: false };                     // 919
          if (typeof quota === "number") content.quota = quota;        // 920
          SignupKeys.insert(content);                                  // 921
          SandstormEmail.send({                                        // 922
            to: email,                                                 // 923
            from: from,                                                // 924
            subject: subject,                                          // 925
            text: message.replace(/\$KEY/g, origin + Router.routes.signup.path({ key: key }))
          });                                                          //
          SignupKeys.update(key, { $set: { definitelySent: true } });  // 928
        }                                                              //
      }                                                                //
                                                                       //
      return { sent: true };                                           // 932
    },                                                                 //
    offerIpNetwork: function (token) {                                 // 934
      checkAuth(token);                                                // 935
      if (!isAdmin()) {                                                // 936
        throw new Meteor.Error(403, "Offering IpNetwork is only allowed for logged in users " + "(a token is not sufficient). Please sign in with an admin account");
      }                                                                //
                                                                       //
      var requirements = [{                                            // 941
        userIsAdmin: Meteor.userId()                                   // 942
      }];                                                              //
      var sturdyRef = waitPromise(saveFrontendRef({ ipNetwork: true }, { webkey: null }, requirements)).sturdyRef;
      return ROOT_URL.protocol + "//" + makeWildcardHost("api") + "#" + sturdyRef;
    },                                                                 //
    offerIpInterface: function (token) {                               // 948
      checkAuth(token);                                                // 949
      if (!isAdmin()) {                                                // 950
        throw new Meteor.Error(403, "Offering IpInterface is only allowed for logged in users " + "(a token is not sufficient). Please sign in with an admin account");
      }                                                                //
                                                                       //
      var requirements = [{                                            // 955
        userIsAdmin: Meteor.userId()                                   // 956
      }];                                                              //
      var sturdyRef = waitPromise(saveFrontendRef({ ipInterface: true }, { webkey: null }, requirements)).sturdyRef;
      return ROOT_URL.protocol + "//" + makeWildcardHost("api") + "#" + sturdyRef;
    },                                                                 //
    adminToggleDisableCap: function (token, capId, value) {            // 962
      checkAuth(token);                                                // 963
      check(capId, String);                                            // 964
      check(value, Boolean);                                           // 965
                                                                       //
      if (value) {                                                     // 967
        ApiTokens.update({ _id: capId }, { $set: { revoked: true } });
      } else {                                                         //
        ApiTokens.update({ _id: capId }, { $set: { revoked: false } });
      }                                                                //
    },                                                                 //
    updateQuotas: function (token, list, quota) {                      // 973
      checkAuth(token);                                                // 974
      check(list, String);                                             // 975
      check(quota, Match.OneOf(undefined, null, Number));              // 976
                                                                       //
      if (!list.trim()) {                                              // 978
        throw new Meteor.Error(400, "Must enter addresses.");          // 979
      }                                                                //
                                                                       //
      list = list.split("\n");                                         // 982
      var invalid = [];                                                // 983
      for (var i in babelHelpers.sanitizeForInObject(list)) {          // 984
        var modifier = typeof quota === "number" ? { $set: { quota: quota } } : { $unset: { quota: "" } };
        var n = SignupKeys.update({ email: list[i] }, modifier, { multi: true });
        n += Meteor.users.update({ signupEmail: list[i] }, modifier, { multi: true });
                                                                       //
        if (n < 1) invalid.push(list[i]);                              // 990
      }                                                                //
                                                                       //
      if (invalid.length > 0) {                                        // 993
        throw new Meteor.Error(404, "These addresses did not map to any user nor invite: " + invalid.join(", "));
      }                                                                //
    },                                                                 //
    dismissAdminStatsNotifications: function (token) {                 // 998
      checkAuth(token);                                                // 999
      globalDb.collections.notifications.remove({ "admin.type": "reportStats" });
    },                                                                 //
    signUpAsAdmin: function (token) {                                  // 1002
      check(token, String);                                            // 1003
      checkAuth(token);                                                // 1004
      if (!this.userId) {                                              // 1005
        throw new Meteor.Error(403, "Must be logged in to sign up as admin.");
      }                                                                //
      if (!Meteor.user().loginIdentities) {                            // 1008
        throw new Meteor.Error(403, "Must be logged into an account to sign up as admin.");
      }                                                                //
      Meteor.users.update({ _id: this.userId }, { $set: { isAdmin: true, signupKey: "admin" } });
      clearAdminToken(token);                                          // 1012
    }                                                                  //
  });                                                                  //
                                                                       //
  var authorizedAsAdmin = function (token, userId) {                   // 1016
    return Match.test(token, Match.OneOf(undefined, null, String)) && (userId && isAdminById(userId) || tokenIsValid(token));
  };                                                                   //
  Meteor.publish("admin", function (token) {                           // 1020
    if (!authorizedAsAdmin(token, this.userId)) return [];             // 1021
    return Settings.find();                                            // 1022
  });                                                                  //
                                                                       //
  Meteor.publish("adminServiceConfiguration", function (token) {       // 1025
    if (!authorizedAsAdmin(token, this.userId)) return [];             // 1026
    return Package['service-configuration'].ServiceConfiguration.configurations.find();
  });                                                                  //
                                                                       //
  Meteor.publish("publicAdminSettings", function () {                  // 1030
    return Settings.find({ _id: { $in: publicAdminSettings } });       // 1031
  });                                                                  //
                                                                       //
  Meteor.publish("adminToken", function (token) {                      // 1034
    check(token, String);                                              // 1035
    this.added("adminToken", "adminToken", { tokenIsValid: tokenIsValid(token) });
    this.ready();                                                      // 1037
  });                                                                  //
                                                                       //
  Meteor.publish("allUsers", function (token) {                        // 1040
    if (!authorizedAsAdmin(token, this.userId)) return [];             // 1041
    return Meteor.users.find();                                        // 1042
  });                                                                  //
  Meteor.publish("activityStats", function (token) {                   // 1044
    if (!authorizedAsAdmin(token, this.userId)) return [];             // 1045
    return ActivityStats.find();                                       // 1046
  });                                                                  //
                                                                       //
  Meteor.publish("statsTokens", function (token) {                     // 1049
    if (!authorizedAsAdmin(token, this.userId)) return [];             // 1050
    return StatsTokens.find();                                         // 1051
  });                                                                  //
                                                                       //
  Meteor.publish("allPackages", function (token) {                     // 1054
    if (!authorizedAsAdmin(token, this.userId)) return [];             // 1055
    return Packages.find({ manifest: { $exists: true } }, { fields: { appId: 1, "manifest.appVersion": 1,
        "manifest.actions": 1, "manifest.appTitle": 1 } });            // 1058
  });                                                                  //
                                                                       //
  Meteor.publish("realTimeStats", function (token) {                   // 1061
    if (!authorizedAsAdmin(token, this.userId)) return [];             // 1062
                                                                       //
    // Last five minutes.                                              //
    this.added("realTimeStats", "now", computeStats(new Date(Date.now() - 5 * 60 * 1000)));
                                                                       //
    // Since last sample.                                              //
    var lastSample = ActivityStats.findOne({}, { sort: { timestamp: -1 } });
    var lastSampleTime = lastSample ? lastSample.timestamp : new Date(0);
    this.added("realTimeStats", "today", computeStats(lastSampleTime));
                                                                       //
    // TODO(someday): Update every few minutes?                        //
                                                                       //
    this.ready();                                                      // 1074
  });                                                                  //
  Meteor.publish("adminLog", function (token) {                        // 1076
    if (!authorizedAsAdmin(token, this.userId)) return [];             // 1077
                                                                       //
    var logfile = SANDSTORM_LOGDIR + "/sandstorm.log";                 // 1079
                                                                       //
    var fd = Fs.openSync(logfile, "r");                                // 1081
    var startSize = Fs.fstatSync(fd).size;                             // 1082
                                                                       //
    // Start tailing at EOF - 8k.                                      //
    var offset = Math.max(0, startSize - 8192);                        // 1085
                                                                       //
    var self = this;                                                   // 1087
    function doTail() {                                                // 1088
      for (;;) {                                                       // 1089
        var buf = new Buffer(Math.max(1024, startSize - offset));      // 1090
        var n = Fs.readSync(fd, buf, 0, buf.length, offset);           // 1091
        if (n <= 0) break;                                             // 1092
        self.added("adminLog", offset, { text: buf.toString("utf8", 0, n) });
        offset += n;                                                   // 1094
      }                                                                //
    }                                                                  //
                                                                       //
    // Watch the file for changes.                                     //
    var watcher = Fs.watch(logfile, { persistent: false }, Meteor.bindEnvironment(doTail));
                                                                       //
    // When the subscription stops, stop watching the file.            //
    this.onStop(function () {                                          // 1102
      watcher.close();                                                 // 1103
      Fs.closeSync(fd);                                                // 1104
    });                                                                //
                                                                       //
    // Read initial 8k tail data immediately.                          //
    doTail();                                                          // 1108
                                                                       //
    // Notify ready.                                                   //
    this.ready();                                                      // 1111
  });                                                                  //
  Meteor.publish("adminApiTokens", function (token) {                  // 1113
    if (!authorizedAsAdmin(token, this.userId)) return [];             // 1114
    return ApiTokens.find({ $or: [{ "frontendRef.ipNetwork": { $exists: true } }, { "frontendRef.ipInterface": { $exists: true } }] }, { fields: { frontendRef: 1, created: 1, requirements: 1, revoked: 1 } });
  });                                                                  //
}                                                                      //
                                                                       //
function serviceEnabled(name) {                                        // 1121
  var setting = Settings.findOne({ _id: name });                       // 1122
  return setting && !!setting.value;                                   // 1123
}                                                                      //
                                                                       //
if (Meteor.server) {                                                   // 1126
  function observeOauthService(name) {                                 // 1127
    Settings.find({ _id: name, value: true }).observe({                // 1128
      added: function () {                                             // 1129
        // Tell the oauth library it should accept login attempts from this service.
        Accounts.oauth.registerService(name);                          // 1131
      },                                                               //
      removed: function () {                                           // 1133
        // Tell the oauth library it should deny login attempts from this service.
        Accounts.oauth.unregisterService(name);                        // 1135
      }                                                                //
    });                                                                //
  }                                                                    //
  observeOauthService("github");                                       // 1139
  observeOauthService("google");                                       // 1140
}                                                                      //
                                                                       //
Accounts.identityServices.github = {                                   // 1143
  isEnabled: function () {                                             // 1144
    return serviceEnabled("github");                                   // 1145
  },                                                                   //
  loginTemplate: {                                                     // 1147
    name: "oauthLoginButton",                                          // 1148
    priority: 1,                                                       // 1149
    data: {                                                            // 1150
      method: "loginWithGithub",                                       // 1151
      name: "github",                                                  // 1152
      displayName: "GitHub"                                            // 1153
    }                                                                  //
  }                                                                    //
};                                                                     //
                                                                       //
Accounts.identityServices.google = {                                   // 1158
  isEnabled: function () {                                             // 1159
    return serviceEnabled("google");                                   // 1160
  },                                                                   //
  loginTemplate: {                                                     // 1162
    name: "oauthLoginButton",                                          // 1163
    priority: 2,                                                       // 1164
    data: {                                                            // 1165
      method: "loginWithGoogle",                                       // 1166
      name: "google",                                                  // 1167
      displayName: "Google"                                            // 1168
    }                                                                  //
  }                                                                    //
};                                                                     //
                                                                       //
Accounts.identityServices.email = {                                    // 1173
  isEnabled: function () {                                             // 1174
    return serviceEnabled("emailToken");                               // 1175
  },                                                                   //
  loginTemplate: {                                                     // 1177
    name: "emailLoginForm",                                            // 1178
    priority: 10 }                                                     // 1179
};                                                                     //
// Put it at the bottom of the list.                                   //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=admin.js.map
