(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// shared/grain.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
// This file implements /grain, i.e. the main view into an app.        //
                                                                       //
var DEFAULT_TITLE = "Sandstorm";                                       // 19
                                                                       //
if (Meteor.isServer) {                                                 // 21
  var Crypto = Npm.require("crypto");                                  // 22
                                                                       //
  Meteor.publish("grainTopBar", function (grainId) {                   // 24
    check(grainId, String);                                            // 25
    var self = this;                                                   // 26
                                                                       //
    var result = [Grains.find({ _id: grainId, $or: [{ userId: this.userId }, { "private": { $ne: true } }] }, { fields: { title: 1, userId: 1, identityId: 1, "private": 1 } })];
    if (this.userId) {                                                 // 30
      var myIdentityIds = SandstormDb.getUserIdentityIds(globalDb.getUser(this.userId));
      result.push(ApiTokens.find({ grainId: grainId,                   // 32
        $or: [{ "owner.user.identityId": { $in: myIdentityIds } }, { identityId: { $in: myIdentityIds } }] }));
    }                                                                  //
    return result;                                                     // 36
  });                                                                  //
                                                                       //
  // We allow users to learn package information about a grain they own.
  // This is used for obtaining icon and app title information for grains
  // you own, which is used in the sidebar. It is not a security/privacy
  // risk since it only exposes this information for grains the user owns.
  Meteor.publish("packageByGrainId", function (grainId) {              // 43
    check(grainId, String);                                            // 44
    var publishThis = [];                                              // 45
    // We need to publish the packageId so that client-side code can   //
    // find the right package.                                         //
    var thisGrainCursor = Grains.find({ _id: grainId, userId: this.userId }, { fields: { packageId: 1 } });
    publishThis.push(thisGrainCursor);                                 // 50
                                                                       //
    if (thisGrainCursor.count()) {                                     // 52
      var thisGrain = thisGrainCursor.fetch()[0];                      // 53
      var thisPackageCursor = Packages.find({ _id: thisGrain.packageId });
      publishThis.push(thisPackageCursor);                             // 55
    }                                                                  //
                                                                       //
    return publishThis;                                                // 58
  });                                                                  //
                                                                       //
  Meteor.publish("tokenInfo", function (token) {                       // 61
    // Allows the client side to map a raw token to its entry in ApiTokens, and the additional
    // metadata that it will need to display the app icon and title.  We do not care about making
    // the metadata reactive.                                          //
    check(token, String);                                              // 65
                                                                       //
    var hashedToken = Crypto.createHash("sha256").update(token).digest("base64");
    var apiToken = ApiTokens.findOne({ _id: hashedToken }, { fields: { grainId: 1, owner: 1 } });
    if (!apiToken) {                                                   // 70
      this.added("tokenInfo", token, { invalidToken: true });          // 71
    } else {                                                           //
      var grainId = apiToken.grainId;                                  // 73
      var grain = Grains.findOne({ _id: grainId }, { fields: { packageId: 1, appId: 1 } });
      if (!grain) {                                                    // 75
        this.added("tokenInfo", token, { invalidToken: true });        // 76
      } else {                                                         //
        if (apiToken.owner && apiToken.owner.user) {                   // 78
          var identity = globalDb.getIdentity(apiToken.owner.user.identityId);
          var metadata = apiToken.owner.user.denormalizedGrainMetadata;
          if (identity && metadata) {                                  // 81
            this.added("tokenInfo", token, { identityOwner: _.pick(identity, "_id", "profile"),
              grainId: grainId,                                        // 84
              grainMetadata: metadata });                              // 85
          } else {                                                     //
            this.added("tokenInfo", token, { invalidToken: true });    // 87
          }                                                            //
        } else if (!apiToken.owner || "webkey" in apiToken.owner) {    //
          var pkg = Packages.findOne({ _id: grain.packageId }, { fields: { manifest: 1 } });
          var appTitle = pkg && pkg.manifest && pkg.manifest.appTitle || { defaultText: "" };
          var appIcon = undefined;                                     // 92
          if (pkg && pkg.manifest && pkg.manifest.metadata && pkg.manifest.metadata.icons) {
            var icons = pkg.manifest.metadata.icons;                   // 94
            appIcon = icons.grain || icons.appGrid;                    // 95
          }                                                            //
          var denormalizedGrainMetadata = {                            // 97
            appTitle: appTitle,                                        // 98
            icon: appIcon,                                             // 99
            appId: appIcon ? undefined : grain.appId                   // 100
          };                                                           //
          this.added("tokenInfo", token, { webkey: true,               // 102
            grainId: grainId,                                          // 104
            grainMetadata: denormalizedGrainMetadata });               // 105
        } else {                                                       //
          this.added("tokenInfo", token, { invalidToken: true });      // 107
        }                                                              //
      }                                                                //
    }                                                                  //
    this.ready();                                                      // 111
    return;                                                            // 112
  });                                                                  //
                                                                       //
  Meteor.publish("grainSize", function (grainId) {                     // 115
    // Publish pseudo-collection containing the size of the grain opened in the given session.
    check(grainId, String);                                            // 117
                                                                       //
    var grain = Grains.findOne(grainId);                               // 119
    if (!grain || grain.userId !== this.userId) {                      // 120
      return [];                                                       // 121
    }                                                                  //
                                                                       //
    var supervisor = globalBackend.cap().getGrain(this.userId, grainId).supervisor;
                                                                       //
    var self = this;                                                   // 126
    var stopped = false;                                               // 127
    var promise = getGrainSize(supervisor);                            // 128
                                                                       //
    function getNext(oldSize) {                                        // 130
      promise = getGrainSize(supervisor, oldSize);                     // 131
      promise.then(function (size) {                                   // 132
        if (!stopped) {                                                // 133
          if (size !== oldSize) {                                      // 134
            // sometimes there are false alarms                        //
            self.changed("grainSizes", grainId, { size: size });       // 135
          }                                                            //
          getNext(size);                                               // 137
        }                                                              //
      }, function (err) {                                              //
        if (!stopped) {                                                // 140
          if (err.kjType === "disconnected") {                         // 141
            self.stop();                                               // 142
          } else {                                                     //
            self.error(err);                                           // 144
          }                                                            //
        }                                                              //
      });                                                              //
    }                                                                  //
                                                                       //
    promise.then(function (size) {                                     // 150
      if (!stopped) {                                                  // 151
        self.added("grainSizes", grainId, { size: size });             // 152
        self.ready();                                                  // 153
        getNext(size);                                                 // 154
      }                                                                //
    }, function (err) {                                                //
      if (!stopped) {                                                  // 157
        if (err.kjType === "disconnected") {                           // 158
          self.stop();                                                 // 159
        } else {                                                       //
          self.error(err);                                             // 161
        }                                                              //
      }                                                                //
    });                                                                //
                                                                       //
    self.onStop(function () {                                          // 166
      stopped = true;                                                  // 167
      promise.cancel();                                                // 168
    });                                                                //
  });                                                                  //
}                                                                      //
                                                                       //
// GrainSizes is used by grainview.js                                  //
GrainSizes = new Mongo.Collection("grainSizes");                       // 174
// TokenInfo is used by grainview.js                                   //
TokenInfo = new Mongo.Collection("tokenInfo");                         // 176
// Pseudo-collections published above.                                 //
                                                                       //
Meteor.methods({                                                       // 179
  deleteGrain: function (grainId) {                                    // 180
    check(grainId, String);                                            // 181
                                                                       //
    if (this.userId) {                                                 // 183
      var grain = Grains.findOne({ _id: grainId, userId: this.userId });
      if (grain) {                                                     // 185
        Grains.remove(grainId);                                        // 186
        ApiTokens.remove({ grainId: grainId, $or: [{ owner: { $exists: false } }, { owner: { webkey: null } }] });
        if (grain.lastUsed) {                                          // 189
          DeleteStats.insert({ type: isDemoUser() ? "demoGrain" : "grain",
            lastActive: grain.lastUsed, appId: grain.appId });         // 191
        }                                                              //
        if (!this.isSimulation) {                                      // 193
          waitPromise(globalBackend.deleteGrain(grainId, this.userId));
          Meteor.call("deleteUnusedPackages", grain.appId);            // 195
        }                                                              //
      }                                                                //
    }                                                                  //
  },                                                                   //
  forgetGrain: function (grainId, identityId) {                        // 200
    // TODO(cleanup): For now we are ignoring `identityId`, but maybe we should expose finer-grained
    //  forgetting.                                                    //
                                                                       //
    check(grainId, String);                                            // 204
    check(identityId, String);                                         // 205
                                                                       //
    if (!this.userId) {                                                // 207
      throw new Meteor.Error(403, "Must be logged in to forget a grain.");
    }                                                                  //
    SandstormDb.getUserIdentityIds(Meteor.user()).forEach(function (identityId) {
      ApiTokens.remove({ grainId: grainId, "owner.user.identityId": identityId });
    });                                                                //
  },                                                                   //
  updateGrainPreferredIdentity: function (grainId, identityId) {       // 214
    check(grainId, String);                                            // 215
    check(identityId, String);                                         // 216
    if (!this.userId) {                                                // 217
      throw new Meteor.Error(403, "Must be logged in.");               // 218
    }                                                                  //
    var grain = globalDb.getGrain(grainId) || {};                      // 220
    if (!grain.userId === this.userId) {                               // 221
      throw new Meteor.Error(403, "Grain not owned by current user.");
    }                                                                  //
    Grains.update({ _id: grainId }, { $set: { identityId: identityId } });
  },                                                                   //
  updateGrainTitle: function (grainId, newTitle, identityId) {         // 226
    check(grainId, String);                                            // 227
    check(newTitle, String);                                           // 228
    check(identityId, String);                                         // 229
    if (this.userId) {                                                 // 230
      var grain = Grains.findOne(grainId);                             // 231
      if (grain) {                                                     // 232
        if (grain.userId === this.userId) {                            // 233
          Grains.update({ _id: grainId, userId: this.userId }, { $set: { title: newTitle } });
        } else {                                                       //
          if (!globalDb.userHasIdentity(this.userId, identityId)) {    // 236
            throw new Meteor.Error(403, "Current user does not have identity " + identityId);
          }                                                            //
          var token = ApiTokens.findOne({ grainId: grainId, objectId: { $exists: false },
            "owner.user.identityId": identityId }, { sort: { created: 1 } });
          if (token) {                                                 // 242
            ApiTokens.update(token._id, { $set: { "owner.user.title": newTitle } });
          }                                                            //
        }                                                              //
      }                                                                //
    }                                                                  //
  },                                                                   //
  privatizeGrain: function (grainId) {                                 // 249
    check(grainId, String);                                            // 250
    if (this.userId) {                                                 // 251
      Grains.update({ _id: grainId, userId: this.userId }, { $set: { "private": true } });
    }                                                                  //
  },                                                                   //
  inviteUsersToGrain: function (origin, identityId, grainId, title, roleAssignment, contacts, message) {
    if (!this.isSimulation) {                                          // 257
      check(origin, String);                                           // 258
      check(identityId, String);                                       // 259
      check(grainId, String);                                          // 260
      check(title, String);                                            // 261
      check(roleAssignment, roleAssignmentPattern);                    // 262
      check(contacts, [{                                               // 263
        _id: String,                                                   // 264
        isDefault: Match.Optional(Boolean),                            // 265
        profile: Match.ObjectIncluding({                               // 266
          service: String,                                             // 267
          name: String,                                                // 268
          intrinsicName: String                                        // 269
        }) }]);                                                        //
      check(message, { text: String, html: String });                  // 271
      if (!this.userId) {                                              // 272
        throw new Meteor.Error(403, "Must be logged in to share by email.");
      }                                                                //
      if (!globalDb.userHasIdentity(this.userId, identityId)) {        // 275
        throw new Meteor.Error(403, "Not an identity of the current user: " + identityId);
      }                                                                //
      if (contacts.length === 0) {                                     // 278
        throw new Meteor.Error(400, "No contacts were provided.");     // 279
      }                                                                //
      var accountId = this.userId;                                     // 281
      var identity = globalDb.getIdentity(identityId);                 // 282
      var sharerDisplayName = identity.profile.name;                   // 283
      var outerResult = { successes: [], failures: [] };               // 284
      contacts.forEach(function (contact) {                            // 285
        if (contact.isDefault && contact.profile.service === "email") {
          var emailAddress = contact.profile.intrinsicName;            // 287
          var result = SandstormPermissions.createNewApiToken(globalDb, { identityId: identityId, accountId: accountId }, grainId, "email invitation for " + emailAddress, roleAssignment, { webkey: { forSharing: true } });
          var url = origin + "/shared/" + result.token;                // 292
          var html = message.html + "<br><br>" + "<a href='" + url + "' style='display:inline-block;text-decoration:none;" + "font-family:sans-serif;width:200px;min-height:30px;line-height:30px;" + "border-radius:4px;text-align:center;background:#428bca;color:white'>" + "Open Shared Grain</a><div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: If you forward this email to other people, they will be able to access " + "the share as well. To prevent this, remove the button before forwarding.</div>";
          try {                                                        // 300
            SandstormEmail.send({                                      // 301
              to: emailAddress,                                        // 302
              from: "Sandstorm server <no-reply@" + HOSTNAME + ">",    // 303
              subject: sharerDisplayName + " has invited you to join a grain: " + title,
              text: message.text + "\n\nFollow this link to open the shared grain:\n\n" + url + "\n\nNote: If you forward this email to other people, they will be able to " + "access the share as well. To prevent this, remove the link before forwarding.",
              html: html                                               // 308
            });                                                        //
          } catch (e) {                                                //
            outerResult.failures.push({ contact: contact, error: e.toString() });
          }                                                            //
        } else {                                                       //
          var _result = SandstormPermissions.createNewApiToken(globalDb, { identityId: identityId, accountId: accountId }, grainId, "direct invitation to " + contact.profile.intrinsicName, roleAssignment, { user: { identityId: contact._id, title: title } });
          try {                                                        // 318
            var identity = Meteor.users.findOne({ _id: contact._id });
            var emailAddress = SandstormDb.getVerifiedEmails(identity)[0];
            var url = origin + "/shared/" + _result.token;             // 321
            if (emailAddress) {                                        // 322
              var intrinsicName = contact.profile.intrinsicName;       // 323
              var loginNote;                                           // 324
              if (contact.profile.service === "google") {              // 325
                loginNote = "Google account with address " + emailAddress;
              } else if (contact.profile.service === "github") {       //
                loginNote = "Github account with username " + intrinsicName;
              } else if (contact.profile.service === "email") {        //
                loginNote = "email address " + intrinsicName;          // 330
              } else {                                                 //
                throw new Meteor.Error(500, "Unknown service to email share link.");
              }                                                        //
              var html = message.html + "<br><br>" + "<a href='" + url + "' style='display:inline-block;text-decoration:none;" + "font-family:sans-serif;width:200px;min-height:30px;line-height:30px;" + "border-radius:4px;text-align:center;background:#428bca;color:white'>" + "Open Shared Grain</a><div style='font-size:8pt;font-style:italic;color:gray'>" + "Note: You will need to log in with your " + loginNote + " to access this grain.";
              SandstormEmail.send({                                    // 341
                to: emailAddress,                                      // 342
                from: "Sandstorm server <no-reply@" + HOSTNAME + ">",  // 343
                subject: sharerDisplayName + " has invited you to join a grain: " + title,
                text: message.text + "\n\nFollow this link to open the shared grain:\n\n" + url + "\n\nNote: You will need to log in with your " + loginNote + " to access this grain.",
                html: html                                             // 348
              });                                                      //
            } else {                                                   //
              outerResult.failures.push({ contact: contact, warning: "User does not have a " + "verified email, so notification of this share was not sent to them. Please " + "manually share " + url + " with them." });
            }                                                          //
          } catch (e) {                                                //
            outerResult.failures.push({ contact: contact, error: e.toString(),
              warning: "Share succeeded, but there was an error emailing the user. Please " + "manually share " + url + " with them." });
          }                                                            //
        }                                                              //
      });                                                              //
      return outerResult;                                              // 362
    }                                                                  //
  }                                                                    //
});                                                                    //
                                                                       //
if (Meteor.isClient) {                                                 // 367
  Tracker.autorun(function () {                                        // 368
    // We need to keep track of certain data about each grain we can view.
    // TODO(cleanup): Do these in GrainView to avoid spurious resubscribes.
    var grains = globalGrains.get();                                   // 371
    grains.forEach(function (grain) {                                  // 372
      grain.depend();                                                  // 373
      var grainId = grain.grainId();                                   // 374
      if (grainId) {                                                   // 375
        Meteor.subscribe("grainTopBar", grainId);                      // 376
        if (grain.isOwner()) {                                         // 377
          Meteor.subscribe("packageByGrainId", grainId);               // 378
        }                                                              //
        var token = grain.token();                                     // 380
        if (token) {                                                   // 381
          Meteor.subscribe("tokenInfo", token);                        // 382
        }                                                              //
      }                                                                //
    });                                                                //
  });                                                                  //
                                                                       //
  Template.layout.events({                                             // 388
    "click .incognito-button": function (event) {                      // 389
      console.log(event);                                              // 390
      var grains = globalGrains.get();                                 // 391
      var token = event.currentTarget.getAttribute("data-token");      // 392
      if (token) {                                                     // 393
        grains.forEach(function (grain) {                              // 394
          if (grain.token() == token) {                                // 395
            grain.doNotRevealIdentity();                               // 396
          }                                                            //
        });                                                            //
      } else {                                                         //
        console.error("Interstitial prompt answered, but no token present?");
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  var promptNewTitle = function () {                                   // 405
    var grain = getActiveGrain(globalGrains.get());                    // 406
    if (grain) {                                                       // 407
      var prompt = "Set new title:";                                   // 408
      if (!grain.isOwner()) {                                          // 409
        prompt = "Set a new personal title: (does not change the owner's title for this grain)";
      }                                                                //
      var title = window.prompt(prompt, grain.title());                // 412
      if (title) {                                                     // 413
        grain.setTitle(title);                                         // 414
      }                                                                //
    }                                                                  //
  };                                                                   //
  Template.grainTitle.events({                                         // 418
    "click": function (event) {                                        // 419
      promptNewTitle();                                                // 420
    },                                                                 //
    "keydown": function (event) {                                      // 422
      if (event.keyCode === 13 || event.keyCode === 32) {              // 423
        // Allow space or enter to trigger renaming the grain - Firefox doesn't treat enter on the
        // focused element as click().                                 //
        promptNewTitle();                                              // 426
        event.preventDefault();                                        // 427
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainDeleteButton.events({                                  // 432
    "click button": function (event) {                                 // 433
      var grains = globalGrains.get();                                 // 434
      var activeIndex = activeGrainIndex(grains);                      // 435
      var activeGrain = grains[activeIndex];                           // 436
      var newActiveIndex = activeIndex == grains.length - 1 ? activeIndex - 1 : activeIndex;
      if (activeGrain.isOwner()) {                                     // 438
        if (window.confirm("Really delete this grain?")) {             // 439
          Meteor.call("deleteGrain", activeGrain.grainId());           // 440
          // TODO: extract globalGrains into a class that has a "close" method for closing the active view
          activeGrain.destroy();                                       // 442
          if (grains.length == 1) {                                    // 443
            globalGrains.set([]);                                      // 444
            Router.go("root"); // will redirect to the right place     // 445
          } else {                                                     //
              grains.splice(activeIndex, 1);                           // 447
              grains[newActiveIndex].setActive(true);                  // 448
              globalGrains.set(grains);                                // 449
              Router.go(grains[newActiveIndex].route());               // 450
            }                                                          //
        }                                                              //
      } else {                                                         //
        if (window.confirm("Really forget this grain?")) {             // 454
          var identityId = activeGrain.identityId();                   // 455
          if (identityId) {                                            // 456
            Meteor.call("forgetGrain", activeGrain.grainId(), identityId);
          }                                                            //
          // TODO: extract globalGrains into a class that has a "close" method for closing the active view
          activeGrain.destroy();                                       // 460
          if (grains.length == 1) {                                    // 461
            globalGrains.set([]);                                      // 462
            Router.go("root"); // will redirect to the right place     // 463
          } else {                                                     //
              grains.splice(activeIndex, 1);                           // 465
              grains[newActiveIndex].setActive(true);                  // 466
              globalGrains.set(grains);                                // 467
              Router.go(grains[newActiveIndex].route());               // 468
            }                                                          //
        }                                                              //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainDebugLogButton.events({                                // 475
    "click button": function (event) {                                 // 476
      this.reset();                                                    // 477
      var activeGrain = getActiveGrain(globalGrains.get());            // 478
      window.open("/grainlog/" + activeGrain.grainId(), "_blank", "menubar=no,status=no,toolbar=no,width=700,height=700");
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainBackupButton.onCreated(function () {                   // 484
    this.isLoading = new ReactiveVar(false);                           // 485
  });                                                                  //
                                                                       //
  Template.grainBackupButton.helpers({                                 // 488
    isLoading: function () {                                           // 489
      return Template.instance().isLoading.get();                      // 490
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainBackupButton.events({                                  // 494
    "click button": function (event, template) {                       // 495
      template.isLoading.set(true);                                    // 496
      this.reset();                                                    // 497
      var activeGrain = getActiveGrain(globalGrains.get());            // 498
      Meteor.call("backupGrain", activeGrain.grainId(), function (err, id) {
        template.isLoading.set(false);                                 // 500
        if (err) {                                                     // 501
          alert("Backup failed: " + err); // TODO(someday): make this better UI
        } else {                                                       //
            // Firefox for some reason decides to kill all websockets when we try to download the file
            // by navigating there. So we're left doing a dirty hack to get around the popup blocker.
            var isFirefox = typeof InstallTrigger !== "undefined";     // 506
                                                                       //
            if (isFirefox) {                                           // 508
              var save = document.createElement("a");                  // 509
              save.href = "/downloadBackup/" + id;                     // 510
                                                                       //
              save.download = activeGrain.title() + ".zip";            // 512
              var event = document.createEvent("MouseEvents");         // 513
              event.initMouseEvent("click", true, false, window, 0, 0, 0, 0, 0, false, false, false, false, 0, null);
              save.dispatchEvent(event);                               // 518
            } else {                                                   //
              window.location = "/downloadBackup/" + id;               // 520
            }                                                          //
          }                                                            //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainRestartButton.events({                                 // 527
    "click button": function (event) {                                 // 528
      this.reset();                                                    // 529
      var activeGrain = getActiveGrain(globalGrains.get());            // 530
      var grainId = activeGrain.grainId();                             // 531
                                                                       //
      Meteor.call("shutdownGrain", grainId, function (err) {           // 533
        if (err) {                                                     // 534
          alert("Restart failed: " + err); // TODO(someday): make this better UI
        } else {                                                       //
            var frames = document.getElementsByClassName("grain-frame");
            for (var i = 0; i < frames.length; i++) {                  // 538
              var frame = frames[i];                                   // 539
              if (frame.dataset.grainid == grainId) {                  // 540
                frame.src = frame.src;                                 // 541
              }                                                        //
            }                                                          //
          }                                                            //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  function selectElementContents(element) {                            // 549
    if (document.body.createTextRange) {                               // 550
      var range = document.body.createTextRange();                     // 551
      range.moveToElementText(element);                                // 552
      range.select();                                                  // 553
    } else if (window.getSelection) {                                  //
      var selection = window.getSelection();                           // 555
      var range = document.createRange();                              // 556
      range.selectNodeContents(element);                               // 557
      selection.removeAllRanges();                                     // 558
      selection.addRange(range);                                       // 559
    }                                                                  //
  }                                                                    //
                                                                       //
  function selectTargetContents(event) {                               // 563
    event.preventDefault();                                            // 564
    selectElementContents(event.currentTarget);                        // 565
  }                                                                    //
                                                                       //
  Template.grainApiTokenPopup.events({                                 // 568
    "click .copy-me": selectTargetContents,                            // 569
    "focus .copy-me": selectTargetContents,                            // 570
    "submit .newApiToken": function (event) {                          // 571
      event.preventDefault();                                          // 572
      var activeGrain = getActiveGrain(globalGrains.get());            // 573
      var grainId = activeGrain.grainId();                             // 574
      activeGrain.setGeneratedApiToken("pending");                     // 575
      var roleList = document.getElementById("api-token-role");        // 576
      // TODO(cleanup): avoid using global ids; select a child of the current template instead
      var assignment = { allAccess: null };                            // 578
      if (roleList && roleList.selectedIndex > 0) {                    // 579
        assignment = { roleId: roleList.selectedIndex - 1 };           // 580
      }                                                                //
      Meteor.call("newApiToken", { identityId: activeGrain.identityId() }, grainId, document.getElementById("api-token-petname").value, assignment, { webkey: { forSharing: false } }, function (error, result) {
        if (error) {                                                   // 586
          activeGrain.setGeneratedApiToken(undefined);                 // 587
          window.alert("Failed to create token.\n" + error);           // 588
          console.error(error.stack);                                  // 589
        } else {                                                       //
          activeGrain.setGeneratedApiToken(result.endpointUrl + "#" + result.token);
        }                                                              //
      });                                                              //
    },                                                                 //
    "click #resetApiToken": function (event) {                         // 595
      var activeGrain = getActiveGrain(globalGrains.get());            // 596
      activeGrain.setGeneratedApiToken(undefined);                     // 597
    },                                                                 //
    "click button.revoke-token": function (event) {                    // 599
      Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { revoked: true });
    },                                                                 //
                                                                       //
    "click .token-petname": function (event) {                         // 604
      // TODO(soon): Find a less-annoying way to get this input, perhaps by allowing the user
      //   to edit the petname in place.                               //
      var petname = window.prompt("Set new label:", this.petname);     // 607
      if (petname) {                                                   // 608
        Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { petname: petname });
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainSharePopup.events({                                    // 615
    "click .copy-me": selectTargetContents,                            // 616
    "focus .copy-me": selectTargetContents,                            // 617
    "click #share-grain-popup-closer": function (event) {              // 618
      Session.set("show-share-grain", false);                          // 619
    },                                                                 //
    "click button.who-has-access": function (event, instance) {        // 621
      event.preventDefault();                                          // 622
      var closer = globalTopbar.addItem({                              // 623
        name: "who-has-access",                                        // 624
        template: Template.whoHasAccess,                               // 625
        popupTemplate: Template.whoHasAccessPopup,                     // 626
        data: new ReactiveVar(instance.data),                          // 627
        startOpen: true,                                               // 628
        onDismiss: function () {                                       // 629
          return "remove";                                             // 630
        }                                                              //
      });                                                              //
    },                                                                 //
    "click #privatize-grain": function (event) {                       // 634
      Meteor.call("privatizeGrain", getActiveGrain(globalGrains.get()).grainId());
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.shareWithOthers.onRendered(function () {                    // 639
    this.find("[role=tab]").focus();                                   // 640
  });                                                                  //
  var activateTargetTab = function (event, instance) {                 // 642
    // Deactivate all tabs and all tab panels.                         //
    instance.findAll("ul[role=tablist]>li[role=tab]").forEach(function (element) {
      element.setAttribute("aria-selected", false);                    // 645
    });                                                                //
    instance.findAll(".tabpanel").forEach(function (element) {         // 647
      element.setAttribute("aria-hidden", true);                       // 648
    });                                                                //
                                                                       //
    // Activate the tab header the user selected.                      //
    event.currentTarget.setAttribute("aria-selected", true);           // 652
    // Show the corresponding tab panel.                               //
    var idToShow = event.currentTarget.getAttribute("aria-controls");  // 654
    var tabPanelToShow = instance.find("#" + idToShow);                // 655
    tabPanelToShow.setAttribute("aria-hidden", false);                 // 656
  };                                                                   //
  Template.shareWithOthers.events({                                    // 658
    "click #send-invite-tab-header": activateTargetTab,                // 659
    "click #shareable-link-tab-header": activateTargetTab,             // 660
    "keydown [role=tab]": function (event, template) {                 // 661
      if (event.keyCode == 38 || event.keyCode == 40) {                // 662
        // up and down arrows                                          //
        event.preventDefault();                                        // 663
      }                                                                //
      var focus = $(template.find(":focus"));                          // 665
      var items = template.$("[role=tab]:visible");                    // 666
      var focusIndex = items.index(focus);                             // 667
      var newFocusIndex;                                               // 668
      if (event.keyCode == 37) {                                       // 669
        // left arrow                                                  //
        event.preventDefault();                                        // 670
        newFocusIndex = focusIndex - 1;                                // 671
        if (newFocusIndex == -1) {                                     // 672
          newFocusIndex = items.length - 1;                            // 673
        }                                                              //
      } else if (event.keyCode == 39) {                                //
        // right arrow                                                 //
        event.preventDefault();                                        // 676
        newFocusIndex = focusIndex + 1;                                // 677
        if (newFocusIndex >= items.length) {                           // 678
          newFocusIndex = 0;                                           // 679
        }                                                              //
      } else if (event.keyCode == 13) {                                //
        // Enter key                                                   //
        event.preventDefault();                                        // 682
        activateTargetTab(event, template);                            // 683
      }                                                                //
      if (newFocusIndex != null) {                                     // 685
        items.attr("tabindex", "-1");                                  // 686
        newFocus = $(items[newFocusIndex]);                            // 687
        newFocus.attr("tabindex", "0");                                // 688
        newFocus.focus();                                              // 689
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.shareableLinkTab.events({                                   // 694
    "change .share-token-role": function (event, instance) {           // 695
      var success = instance.completionState.get().success;            // 696
      if (success) {                                                   // 697
        var roleList = event.target;                                   // 698
        var assignment;                                                // 699
        if (roleList) {                                                // 700
          assignment = { roleId: roleList.selectedIndex };             // 701
        } else {                                                       //
          assignment = { none: null };                                 // 703
        }                                                              //
        Meteor.call("updateApiToken", success.id, { roleAssignment: assignment }, function (error) {
          if (error) {                                                 // 706
            console.error(error.stack);                                // 707
          }                                                            //
        });                                                            //
      }                                                                //
    },                                                                 //
    "change .label": function (event, instance) {                      // 712
      var success = instance.completionState.get().success;            // 713
      if (success) {                                                   // 714
        var label = event.target.value;                                // 715
        Meteor.call("updateApiToken", success.id, { petname: label }, function (error) {
          if (error) {                                                 // 717
            console.error(error.stack);                                // 718
          }                                                            //
        });                                                            //
      }                                                                //
    },                                                                 //
    "submit form.new-share-token": function (event, instance) {        // 723
      event.preventDefault();                                          // 724
      if (!instance.completionState.get().clear) {                     // 725
        return;                                                        // 726
      }                                                                //
      var currentGrain = getActiveGrain(globalGrains.get());           // 728
      var grainId = currentGrain.grainId();                            // 729
      var roleList = event.target.getElementsByClassName("share-token-role")[0];
      var assignment;                                                  // 731
      if (roleList) {                                                  // 732
        assignment = { roleId: roleList.selectedIndex };               // 733
      } else {                                                         //
        assignment = { none: null };                                   // 735
      }                                                                //
      instance.completionState.set({ "pending": true });               // 737
      Meteor.call("newApiToken", { identityId: currentGrain.identityId() }, grainId, event.target.getElementsByClassName("label")[0].value, assignment, { webkey: { forSharing: true } }, function (error, result) {
        if (error) {                                                   // 742
          console.error(error.stack);                                  // 743
        } else {                                                       //
          result.url = getOrigin() + "/shared/" + result.token;        // 745
          instance.completionState.set({ "success": result });         // 746
          // On the next render, .copy-me will exist, and we should focus it then.
          Meteor.defer(function () {                                   // 748
            var element = instance.find(".copy-me");                   // 749
            element.focus();                                           // 750
            selectElementContents(element);                            // 751
          });                                                          //
        }                                                              //
      });                                                              //
    },                                                                 //
    "click .reset-share-token": function (event, instance) {           // 756
      instance.completionState.set({ clear: true });                   // 757
      instance.find("form").reset();                                   // 758
      instance.find("form option[data-default-selected=true]").selected = true;
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainPowerboxRequestPopup.events({                          // 763
    "submit #powerbox-request-form": function (event) {                // 764
      event.preventDefault();                                          // 765
      var powerboxRequestInfo = this;                                  // 766
      var identityId = getActiveGrain(globalGrains.get()).identityId();
      Meteor.call("finishPowerboxRequest", event.target.token.value, powerboxRequestInfo.saveLabel, identityId, powerboxRequestInfo.grainId, function (err, token) {
        if (err) {                                                     // 771
          powerboxRequestInfo.error.set(err.toString());               // 772
        } else {                                                       //
          powerboxRequestInfo.source.postMessage({                     // 774
            rpcId: powerboxRequestInfo.rpcId,                          // 776
            token: token                                               // 777
          }, powerboxRequestInfo.origin);                              //
          powerboxRequestInfo.closer.close();                          // 779
        }                                                              //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainPowerboxOfferPopup.events({                            // 786
    "click button.dismiss": function (event, instance) {               // 787
      var sessionId = instance.data.sessionId;                         // 788
      if (sessionId) {                                                 // 789
        Meteor.call("finishPowerboxOffer", sessionId, function (err) {
          // TODO(someday): display the error nicely to the user       //
          if (err) {                                                   // 792
            console.error(err);                                        // 793
          }                                                            //
        });                                                            //
      } else {                                                         //
        // TODO(cleanup): This path is used by the admin UI. This is really hacky, though.
        Iron.controller().state.set("powerboxOfferUrl", null);         // 798
      }                                                                //
    },                                                                 //
    "click .copy-me": selectTargetContents,                            // 801
    "focus .copy-me": selectTargetContents                             // 802
  });                                                                  //
                                                                       //
  Template.grainSharePopup.helpers({                                   // 805
    "currentGrain": function () {                                      // 806
      return getActiveGrain(globalGrains.get());                       // 807
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainView.helpers({                                         // 811
    unpackedGrainState: function () {                                  // 812
      return mapGrainStateToTemplateData(this);                        // 813
    },                                                                 //
    identityPickerData: function () {                                  // 815
      var identities = SandstormDb.getUserIdentityIds(Meteor.user()).map(function (id) {
        var identity = Meteor.users.findOne({ _id: id });              // 817
        if (identity) {                                                // 818
          SandstormDb.fillInProfileDefaults(identity);                 // 819
          SandstormDb.fillInIntrinsicName(identity);                   // 820
          SandstormDb.fillInPictureUrl(identity);                      // 821
          return identity;                                             // 822
        }                                                              //
      });                                                              //
      var grain = getActiveGrain(globalGrains.get());                  // 825
      return { identities: identities,                                 // 826
        onPicked: function (identityId) {                              // 827
          grain.revealIdentity(identityId);                            // 827
        } };                                                           //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grain.helpers({                                             // 831
    currentGrain: function () {                                        // 832
      return getActiveGrain(globalGrains.get());                       // 833
    },                                                                 //
    isOwner: function () {                                             // 835
      var current = getActiveGrain(globalGrains.get());                // 836
      return current && current.isOwner();                             // 837
    },                                                                 //
                                                                       //
    grainSize: function () {                                           // 840
      var current = getActiveGrain(globalGrains.get());                // 841
      var sizeBytes = current && current.size();                       // 842
      if (sizeBytes !== undefined) {                                   // 843
        return "(" + prettySize(sizeBytes) + ")";                      // 844
      } else {                                                         //
        return "";                                                     // 846
      }                                                                //
    },                                                                 //
                                                                       //
    setGrainWindowTitle: function () {                                 // 850
      var current = getActiveGrain(globalGrains.get());                // 851
      if (current) {                                                   // 852
        current.updateDocumentTitle();                                 // 853
      }                                                                //
    },                                                                 //
                                                                       //
    displayWebkeyButton: function () {                                 // 857
      // TODO: figure out what this should do                          //
      return Meteor.userId() || !this.oldSharingModel;                 // 859
    },                                                                 //
                                                                       //
    showPowerboxOffer: function () {                                   // 862
      var current = getActiveGrain(globalGrains.get());                // 863
      if (current) {                                                   // 864
        var session = Sessions.findOne({ _id: current.sessionId() }, { fields: { powerboxView: 1 } });
        return session && session.powerboxView && !!session.powerboxView.offer;
      }                                                                //
      return false;                                                    // 868
    }                                                                  //
                                                                       //
  });                                                                  //
                                                                       //
  Template.grainTitle.helpers({                                        // 873
    title: function () {                                               // 874
      var grain = getActiveGrain(globalGrains.get());                  // 875
      return grain && grain.title() || "Untitled grain";               // 876
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainApiTokenPopup.helpers({                                // 880
    displayToken: function () {                                        // 881
      return !this.revoked && !this.expiresIfUnused && !this.parentToken;
    },                                                                 //
    existingTokens: function () {                                      // 884
      var current = getActiveGrain(globalGrains.get());                // 885
      return current && ApiTokens.find({ grainId: current.grainId(),   // 886
        identityId: current.identityId(),                              // 887
        forSharing: { $ne: true },                                     // 888
        $or: [{ owner: { webkey: null } }, { owner: { $exists: false } }],
        expiresIfUnused: null });                                      // 891
    },                                                                 //
    generatedApiToken: function () {                                   // 893
      var current = getActiveGrain(globalGrains.get());                // 894
      return current && current.generatedApiToken();                   // 895
    },                                                                 //
    generatedApiTokenPending: function () {                            // 897
      var current = getActiveGrain(globalGrains.get());                // 898
      return (current && current.generatedApiToken()) == "pending";    // 899
    },                                                                 //
    viewInfo: function () {                                            // 901
      var current = getActiveGrain(globalGrains.get());                // 902
      return current && current.viewInfo();                            // 903
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.whoHasAccessPopup.onCreated(function () {                   // 907
    var instance = this;                                               // 908
    this.subscribe("contactProfiles");                                 // 909
    var currentGrain = getActiveGrain(globalGrains.get());             // 910
    instance.identityId = currentGrain.identityId();                   // 911
    instance.grainId = currentGrain.grainId();                         // 912
    instance.transitiveShares = new ReactiveVar(null);                 // 913
    instance.downstreamTokensById = new ReactiveVar({});               // 914
    this.resetTransitiveShares = function () {                         // 915
      Meteor.call("transitiveShares", instance.identityId, instance.grainId, function (error, downstream) {
        if (error) {                                                   // 918
          console.error(error.stack);                                  // 919
        } else {                                                       //
          var result;                                                  //
                                                                       //
          (function () {                                               //
            var downstreamTokensById = {};                             // 921
            var sharesByRecipient = {};                                // 922
            downstream.forEach(function (token) {                      // 923
              downstreamTokensById[token._id] = token;                 // 924
              if (Match.test(token.owner, { user: Match.ObjectIncluding({ identityId: String }) })) {
                var recipient = token.owner.user.identityId;           // 926
                if (!sharesByRecipient[recipient]) {                   // 927
                  sharesByRecipient[recipient] = {                     // 928
                    recipient: recipient,                              // 929
                    dedupedShares: [],                                 // 930
                    allShares: []                                      // 931
                  };                                                   //
                }                                                      //
                sharesByRecipient[recipient].allShares.push(token);    // 934
                var dedupedShares = sharesByRecipient[recipient].dedupedShares;
                if (!dedupedShares.some(function (share) {             // 936
                  return share.identityId === token.identityId;        //
                })) {                                                  //
                  dedupedShares.push(token);                           // 937
                }                                                      //
              }                                                        //
            });                                                        //
            result = _.values(sharesByRecipient);                      // 941
                                                                       //
            if (result.length == 0) {                                  // 942
              result = { empty: true };                                // 943
            }                                                          //
            instance.transitiveShares.set(result);                     // 945
            instance.downstreamTokensById.set(downstreamTokensById);   // 946
          })();                                                        //
        }                                                              //
      });                                                              //
    };                                                                 //
    this.resetTransitiveShares();                                      // 950
  });                                                                  //
                                                                       //
  Template.whoHasAccessPopup.events({                                  // 953
    "change .share-token-role": function (event, instance) {           // 954
      var roleList = event.target;                                     // 955
      var assignment;                                                  // 956
      if (roleList) {                                                  // 957
        assignment = { roleId: roleList.selectedIndex };               // 958
      } else {                                                         //
        assignment = { none: null };                                   // 960
      }                                                                //
      Meteor.call("updateApiToken", roleList.getAttribute("data-token-id"), { roleAssignment: assignment }, function (error) {
        if (error) {                                                   // 964
          console.error(error.stack);                                  // 965
        }                                                              //
      });                                                              //
    },                                                                 //
    "click button.revoke-token": function (event, instance) {          // 969
      Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { revoked: true });
      instance.resetTransitiveShares();                                // 972
    },                                                                 //
    "click button.revoke-access": function (event, instance) {         // 974
      var recipient = event.currentTarget.getAttribute("data-recipient");
      var transitiveShares = instance.transitiveShares.get();          // 976
      var tokensById = instance.downstreamTokensById.get();            // 977
      var recipientShares = _.findWhere(transitiveShares, { recipient: recipient });
      var currentIdentityId = Accounts.getCurrentIdentityId();         // 979
      var recipientTokens = _.where(recipientShares.allShares, { identityId: currentIdentityId });
                                                                       //
      // Two cases:                                                    //
      // 1. All of the tokens are direct shares. Easy. Just revoke them.
      // 2. Some of the links are child tokens. For each, we walk up the chain of parents to the
      //    root. Then we walk downwards to see whether any other identities would be immediately
      //    affected by revoking those roots. We then collect that data and display it in a
      //    confirmation dialog.                                       //
                                                                       //
      var directTokens = [];                                           // 989
      // Direct shares from the current identity to the recipient.     //
                                                                       //
      var rootTokens = [];                                             // 992
      // Roots of chains of child tokens starting at the current identity and leading to the
      // recipient.                                                    //
                                                                       //
      recipientTokens.forEach(function (token) {                       // 996
        if (token.parentToken) {                                       // 997
          var currentToken = token;                                    // 998
          do {                                                         // 999
            currentToken = tokensById[token.parentToken];              // 1000
          } while (currentToken.parentToken);                          //
                                                                       //
          rootTokens.push(currentToken);                               // 1003
        } else {                                                       //
          directTokens.push(token);                                    // 1005
        }                                                              //
      });                                                              //
                                                                       //
      if (rootTokens.length > 0) {                                     // 1009
        var _ret2 = (function () {                                     //
          // Some of the shares are not direct.                        //
                                                                       //
          var tokensByParent = {};                                     // 1012
          for (var id in babelHelpers.sanitizeForInObject(tokensById)) {
            var token = tokensById[id];                                // 1014
            if (token.parentToken) {                                   // 1015
              if (!tokensByParent[token.parentToken]) {                // 1016
                tokensByParent[token.parentToken] = [];                // 1017
              }                                                        //
              tokensByParent[token.parentToken].push(token);           // 1019
            }                                                          //
          }                                                            //
                                                                       //
          var otherAffectedIdentities = {};                            // 1023
          var tokenStack = rootTokens.slice(0);                        // 1024
          while (tokenStack.length > 0) {                              // 1025
            var current = tokenStack.pop();                            // 1026
            if (Match.test(current.owner, { user: Match.ObjectIncluding({ identityId: String }) })) {
              if (current.owner.user.identityId != recipient) {        // 1029
                otherAffectedIdentities[current.owner.user.identityId] = true;
              }                                                        //
            } else {                                                   //
              if (tokensByParent[current._id]) {                       // 1033
                var children = tokensByParent[current._id];            // 1034
                children.forEach(function (child) {                    // 1035
                  tokenStack.push(child);                              // 1036
                });                                                    //
              }                                                        //
            }                                                          //
          }                                                            //
                                                                       //
          var recipientIdentity = ContactProfiles.findOne({ _id: recipient });
          var recipientName = recipientIdentity && recipientIdentity.profile.name || "Unknown User";
          var singular = rootTokens.length == 1;                       // 1045
                                                                       //
          var tokenLabels = _.pluck(rootTokens, "petname").map(function (petname) {
            return petname || "Unlabeled Link";                        //
          }).map(function (petname) {                                  //
            return "\"" + petname + "\"";                              //
          }).join(", ");                                               //
                                                                       //
          var confirmText = "This will revoke the following sharing link" + (singular ? "" : "s") + ":\n\n    " + tokenLabels + "\n\n";
                                                                       //
          var othersNote = "(No signed-in user other than " + recipientName + " has opened " + (singular ? "this link" : "these links") + " yet.)";
          if (Object.keys(otherAffectedIdentities).length > 0) {       // 1057
            var othersNames = Object.keys(otherAffectedIdentities).map(function (identityId) {
              return ContactProfiles.findOne({ _id: identityId });     //
            }).map(function (identity) {                               //
              return identity && identity.profile.name || "Unknown User";
            }).join(", ");                                             //
            othersNote = (singular ? "This link has" : "These links have") + " also been opened by:\n\n    " + othersNames;
          }                                                            //
                                                                       //
          if (window.confirm(confirmText + othersNote)) {              // 1066
            rootTokens.forEach(function (token) {                      // 1067
              Meteor.call("updateApiToken", token._id, { revoked: true });
            });                                                        //
          } else {                                                     //
            // Cancel.                                                 //
            return {                                                   // 1072
              v: undefined                                             //
            };                                                         //
          }                                                            //
        })();                                                          //
                                                                       //
        if (typeof _ret2 === "object") return _ret2.v;                 //
      }                                                                //
                                                                       //
      directTokens.forEach(function (token) {                          // 1076
        Meteor.call("updateApiToken", token._id, { revoked: true });   // 1077
      });                                                              //
                                                                       //
      instance.resetTransitiveShares();                                // 1080
    },                                                                 //
                                                                       //
    "click .token-petname": function (event, instance) {               // 1083
      // TODO(soon): Find a less-annoying way to get this input, perhaps by allowing the user
      //   to edit the petname in place.                               //
      var petname = window.prompt("Set new label:", this.petname);     // 1086
      if (petname) {                                                   // 1087
        Meteor.call("updateApiToken", event.currentTarget.getAttribute("data-token-id"), { petname: petname });
      }                                                                //
      instance.resetTransitiveShares();                                // 1091
    }                                                                  //
  });                                                                  //
                                                                       //
  function isEmptyPermissionSet(permissionSet) {                       // 1095
    if (!permissionSet) {                                              // 1096
      return true;                                                     // 1097
    }                                                                  //
    for (var ii = 0; ii < permissionSet.length; ++ii) {                // 1099
      if (permissionSet[ii]) {                                         // 1100
        return false;                                                  // 1101
      }                                                                //
    }                                                                  //
    return true;                                                       // 1104
  }                                                                    //
                                                                       //
  Template.whoHasAccessPopup.helpers({                                 // 1107
    existingShareTokens: function () {                                 // 1108
      if (Meteor.userId()) {                                           // 1109
        return ApiTokens.find({ grainId: Template.instance().grainId,  // 1110
          identityId: Template.instance().identityId,                  // 1111
          forSharing: true,                                            // 1112
          $or: [{ owner: { webkey: null } }, { owner: { $exists: false } }]
        }).fetch();                                                    //
      }                                                                //
    },                                                                 //
    isCurrentIdentity: function () {                                   // 1118
      if (this.identityId === Accounts.getCurrentIdentityId()) {       // 1119
        return true;                                                   // 1120
      }                                                                //
    },                                                                 //
    getPetname: function () {                                          // 1123
      if (this.petname) {                                              // 1124
        return this.petname;                                           // 1125
      } else {                                                         //
        return "Unlabeled Link";                                       // 1127
      }                                                                //
    },                                                                 //
    displayName: function (identityId) {                               // 1130
      var identity = ContactProfiles.findOne({ _id: identityId });     // 1131
      if (!identity) {                                                 // 1132
        identity = Meteor.users.findOne({ _id: identityId });          // 1133
      }                                                                //
      if (identity) {                                                  // 1135
        return identity.profile.name;                                  // 1136
      } else {                                                         //
        return "Unknown User (" + identityId.slice(0, 16) + ")";       // 1138
      }                                                                //
    },                                                                 //
    transitiveShares: function () {                                    // 1141
      return Template.instance().transitiveShares.get();               // 1142
    },                                                                 //
    indexedRoles: function () {                                        // 1144
      var result = [];                                                 // 1145
      var instance = Template.instance();                              // 1146
      var currentGrain = getActiveGrain(globalGrains.get());           // 1147
      var roles = currentGrain.viewInfo().roles;                       // 1148
      for (var ii = 0; ii < roles.length; ++ii) {                      // 1149
        result.push({ idx: ii, title: roles[ii].title, verbPhrase: roles[ii].verbPhrase });
      }                                                                //
      return result;                                                   // 1152
    },                                                                 //
    roleText: function () {                                            // 1154
      if (this.verbPhrase) {                                           // 1155
        return this.verbPhrase.defaultText;                            // 1156
      } else {                                                         //
        return "is " + this.title.defaultText;                         // 1158
      }                                                                //
    },                                                                 //
    hasCustomRole: function (token) {                                  // 1161
      var role = token.roleAssignment;                                 // 1162
      if ("roleId" in role && isEmptyPermissionSet(role.addPermissions) && isEmptyPermissionSet(role.removePermissions)) {
        return false;                                                  // 1166
      }                                                                //
      return true;                                                     // 1168
    },                                                                 //
    hasCurrentRole: function (token) {                                 // 1170
      var role = token.roleAssignment;                                 // 1171
      if ("roleId" in role && role.roleId == this.idx && isEmptyPermissionSet(role.addPermissions) && isEmptyPermissionSet(role.removePermissions)) {
        return true;                                                   // 1175
      }                                                                //
      return false;                                                    // 1177
    },                                                                 //
    displayToken: function () {                                        // 1179
      return !this.revoked && !this.expiresIfUnused && !this.parentToken;
    },                                                                 //
    viewInfo: function () {                                            // 1182
      var activeGrain = getActiveGrain(globalGrains.get());            // 1183
      return activeGrain && activeGrain.viewInfo();                    // 1184
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.selectRole.helpers({                                        // 1188
    roleText: function () {                                            // 1189
      if (this.verbPhrase) {                                           // 1190
        return this.verbPhrase.defaultText;                            // 1191
      } else {                                                         //
        return "is " + this.title.defaultText;                         // 1193
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.shareableLinkTab.onCreated(function () {                    // 1198
    this.completionState = new ReactiveVar({ clear: true });           // 1199
  });                                                                  //
                                                                       //
  Template.emailInviteTab.onCreated(function () {                      // 1202
    this.completionState = new ReactiveVar({ clear: true });           // 1203
    this.contacts = new ReactiveVar([]);                               // 1204
  });                                                                  //
                                                                       //
  Template.shareableLinkTab.helpers({                                  // 1207
    completionState: function () {                                     // 1208
      var instance = Template.instance();                              // 1209
      return instance.completionState.get();                           // 1210
    }                                                                  //
  });                                                                  //
  Template.emailInviteTab.helpers({                                    // 1213
    completionState: function () {                                     // 1214
      var instance = Template.instance();                              // 1215
      return instance.completionState.get();                           // 1216
    },                                                                 //
    contacts: function () {                                            // 1218
      return Template.instance().contacts;                             // 1219
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.emailInviteTab.events({                                     // 1223
    "submit form.email-invite": function (event, instance) {           // 1224
      event.preventDefault();                                          // 1225
      return false;                                                    // 1226
    },                                                                 //
    "click form.email-invite button": function (event, instance) {     // 1228
      event.preventDefault();                                          // 1229
      if (!instance.completionState.get().clear) {                     // 1230
        return;                                                        // 1231
      }                                                                //
      var grainId = instance.data.grainId;                             // 1233
      var title = instance.data.title;                                 // 1234
                                                                       //
      var roleList = instance.find(".share-token-role");               // 1236
      var assignment;                                                  // 1237
      if (roleList) {                                                  // 1238
        assignment = { roleId: roleList.selectedIndex };               // 1239
      } else {                                                         //
        assignment = { none: null };                                   // 1241
      }                                                                //
      var message = instance.find(".personal-message").value;          // 1243
      instance.completionState.set({ pending: true });                 // 1244
                                                                       //
      var currentGrain = getActiveGrain(globalGrains.get());           // 1246
                                                                       //
      // HTML-escape the message.                                      //
      var div = document.createElement('div');                         // 1249
      div.appendChild(document.createTextNode(message));               // 1250
      var htmlMessage = div.innerHTML.replace(/\n/g, "<br>");          // 1251
                                                                       //
      var contacts = instance.contacts.get();                          // 1253
      var emails = instance.find("input.emails");                      // 1254
      var emailsValue = emails.value;                                  // 1255
      if (emailsValue) {                                               // 1256
        if (emailsValue.match(/.+\@.+\..+/)) {                         // 1257
          contacts.push({                                              // 1258
            _id: emailsValue,                                          // 1259
            profile: {                                                 // 1260
              service: "email",                                        // 1261
              name: emailsValue,                                       // 1262
              intrinsicName: emailsValue,                              // 1263
              pictureUrl: "/email.svg"                                 // 1264
            },                                                         //
            isDefault: true                                            // 1266
          });                                                          //
          instance.contacts.set(contacts);                             // 1268
          emails.value = "";                                           // 1269
        } else {                                                       //
          instance.completionState.set({ error: "Invalid text in contact input box: " + emailsValue });
          return;                                                      // 1273
        }                                                              //
      }                                                                //
                                                                       //
      Meteor.call("inviteUsersToGrain", getOrigin(), currentGrain.identityId(), grainId, title, assignment, contacts, { text: message, html: htmlMessage }, function (error, result) {
        if (error) {                                                   // 1280
          instance.completionState.set({ error: error.toString() });   // 1281
        } else {                                                       //
          if (result.failures.length > 0) {                            // 1283
            var message = "Failed to send to: ";                       // 1284
            for (var ii = 0; ii < result.failures.length; ++ii) {      // 1285
              console.error(result.failures[ii].error);                // 1286
              if (ii != 0) {                                           // 1287
                message += ", ";                                       // 1288
              }                                                        //
              message += result.failures[ii].contact.profile.name;     // 1290
              var warning = result.failures[ii].warning;               // 1291
              if (warning) {                                           // 1292
                message += ". " + warning;                             // 1293
              }                                                        //
            }                                                          //
            instance.completionState.set({ error: message });          // 1296
          } else {                                                     //
            instance.completionState.set({ success: "success" });      // 1298
          }                                                            //
        }                                                              //
      });                                                              //
    },                                                                 //
    "click .reset-invite": function (event, instance) {                // 1303
      instance.contacts.set([]);                                       // 1304
      instance.completionState.set({ clear: true });                   // 1305
      instance.find("form").reset();                                   // 1306
      instance.find("form option[data-default-selected=true]").selected = true;
    },                                                                 //
    "click .start-over-invite": function (event, instance) {           // 1309
      instance.completionState.set({ clear: true });                   // 1310
    }                                                                  //
  });                                                                  //
                                                                       //
  Template.grainPowerboxOfferPopup.helpers({                           // 1314
    powerboxOfferUrl: function () {                                    // 1315
      if (this.powerboxOfferUrl) {                                     // 1316
        // TODO(cleanup): This path is used by the admin UI. This is really hacky, though.
        return this.powerboxOfferUrl;                                  // 1318
      }                                                                //
                                                                       //
      var activeGrain = getActiveGrain(globalGrains.get());            // 1321
      var session = Sessions.findOne({ _id: activeGrain.sessionId() }, { fields: { powerboxView: 1 } });
      return session && session.powerboxView && session.powerboxView.offer;
    }                                                                  //
  });                                                                  //
                                                                       //
  // Send a keep-alive for each grain every now and then.              //
  Meteor.setInterval(function () {                                     // 1328
    var grains = globalGrains.get();                                   // 1329
    if (!grains) return;                                               // 1330
                                                                       //
    // Meteor has an exponential backoff of up to 5 minutes for reconnect. This is unnacceptable
    // for us, since we rely on Sessions being re-established in under 60s.
    if (Meteor.status().status === "waiting") {                        // 1334
      console.log("Sandstorm is trying to reconnect...");              // 1335
      Meteor.reconnect();                                              // 1336
    }                                                                  //
                                                                       //
    grains.forEach(function (grain) {                                  // 1339
      if (grain.sessionId()) {                                         // 1340
        // TODO(soon):  Investigate what happens in background tabs.  Maybe arrange to re-open the
        //   app if it dies while in the background.                   //
        Meteor.call("keepSessionAlive", grain.sessionId(), function (error, result) {
          // Sessions will automatically resume if possible, otherwise they will refresh.
        });                                                            //
      }                                                                //
    });                                                                //
  }, 60000);                                                           //
                                                                       //
  // Message handler for Sandstorm's client-side postMessage API.      //
  Meteor.startup(function () {                                         // 1351
    var messageListener = function (event) {                           // 1352
      if (event.origin === getOrigin()) {                              // 1353
        // Meteor likes to postMessage() to itself sometimes, so we ignore these messages.
        return;                                                        // 1355
      }                                                                //
                                                                       //
      // Look up the grain index of which grain this postmessage came from, so we can map behavior
      // into that particular grain's state                            //
      var grains = globalGrains.get();                                 // 1360
      var senderGrainIndex = grainOriginToIndex(grains, event.origin);
      if (senderGrainIndex == -1) {                                    // 1362
        // We got a postMessage from an origin that is not a grain we currently believe is open.
        // Ignore it. (It may be aimed at some other message listener registered elsewhere...)
        return;                                                        // 1365
      }                                                                //
      var senderGrain = grains[senderGrainIndex];                      // 1367
                                                                       //
      if (event.data.setPath || event.data.setPath === '') {           // 1369
        var path = event.data.setPath || '/';                          // 1370
        check(path, String);                                           // 1371
        check(path.charAt(0), '/');                                    // 1372
        // TODO(security): More sanitization of this path. E.g. reject "/../../".
        senderGrain.setPath(path);                                     // 1374
      } else if (event.data.startSharing) {                            //
        // Allow the current grain to request that the "Share Access" menu be shown.
        // Only show this popup if no other popup is currently active.
        // TODO(security): defend against malicious apps spamming this call, blocking all other UI.
        var currentGrain = getActiveGrain(globalGrains.get());         // 1379
        if (senderGrain === currentGrain && !globalTopbar.isPopupOpen()) {
          globalTopbar.openPopup("share");                             // 1381
        }                                                              //
      } else if (event.data.setTitle || event.data.setTitle === '') {  //
        senderGrain.setFrameTitle(event.data.setTitle);                // 1384
      } else if (event.data.renderTemplate) {                          //
        // Request creation of a single-use template with a privileged API token.
        // Why?  Apps should not be able to obtain capabilities-as-keys to
        // themselves directly, because those can be leaked through an
        // arbitrary bit stream or covert channel.  However, apps often need a
        // way to provide instructions to users to copy/paste with some
        // privileged token contained within.  By providing this templating in
        // the platform, we can ensure that the token is only visible to the
        // shell's origin.                                             //
        var call = event.data.renderTemplate;                          // 1394
        check(call, Object);                                           // 1395
        var rpcId = call.rpcId;                                        // 1396
        try {                                                          // 1397
          check(call, { rpcId: String, template: String, petname: Match.Optional(String),
            roleAssignment: Match.Optional(roleAssignmentPattern),     // 1399
            forSharing: Match.Optional(Boolean) });                    // 1400
        } catch (error) {                                              //
          event.source.postMessage({ rpcId: rpcId, error: error.toString() }, event.origin);
          return;                                                      // 1403
        }                                                              //
        var template = call.template;                                  // 1405
        var petname = "connected external app";                        // 1406
        if (call.petname) {                                            // 1407
          petname = call.petname;                                      // 1408
        }                                                              //
        var assignment = { allAccess: null };                          // 1410
        if (call.roleAssignment) {                                     // 1411
          assignment = call.roleAssignment;                            // 1412
        }                                                              //
        var forSharing = call.forSharing ? call.forSharing : false;    // 1414
        // Tokens expire by default in 5 minutes from generation date  //
        var selfDestructDuration = 5 * 60 * 1000;                      // 1416
                                                                       //
        var provider;                                                  // 1418
        if (Router.current().route.getName() === "shared") {           // 1419
          provider = { rawParentToken: Router.current().params.token };
        } else {                                                       //
          provider = { identityId: senderGrain.identityId() };         // 1422
        }                                                              //
        var owner = { webkey: { forSharing: forSharing,                // 1424
            expiresIfUnusedDuration: selfDestructDuration } };         // 1425
                                                                       //
        Meteor.call("newApiToken", provider, senderGrain.grainId(), petname, assignment, owner, function (error, result) {
          if (error) {                                                 // 1429
            event.source.postMessage({ rpcId: rpcId, error: error.toString() }, event.origin);
          } else {                                                     //
            var tokenId = result.token;                                // 1432
            // Generate random key id2.                                //
            var id2 = Random.secret();                                 // 1434
            // Store apitoken id1 and template in session storage in the offer
            // template namespace under key id2.                       //
            var key = "offerTemplate" + id2;                           // 1437
            var renderedTemplate = template.replace(/\$API_TOKEN/g, tokenId).replace(/\$API_HOST/g, makeWildcardHost("api"));
            sessionStorage.setItem(key, JSON.stringify({               // 1440
              "token": tokenId,                                        // 1441
              "renderedTemplate": renderedTemplate,                    // 1442
              "expires": Date.now() + selfDestructDuration             // 1443
            }));                                                       //
            sessionStorage.setItem("apiHost", makeWildcardHost("api"));
                                                                       //
            // Send message to event.source with URL containing id2    //
            templateLink = window.location.origin + "/offer-template.html#" + id2;
            event.source.postMessage({ rpcId: rpcId, uri: templateLink }, event.origin);
          }                                                            //
        });                                                            //
      } else if (event.data.powerboxRequest) {                         //
        // TODO(now): make this work with GrainView                    //
        var powerboxRequest = event.data.powerboxRequest;              // 1455
        check(powerboxRequest, Object);                                // 1456
        var rpcId = powerboxRequest.rpcId;                             // 1457
                                                                       //
        var powerboxRequestInfo = {                                    // 1459
          source: event.source,                                        // 1460
          rpcId: rpcId,                                                // 1461
          grainId: senderGrain.grainId(),                              // 1462
          origin: event.origin,                                        // 1463
          saveLabel: powerboxRequest.saveLabel,                        // 1464
          error: new ReactiveVar(null)                                 // 1465
        };                                                             //
                                                                       //
        powerboxRequestInfo.closer = globalTopbar.addItem({            // 1468
          name: "request",                                             // 1469
          template: Template.grainPowerboxRequest,                     // 1470
          popupTemplate: Template.grainPowerboxRequestPopup,           // 1471
          data: new ReactiveVar(powerboxRequestInfo),                  // 1472
          startOpen: true,                                             // 1473
          onDismiss: function () {                                     // 1474
            powerboxRequestInfo.source.postMessage({                   // 1475
              rpcId: powerboxRequestInfo.rpcId,                        // 1477
              error: "User canceled request"                           // 1478
            }, powerboxRequestInfo.origin);                            //
            return "remove";                                           // 1480
          }                                                            //
        });                                                            //
      } else {                                                         //
        console.log("postMessage from app not understood: " + event.data);
        console.log(event);                                            // 1485
      }                                                                //
    };                                                                 //
                                                                       //
    window.addEventListener("message", messageListener, false);        // 1489
  });                                                                  //
}                                                                      //
                                                                       //
if (Meteor.isClient) {                                                 // 1493
  function maybeScrollLog() {                                          // 1494
    var elem = document.getElementById("grainLog");                    // 1495
    if (elem) {                                                        // 1496
      // The log already exists. It's about to be updated. Check if it's scrolled to the bottom
      // before the update.                                            //
      if (elem.scrollHeight - elem.scrollTop === elem.clientHeight) {  // 1499
        // Indeed, so we want to scroll it back to the bottom after the update.
        Tracker.afterFlush(function () {                               // 1501
          scrollLogToBottom(elem);                                     // 1501
        });                                                            //
      }                                                                //
    } else {                                                           //
      // No element exists yet, but it's probably about to be created, in which case we definitely
      // want to scroll it.                                            //
      Tracker.afterFlush(function () {                                 // 1506
        var elem2 = document.getElementById("grainLog");               // 1507
        if (elem2) scrollLogToBottom(elem2);                           // 1508
      });                                                              //
    }                                                                  //
  }                                                                    //
                                                                       //
  function scrollLogToBottom(elem) {                                   // 1513
    elem.scrollTop = elem.scrollHeight;                                // 1514
  }                                                                    //
}                                                                      //
                                                                       //
function makeGrainIdActive(grainId) {                                  // 1518
  var grains = globalGrains.get();                                     // 1519
  for (var i = 0; i < grains.length; i++) {                            // 1520
    var grain = grains[i];                                             // 1521
    if (grain.grainId() === grainId) {                                 // 1522
      if (!grain.isActive()) {                                         // 1523
        grain.setActive(true);                                         // 1524
      }                                                                //
    } else {                                                           //
      if (grain.isActive()) {                                          // 1527
        grain.setActive(false);                                        // 1528
      }                                                                //
    }                                                                  //
  }                                                                    //
}                                                                      //
                                                                       //
function getActiveGrain(grains) {                                      // 1534
  var idx = activeGrainIndex(grains);                                  // 1535
  return idx == -1 ? undefined : grains[idx];                          // 1536
}                                                                      //
                                                                       //
function activeGrainIndex(grains) {                                    // 1539
  for (var i = 0; i < grains.length; i++) {                            // 1540
    if (grains[i].isActive()) {                                        // 1541
      return i;                                                        // 1542
    }                                                                  //
  }                                                                    //
  return -1;                                                           // 1545
}                                                                      //
                                                                       //
function grainOriginToIndex(grains, origin) {                          // 1548
  for (var i = 0; i < grains.length; i++) {                            // 1549
    if (grains[i].origin() === origin) {                               // 1550
      return i;                                                        // 1551
    }                                                                  //
  }                                                                    //
  return -1;                                                           // 1554
}                                                                      //
                                                                       //
function grainIdToIndex(grains, grainId) {                             // 1557
  for (var i = 0; i < grains.length; i++) {                            // 1558
    var grain = grains[i];                                             // 1559
    if (grains[i].grainId() === grainId) {                             // 1560
      return i;                                                        // 1561
    }                                                                  //
  }                                                                    //
  return -1;                                                           // 1564
}                                                                      //
                                                                       //
function mapGrainStateToTemplateData(grainState) {                     // 1567
  var templateData = {                                                 // 1568
    grainId: grainState.grainId(),                                     // 1569
    active: grainState.isActive(),                                     // 1570
    title: grainState.title(),                                         // 1571
    error: grainState.error(),                                         // 1572
    appOrigin: grainState.origin(),                                    // 1573
    hasNotLoaded: !grainState.hasLoaded(),                             // 1574
    sessionId: grainState.sessionId(),                                 // 1575
    originalPath: grainState._originalPath,                            // 1576
    interstitial: grainState.shouldShowInterstitial(),                 // 1577
    token: grainState.token(),                                         // 1578
    viewInfo: grainState.viewInfo()                                    // 1579
  };                                                                   //
  return templateData;                                                 // 1581
}                                                                      //
                                                                       //
GrainLog = new Mongo.Collection("grainLog");                           // 1584
// Pseudo-collection created by subscribing to "grainLog", implemented in proxy.js.
                                                                       //
Router.map(function () {                                               // 1587
  this.route("apps", {                                                 // 1588
    path: "/apps",                                                     // 1589
    waitOn: function () {                                              // 1590
      return globalSubs;                                               // 1590
    },                                                                 //
    data: function () {                                                // 1591
      if (!this.ready()) return;                                       // 1592
      if (!Meteor.userId() && !Meteor.loggingIn()) {                   // 1593
        Router.go("root", {}, { replaceState: true });                 // 1594
      }                                                                //
                                                                       //
      return new SandstormAppList(globalDb, globalQuotaEnforcer);      // 1597
    }                                                                  //
  });                                                                  //
  this.route("newGrainRedirect", {                                     // 1600
    // The path /grain/new used to be where you would go to create new grains.
    // Its functionality has been superceded by the apps route, so redirect in
    // case anyone has the old link saved somewhere.                   //
    path: "/grain/new",                                                // 1604
    onBeforeAction: function () {                                      // 1605
      Router.go("apps", {}, { replaceState: true });                   // 1606
    }                                                                  //
  });                                                                  //
  this.route("grains", {                                               // 1609
    path: "/grain",                                                    // 1610
    waitOn: function () {                                              // 1611
      return globalSubs;                                               // 1611
    },                                                                 //
    data: function () {                                                // 1612
      if (!this.ready()) return;                                       // 1613
      if (!Meteor.userId() && !Meteor.loggingIn()) {                   // 1614
        Router.go("root", {}, { replaceState: true });                 // 1615
      }                                                                //
                                                                       //
      return new SandstormGrainListPage(globalDb, globalQuotaEnforcer);
    }                                                                  //
  });                                                                  //
  this.route("grain", {                                                // 1621
    path: "/grain/:grainId/:path(.*)?",                                // 1622
    loadingTemplate: "loadingNoMessage",                               // 1623
                                                                       //
    waitOn: function () {                                              // 1625
      return globalSubs;                                               // 1626
    },                                                                 //
                                                                       //
    onBeforeAction: function () {                                      // 1629
      // Run this hook only once.                                      //
      if (this.state.get("beforeActionHookRan")) {                     // 1631
        return this.next();                                            // 1631
      }                                                                //
      this.state.set("beforeActionHookRan", true);                     // 1632
      var grainId = this.params.grainId;                               // 1633
      var path = "/" + (this.params.path || "") + (this.originalUrl.match(/[#?].*$/) || "");
      var grains = globalGrains.get();                                 // 1635
      var grainIndex = grainIdToIndex(grains, grainId);                // 1636
      if (grainIndex == -1) {                                          // 1637
        // The element we need to attach our Blaze view to may not exist yet.
        // In that case, defer creating the GrainView until we're sure it's
        // had a chance to render.                                     //
        var openView = (function () {                                  // 1641
          function openView() {                                        // 1641
            var mainContentElement = document.querySelector("body>.main-content");
            if (mainContentElement) {                                  // 1643
              var grains = globalGrains.get();                         // 1644
              var grainToOpen = new GrainView(grainId, path, undefined, mainContentElement);
              grainToOpen.openSession();                               // 1646
              grainIndex = grains.push(grainToOpen) - 1;               // 1647
              globalGrains.set(grains);                                // 1648
              makeGrainIdActive(grainId);                              // 1649
            } else {                                                   //
              Meteor.defer(openView);                                  // 1651
            }                                                          //
          }                                                            //
                                                                       //
          return openView;                                             //
        })();                                                          //
        openView();                                                    // 1654
      } else {                                                         //
        makeGrainIdActive(grainId);                                    // 1656
      }                                                                //
      this.next();                                                     // 1658
    },                                                                 //
                                                                       //
    onStop: function () {                                              // 1661
      this.state.set("beforeActionHookRan", undefined);                // 1662
      globalGrains.get().forEach(function (grain) {                    // 1663
        if (grain.isActive()) {                                        // 1664
          grain.setActive(false);                                      // 1665
        }                                                              //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("/shared/:token/:path(.*)?", {                            // 1671
    name: "shared",                                                    // 1672
    template: "grain",                                                 // 1673
    loadingTemplate: "loadingNoMessage",                               // 1674
                                                                       //
    waitOn: function () {                                              // 1676
      return [Meteor.subscribe("devPackages"), Meteor.subscribe("tokenInfo", this.params.token), Meteor.subscribe("grainsMenu")
      // This subscription gives us the data we need for deciding whether to automatically reveal
      // our identity.                                                 //
      // TODO(soon): Subscribe to contacts instead.                    //
      ];                                                               //
    },                                                                 //
                                                                       //
    onBeforeAction: function () {                                      // 1688
      // Run this hook only once. We could accomplish the same thing by using the `onRun()` hook
      // and waiting for `this.ready()`, but for some reason that fails in the case when a user
      // logs in while visiting a /shared/ link.                       //
      if (this.state.get("beforeActionHookRan")) {                     // 1692
        return this.next();                                            // 1692
      }                                                                //
      this.state.set("beforeActionHookRan", true);                     // 1693
                                                                       //
      var token = this.params.token;                                   // 1695
      var path = "/" + (this.params.path || "") + (this.originalUrl.match(/[#?].*$/) || "");
      var hash = this.params.hash;                                     // 1697
                                                                       //
      var tokenInfo = TokenInfo.findOne({ _id: token });               // 1699
      if (tokenInfo && tokenInfo.invalidToken) {                       // 1700
        this.state.set("invalidToken", true);                          // 1701
      } else {                                                         //
        var grainId = tokenInfo.grainId;                               // 1703
        var grains = globalGrains.get();                               // 1704
        var grainIndex = grainIdToIndex(grains, grainId);              // 1705
        if (grainIndex == -1) {                                        // 1706
          var openView = (function () {                                // 1707
            function openView() {                                      // 1707
              var mainContentElement = document.querySelector("body>.main-content");
              if (mainContentElement) {                                // 1709
                var grains = globalGrains.get();                       // 1710
                var grainToOpen = new GrainView(grainId, path, tokenInfo, mainContentElement);
                grainToOpen.openSession();                             // 1712
                grainIndex = grains.push(grainToOpen) - 1;             // 1713
                globalGrains.set(grains);                              // 1714
                makeGrainIdActive(grainId);                            // 1715
              } else {                                                 //
                Meteor.defer(openView);                                // 1717
              }                                                        //
            }                                                          //
                                                                       //
            return openView;                                           //
          })();                                                        //
          openView();                                                  // 1720
        } else {                                                       //
          makeGrainIdActive(grainId);                                  // 1722
        }                                                              //
      }                                                                //
      this.next();                                                     // 1725
    },                                                                 //
                                                                       //
    action: function () {                                              // 1728
      if (this.state.get("invalidToken")) {                            // 1729
        this.render("invalidToken", { data: { token: this.params.token } });
      } else {                                                         //
        this.render();                                                 // 1732
      }                                                                //
    },                                                                 //
                                                                       //
    onStop: function () {                                              // 1736
      this.state.set("beforeActionHookRan", undefined);                // 1737
      this.state.set("invalidToken", undefined);                       // 1738
      globalGrains.get().forEach(function (grain) {                    // 1739
        if (grain.isActive()) {                                        // 1740
          grain.setActive(false);                                      // 1741
        }                                                              //
      });                                                              //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("grainLog", {                                             // 1747
    path: "/grainlog/:grainId",                                        // 1748
    layoutTemplate: "noLayout",                                        // 1749
                                                                       //
    waitOn: function () {                                              // 1751
      return [Meteor.subscribe("grainTopBar", this.params.grainId), Meteor.subscribe("grainLog", this.params.grainId)];
    },                                                                 //
                                                                       //
    data: function () {                                                // 1758
      if (this.ready()) {                                              // 1759
        maybeScrollLog();                                              // 1760
        var grain = Grains.findOne(this.params.grainId);               // 1761
        return {                                                       // 1762
          title: grain ? grain.title : "(deleted grain)",              // 1763
          html: AnsiUp.ansi_to_html(GrainLog.find({}, { $sort: { _id: 1 } }).map(function (entry) {
            return entry.text;                                         // 1765
          }).join(""), { use_classes: true })                          //
        };                                                             //
      }                                                                //
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("appDetails", {                                           // 1772
    path: "/apps/:appId",                                              // 1773
    template: "appDetails",                                            // 1774
    waitOn: function () {                                              // 1775
      return globalSubs.concat(Meteor.subscribe("appIndex", this.params.appId));
    },                                                                 //
    data: function () {                                                // 1779
      return new SandstormAppDetails(globalDb, globalQuotaEnforcer, this.params.appId);
    }                                                                  //
  });                                                                  //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=grain.js.map
