(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// shared/install.js                                                   //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
// This file covers /install and /upload.                              //
                                                                       //
var localizedTextPattern = {                                           // 19
  defaultText: String,                                                 // 20
  localizations: Match.Optional([{ locale: String, text: String }])    // 21
};                                                                     //
                                                                       //
if (Meteor.isServer) {                                                 // 24
  UserActions.allow({                                                  // 25
    insert: function (userId, action) {                                // 26
      // TODO(cleanup): This check keeps breaking. Use a method instead that takes the package
      //   ID as an argument.                                          //
      check(action, {                                                  // 29
        userId: String,                                                // 30
        packageId: String,                                             // 31
        appId: String,                                                 // 32
        appTitle: Match.Optional(localizedTextPattern),                // 33
        appMarketingVersion: Match.Optional(Object),                   // 34
        appVersion: Match.Integer,                                     // 35
        title: localizedTextPattern,                                   // 36
        nounPhrase: Match.Optional(localizedTextPattern),              // 37
        command: {                                                     // 38
          executablePath: Match.Optional(String),                      // 39
          deprecatedExecutablePath: Match.Optional(String),            // 40
          args: Match.Optional([String]),                              // 41
          argv: Match.Optional([String]),                              // 42
          environ: Match.Optional([{ key: String, value: String }])    // 43
        }                                                              //
      });                                                              //
      return userId && isSignedUpOrDemo() && action.userId === userId;
    },                                                                 //
    remove: function (userId, action) {                                // 48
      return userId && action.userId === userId;                       // 49
    }                                                                  //
  });                                                                  //
                                                                       //
  var uploadTokens = {};                                               // 53
  // Not all users are allowed to upload apps. We need to manually implement authorization
  // because Meteor.userId() is not available in server-side routes.   //
                                                                       //
  Meteor.methods({                                                     // 57
    cancelDownload: function (packageId) {                             // 58
      check(packageId, String);                                        // 59
                                                                       //
      // TODO(security):  Only let user cancel download if they initiated it.
      cancelDownload(packageId);                                       // 62
    },                                                                 //
                                                                       //
    newUploadToken: function () {                                      // 65
      if (!isSignedUp()) {                                             // 66
        throw new Meteor.Error(403, "Unauthorized", "Only invited users can upload apps.");
      }                                                                //
      if (globalDb.isUninvitedFreeUser()) {                            // 69
        throw new Meteor.Error(403, "Unauthorized", "Only paid users can upload apps.");
      }                                                                //
      var token = Random.id();                                         // 72
      uploadTokens[token] = setTimeout(function () {                   // 73
        delete uploadTokens[token];                                    // 74
      }, 20 * 60 * 1000);                                              //
      return token;                                                    // 76
    },                                                                 //
                                                                       //
    upgradeGrains: function (appId, version, packageId) {              // 79
      this.connection.sandstormDb.upgradeGrains(appId, version, packageId, globalBackend);
    }                                                                  //
  });                                                                  //
}                                                                      //
                                                                       //
function isSafeDemoAppUrl(url) {                                       // 85
  // For demo accounts, we allow using a bare hash with no URL (which will never upload a new app)
  // and we allow specifying a sandstorm.io URL.                       //
  return !url || url.lastIndexOf("http://sandstorm.io/", 0) === 0 || url.lastIndexOf("https://sandstorm.io/", 0) === 0 || url.lastIndexOf("https://alpha-j7uny7u376jnimcsx34c.sandstorm.io/", 0) === 0 || url.lastIndexOf("https://app-index.sandstorm.io/", 0) === 0;
}                                                                      //
                                                                       //
Meteor.methods({                                                       // 95
  ensureInstalled: function (packageId, url, isRetry) {                // 96
    check(packageId, String);                                          // 97
    check(url, Match.OneOf(String, undefined, null));                  // 98
    check(isRetry, Boolean);                                           // 99
                                                                       //
    if (!packageId.match(/^[a-zA-Z0-9]*$/)) {                          // 101
      throw new Meteor.Error(400, "The package ID contains illegal characters.");
    }                                                                  //
                                                                       //
    if (!this.userId) {                                                // 105
      if (allowDemo && isSafeDemoAppUrl(url)) {                        // 106
        // continue on                                                 //
      } else {                                                         //
          throw new Meteor.Error(403, "You must be logged in to install packages.");
        }                                                              //
    } else if (!isSignedUp() && !isDemoUser()) {                       //
      throw new Meteor.Error(403, "This Sandstorm server requires you to get an invite before installing apps.");
    } else if (isUserOverQuota(Meteor.user())) {                       //
      throw new Meteor.Error(402, "You are out of storage space. Please delete some things and try again.");
    }                                                                  //
                                                                       //
    if (!this.isSimulation) {                                          // 119
      var pkg = Packages.findOne(packageId);                           // 120
                                                                       //
      if (!pkg || pkg.status !== "ready") {                            // 122
        if (!this.userId || isDemoUser() || globalDb.isUninvitedFreeUser()) {
          if (!isSafeDemoAppUrl(url)) {                                // 124
            // TODO(someday): Billing prompt on client side.           //
            throw new Meteor.Error(403, "Sorry, demo and free users cannot upload custom apps; " + "they may only install apps from apps.sandstorm.io.");
          }                                                            //
        }                                                              //
      }                                                                //
                                                                       //
      if (!pkg || isRetry) {                                           // 132
        globalDb.startInstall(packageId, url, isRetry);                // 133
      }                                                                //
    }                                                                  //
  }                                                                    //
});                                                                    //
                                                                       //
Router.map(function () {                                               // 139
  this.route("install", {                                              // 140
    path: "/install/:packageId",                                       // 141
                                                                       //
    waitOn: function () {                                              // 143
      return [Meteor.subscribe("packageInfo", this.params.packageId), Meteor.subscribe("credentials"),
      // We need UserActions and Grains populated so we can check if we need to upgrade them.
      Meteor.subscribe("grainsMenu")];                                 // 148
    },                                                                 //
                                                                       //
    data: function () {                                                // 152
      if (!this.ready()) return;                                       // 153
                                                                       //
      var packageId = this.params.packageId;                           // 155
      var packageUrl = this.params.query && this.params.query.url;     // 156
      var handle = new SandstormAppInstall(packageId, packageUrl, globalDb);
                                                                       //
      var pkg = Packages.findOne(packageId);                           // 159
      if (!Meteor.userId()) {                                          // 160
        if (allowDemo && isSafeDemoAppUrl(packageUrl)) {               // 161
          if (pkg && pkg.status === "ready") {                         // 162
            Router.go("appdemo", { appId: pkg.appId }, { replaceState: true });
            return handle;                                             // 164
          } else {                                                     //
            // continue on and install...                              //
          }                                                            //
        } else {                                                       //
            handle.setError("You must sign in to install packages.");  // 169
            return handle;                                             // 170
          }                                                            //
      } else {                                                         //
        try {                                                          // 173
          // When the user clicks to install an app, the app store is opened in a new tab. When they
          // choose an app in the app store, they are redirected back to Sandstorm. But we'd really
          // like to bring them back to the tab where they clicked "install apps". It turns out we
          // can exploit a terrible feature of the web platform to do this: window.opener is a
          // pointer to the tab which opened this tab, and we can actually reach right into it and
          // call functions in it. So, we redirect the original tab to install the app, then close
          // this one.                                                 //
          if (globalGrains.get().length === 0 && window.opener.location.hostname === window.location.hostname && window.opener.Router) {
            window.opener.Router.go("install", { packageId: packageId }, { query: this.params.query });
            window.close();                                            // 185
            return handle;                                             // 186
          }                                                            //
        } catch (err) {                                                //
          // Probably security error because window.opener is in a different domain.
        }                                                              //
                                                                       //
        if (!isSignedUp() && !isDemoUser()) {                          // 192
          handle.setError("This Sandstorm server requires you to get an invite before installing apps.");
          return handle;                                               // 194
        }                                                              //
      }                                                                //
                                                                       //
      Meteor.call("ensureInstalled", packageId, packageUrl, false, function (err, result) {
        if (err) {                                                     // 200
          console.log(err);                                            // 201
          handle.setError(err.message);                                // 202
        }                                                              //
      });                                                              //
                                                                       //
      if (pkg === undefined) {                                         // 206
        if (!packageUrl) {                                             // 207
          handle.setError("Unknown package ID: " + packageId + "\nPerhaps it hasn't been uploaded?");
        }                                                              //
        return handle;                                                 // 211
      }                                                                //
                                                                       //
      if (pkg.status !== "ready") {                                    // 214
        if (pkg.status === "failed") {                                 // 215
          handle.setError(pkg.error);                                  // 216
        }                                                              //
        return handle;                                                 // 218
      }                                                                //
                                                                       //
      // From here on, we know the package is installed in the global store, but we
      // might have some per-user things to attend to.                 //
      if (handle.isInstalled()) {                                      // 223
        // The app is installed, so let's send the user to it!  We use `replaceState` so that if
        // the user clicks "back" they don't just get redirected forward again, but end up back
        // at the app list.                                            //
        Router.go("appDetails", { appId: handle.appId() }, { replaceState: true });
      }                                                                //
      return handle;                                                   // 229
    }                                                                  //
  });                                                                  //
                                                                       //
  this.route("upload", {                                               // 233
    where: "server",                                                   // 234
    path: "/upload/:token",                                            // 235
                                                                       //
    action: function () {                                              // 237
      if (!this.params.token || !uploadTokens[this.params.token]) {    // 238
        this.response.writeHead(403, {                                 // 239
          "Content-Type": "text/plain"                                 // 240
        });                                                            //
        this.response.write("Invalid upload token.");                  // 242
        this.response.end();                                           // 243
      } else if (this.request.method === "POST") {                     //
        try {                                                          // 245
          var self = this;                                             // 246
          var packageId = promiseToFuture(doClientUpload(this.request)).wait();
          self.response.writeHead(200, {                               // 248
            "Content-Length": packageId.length,                        // 249
            "Content-Type": "text/plain"                               // 250
          });                                                          //
          self.response.write(packageId);                              // 252
          self.response.end();                                         // 253
          clearTimeout(uploadTokens[this.params.token]);               // 254
          delete uploadTokens[this.params.token];                      // 255
        } catch (error) {                                              //
          console.error(error.stack);                                  // 257
          self.response.writeHead(500, {                               // 258
            "Content-Type": "text/plain"                               // 259
          });                                                          //
          self.response.write("Unpacking SPK failed; is it valid?");   // 261
          self.response.end();                                         // 262
        };                                                             //
      } else {                                                         //
        this.response.writeHead(405, {                                 // 265
          "Content-Type": "text/plain"                                 // 266
        });                                                            //
        this.response.write("You can only POST here.");                // 268
        this.response.end();                                           // 269
      }                                                                //
    }                                                                  //
  });                                                                  //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=install.js.map
