(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// shared/signup.js                                                    //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
// This file covers creation and consumption of invite keys (i.e. to invite people to become
// users of the Sandstorm server).                                     //
                                                                       //
if (Meteor.isServer) {                                                 // 20
  Meteor.publish("signupKey", function (key) {                         // 21
    check(key, String);                                                // 22
    return SignupKeys.find(key);                                       // 23
  });                                                                  //
                                                                       //
  Meteor.methods({                                                     // 26
    useSignupKey: function (key) {                                     // 27
      check(key, String);                                              // 28
                                                                       //
      if (!this.userId) {                                              // 30
        throw new Meteor.Error(403, "Must be signed in.");             // 31
      }                                                                //
                                                                       //
      var user = Meteor.user();                                        // 34
      if (user.signupKey) {                                            // 35
        // Don't waste it.                                             //
        return;                                                        // 37
      }                                                                //
                                                                       //
      if (isDemoUser()) {                                              // 40
        throw new Meteor.Error(403, "Demo users cannot accept invite keys. Please sign in as a real user.");
      }                                                                //
                                                                       //
      if (!user.loginIdentities) {                                     // 45
        // Don't consume signup key as identity user.                  //
        return;                                                        // 47
      }                                                                //
                                                                       //
      var keyInfo = SignupKeys.findOne(key);                           // 50
      if (!keyInfo || keyInfo.used) {                                  // 51
        throw new Meteor.Error(403, "Invalid key or already used.");   // 52
      }                                                                //
                                                                       //
      if (isSignedUp() && user.payments) {                             // 55
        // This user is already signed up with a payment account. Possibly, they signed up before
        // using their invite, and then went back and clicked on the invite. As a result they
        // probably now have two payment accounts. Mark this invite as used but also add a special
        // flag so we can find it later and cancel the dupe payment account. Record who tried to
        // use it so that we can transfer credits over if needed.      //
        SignupKeys.update(key, { $set: { used: true, rejectedBy: this.userId } });
        return;                                                        // 62
      }                                                                //
                                                                       //
      var userFields = {                                               // 65
        signupKey: key,                                                // 66
        signupNote: keyInfo.note                                       // 67
      };                                                               //
      if (keyInfo.email) {                                             // 69
        userFields.signupEmail = keyInfo.email;                        // 70
      }                                                                //
      if ("quota" in keyInfo) {                                        // 72
        userFields.quota = keyInfo.quota;                              // 73
      }                                                                //
      if (keyInfo.plan) {                                              // 75
        userFields.plan = keyInfo.plan;                                // 76
      }                                                                //
      if (keyInfo.payments) {                                          // 78
        userFields.payments = keyInfo.payments;                        // 79
      }                                                                //
                                                                       //
      Meteor.users.update(this.userId, { $set: userFields });          // 82
      SignupKeys.update(key, { $set: { used: true } });                // 83
    }                                                                  //
  });                                                                  //
}                                                                      //
                                                                       //
if (Meteor.isClient) {                                                 // 88
  Template.signup.helpers({                                            // 89
    signupDialog: function () {                                        // 90
      var setting = Settings.findOne("signupDialog");                  // 91
      return setting && setting.value || DEFAULT_SIGNUP_DIALOG;        // 92
    }                                                                  //
  });                                                                  //
}                                                                      //
                                                                       //
Router.map(function () {                                               // 97
  this.route("signup", {                                               // 98
    path: "/signup/:key",                                              // 99
                                                                       //
    waitOn: function () {                                              // 101
      return [Meteor.subscribe("signupKey", this.params.key), Meteor.subscribe("credentials")];
    },                                                                 //
                                                                       //
    data: function () {                                                // 108
      var keyInfo = SignupKeys.findOne(this.params.key);               // 109
      var user = Meteor.user();                                        // 110
                                                                       //
      var result = {                                                   // 112
        keyIsValid: !!keyInfo,                                         // 113
        keyIsUsed: keyInfo && keyInfo.used,                            // 114
        origin: getOrigin(),                                           // 115
        alreadySignedUp: user && !!user.signupKey || keyInfo && user && keyInfo.rejectedBy === user._id,
        hasPaymentInfo: keyInfo && !!keyInfo.payments                  // 118
      };                                                               //
                                                                       //
      if (result.keyIsValid && !result.keyIsUsed && Meteor.userId()) {
        Meteor.call("useSignupKey", this.params.key);                  // 122
      }                                                                //
                                                                       //
      return result;                                                   // 125
    }                                                                  //
  });                                                                  //
});                                                                    //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=signup.js.map
