(function(){

/////////////////////////////////////////////////////////////////////////
//                                                                     //
// shared/stats.js                                                     //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
                                                                       //
// Sandstorm - Personal Cloud Sandbox                                  //
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors
// All rights reserved.                                                //
//                                                                     //
// Licensed under the Apache License, Version 2.0 (the "License");     //
// you may not use this file except in compliance with the License.    //
// You may obtain a copy of the License at                             //
//                                                                     //
//   http://www.apache.org/licenses/LICENSE-2.0                        //
//                                                                     //
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,   //
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.                                      //
                                                                       //
var DAY_MS = 24 * 60 * 60 * 1000;                                      // 17
                                                                       //
if (Meteor.isServer) {                                                 // 19
  if (Mongo.Collection.prototype.aggregate) {                          // 20
    throw new Error("Looks like Meteor wrapped the Collection.aggregate() call. Make sure it " + "works then delete our own wrapper.");
  }                                                                    //
  Mongo.Collection.prototype.aggregate = function () {                 // 24
    // Meteor doesn't wrapp Mongo's aggregate() method.                //
    var raw = this.rawCollection();                                    // 26
    return Meteor.wrapAsync(raw.aggregate, raw).apply(raw, arguments);
  };                                                                   //
                                                                       //
  computeStats = function (since) {                                    // 30
    // We'll need this for a variety of queries.                       //
    var timeConstraint = { $gt: since };                               // 32
                                                                       //
    // This calculates the number of user accounts that have been used
    // during the requested time period.                               //
    var currentlyActiveUsersCount = Meteor.users.find({ expires: { $exists: false }, loginIdentities: { $exists: true },
      lastActive: timeConstraint }).count();                           // 38
                                                                       //
    // This calculates the number of grains that have been used during
    // the requested time period.                                      //
    var activeGrainsCount = Grains.find({ lastUsed: timeConstraint }).count();
                                                                       //
    // If Meteor.settings.allowDemoAccounts is true, DeleteStats       //
    // contains records of type `user` and `appDemoUser`, indicating   //
    // the number of those types of accounts that were created and     //
    // then auto-expired through the demo mode's auto-account-expiry.  //
    var deletedDemoUsersCount = DeleteStats.find({ type: "demoUser", lastActive: timeConstraint }).count();
    var deletedAppDemoUsersCount = DeleteStats.find({ type: "appDemoUser", lastActive: timeConstraint }).count();
                                                                       //
    // Similarly, if the demo is enabled, we auto-delete grains; we store that
    // fact in DeleteStats with type: "grain".                         //
    var deletedGrainsCount = DeleteStats.find({ type: "grain", lastActive: timeConstraint }).count();
                                                                       //
    var apps = Grains.aggregate([{ $match: { lastUsed: timeConstraint } }, { $group: { _id: "$appId", grains: { $sum: 1 }, userIds: { $addToSet: "$userId" } } }, { $project: { grains: 1, owners: { $size: "$userIds" } } }]);
    apps = _.indexBy(apps, "_id");                                     // 63
                                                                       //
    for (var appId in babelHelpers.sanitizeForInObject(apps)) {        // 65
      // We need to count ApiTokens, which don't have appId denormalized into them. We therefore
      // have to fetch a list of grainIds first.                       //
      // TODO(perf): If stats are getting slow, denormalize appId into ApiTokens. Note that it is
      //   already denormalized for the specific case of apps that don't have icons, but the data
      //   for that use case is NOT safe to use here because it's intended that we might allow an
      //   app to mimic another app's icon by spoofing that app ID. In other words, the existing
      //   denormalization of appId should be considered "app ID for identicon purposes only". We'll
      //   need to add a new denormalization for stats purposes -- and make sure that it is not
      //   revealed to the user.                                       //
      var app = apps[appId];                                           // 75
      delete app["_id"];                                               // 76
      var grains = Grains.find({ lastUsed: timeConstraint, appId: appId }, { fields: { _id: 1 } }).fetch();
      var grainIds = _.pluck(grains, "_id");                           // 78
                                                                       //
      var counts = ApiTokens.aggregate([{ $match: { "owner.user.lastUsed": timeConstraint, "grainId": { $in: grainIds } } }, { $group: { _id: "$owner.user.identityId" } }, { $group: { _id: "count", count: { $sum: 1 } } }]);
                                                                       //
      if (counts.length > 0) {                                         // 86
        if (counts.length !== 1) {                                     // 87
          console.error("error: sharedUsers aggregation returned multiple rows");
        }                                                              //
        app.sharedUsers = counts[0].count;                             // 90
      }                                                                //
    }                                                                  //
                                                                       //
    // Count per-app appdemo users and deleted grains.                 //
    DeleteStats.aggregate([{ $match: { lastActive: timeConstraint, appId: { $exists: true } } }, { $group: { _id: { appId: "$appId", type: "$type" }, count: { $sum: 1 } } }]).forEach(function (deletion) {
      var app = apps[deletion._id.appId];                              // 99
      if (!app) {                                                      // 100
        app = apps[deletion.appId] = {};                               // 101
      }                                                                //
      if (deletion._id.type === "appDemoUser") {                       // 103
        app.appDemoUsers = deletion.count;                             // 104
      } else if (deletion._id.type === "grain") {                      //
        app.deleted = deletion.count;                                  // 106
      } else if (deletion._id.type === "demoGrain") {                  //
        app.demoed = deletion.count;                                   // 108
      }                                                                //
    });                                                                //
                                                                       //
    return {                                                           // 112
      activeUsers: currentlyActiveUsersCount,                          // 113
      demoUsers: deletedDemoUsersCount,                                // 114
      appDemoUsers: deletedAppDemoUsersCount,                          // 115
      activeGrains: activeGrainsCount + deletedGrainsCount,            // 116
      apps: apps                                                       // 117
    };                                                                 //
  };                                                                   //
                                                                       //
  function recordStats() {                                             // 121
    var now = new Date();                                              // 122
                                                                       //
    var planStats = _.countBy(Meteor.users.find({ expires: { $exists: false }, "payments.id": { $exists: true } }, { fields: { plan: 1 } }).fetch(), "plan");
                                                                       //
    var record = {                                                     // 130
      timestamp: now,                                                  // 131
      daily: computeStats(new Date(now.getTime() - DAY_MS)),           // 132
      weekly: computeStats(new Date(now.getTime() - 7 * DAY_MS)),      // 133
      monthly: computeStats(new Date(now.getTime() - 30 * DAY_MS)),    // 134
      forever: computeStats(new Date(0)),                              // 135
      plans: planStats                                                 // 136
    };                                                                 //
    record.computeTime = Date.now() - now;                             // 138
                                                                       //
    ActivityStats.insert(record);                                      // 140
    var age = ActivityStats.find().count();                            // 141
    if (age > 3) {                                                     // 142
      var reportSetting = Settings.findOne({ _id: "reportStats" });    // 143
      if (!reportSetting) {                                            // 144
        // Setting not set yet, send out notifications and set it to false
        globalDb.sendAdminNotification("You can help Sandstorm by sending us some anonymous " + "usage stats. Click here for more info.", "/admin/stats");
        Settings.insert({ _id: "reportStats", value: "unset" });       // 148
      } else if (reportSetting.value === true) {                       //
        // The stats page which the user agreed we can send actually displays the whole history
        // of the server, but we're only sending stats from the last day. Let's also throw in the
        // length of said history. This is still strictly less information than what the user said
        // we're allowed to send.                                      //
        record.serverAge = age;                                        // 154
                                                                       //
        HTTP.post("https://alpha-api.sandstorm.io/data", {             // 156
          data: record,                                                // 157
          headers: {                                                   // 158
            Authorization: "Bearer aT-mGyNwsgwZBbZvd5FWr0Ma79O9IehI4NiEO94y_oR",
            "Content-Type": "application/json"                         // 160
          }                                                            //
        });                                                            //
      }                                                                //
    }                                                                  //
  }                                                                    //
                                                                       //
  if (!Meteor.settings.replicaNumber) {                                // 167
    // Wait until 10:00 UTC (2:00 PST / 5:00 EST), then start recording stats every 24 hours.
    // (Only on the first replica to avoid conflicts.)                 //
    Meteor.setTimeout(function () {                                    // 170
      Meteor.setInterval(function () {                                 // 171
        recordStats();                                                 // 172
      }, DAY_MS);                                                      //
                                                                       //
      recordStats();                                                   // 175
    }, DAY_MS - (Date.now() - 10 * 60 * 60 * 1000) % DAY_MS);          //
                                                                       //
    Meteor.startup(function () {                                       // 178
      if (StatsTokens.find().count() === 0) {                          // 179
        StatsTokens.remove({});                                        // 180
        StatsTokens.insert({ _id: Random.id(22) });                    // 181
      }                                                                //
    });                                                                //
  }                                                                    //
                                                                       //
  Meteor.methods({                                                     // 186
    regenerateStatsToken: function () {                                // 187
      if (!isAdmin()) {                                                // 188
        throw new Meteor.Error(403, "Unauthorized", "User must be admin");
      }                                                                //
                                                                       //
      StatsTokens.remove({});                                          // 192
      var token = StatsTokens.insert({ _id: Random.id(22) });          // 193
      return token._id;                                                // 194
    }                                                                  //
  });                                                                  //
}                                                                      //
                                                                       //
// Pseudo-collection defined via publish, above.                       //
RealTimeStats = new Mongo.Collection("realTimeStats");                 // 200
                                                                       //
Router.map(function () {                                               // 202
  this.route("fetchStats", {                                           // 203
    where: "server",                                                   // 204
    path: "/fetchStats/:tokenId",                                      // 205
    action: function () {                                              // 206
      var token = StatsTokens.findOne({ _id: this.params.tokenId });   // 207
                                                                       //
      if (!token) {                                                    // 209
        this.response.writeHead(404, {                                 // 210
          "Content-Type": "text/plain"                                 // 211
        });                                                            //
        this.response.write("Token not found");                        // 213
        return this.response.end();                                    // 214
      }                                                                //
                                                                       //
      try {                                                            // 217
        var stats = ActivityStats.find().fetch();                      // 218
        var statsString = JSON.stringify(stats);                       // 219
                                                                       //
        this.response.writeHead(200, {                                 // 221
          "Content-Type": "application/json"                           // 222
        });                                                            //
        this.response.write(statsString);                              // 224
      } catch (error) {                                                //
        console.error(error.stack);                                    // 226
        this.response.writeHead(500, {                                 // 227
          "Content-Type": "text/plain"                                 // 228
        });                                                            //
        this.response.write(error.stack);                              // 230
      }                                                                //
      return this.response.end();                                      // 232
    }                                                                  //
  });                                                                  //
});                                                                    //
                                                                       //
if (Meteor.isClient) {                                                 // 237
  var saveReportStats = function (newValue, template) {                // 238
    var state = Iron.controller().state;                               // 239
    var token = state.get("token");                                    // 240
    template.reportStatsSaved.set(false);                              // 241
    template.fadeCheckmark.set(false);                                 // 242
    if (template.fadeTimeoutId) {                                      // 243
      Meteor.clearTimeout(template.fadeTimeoutId);                     // 244
    }                                                                  //
                                                                       //
    Meteor.call("setSetting", token, "reportStats", newValue, function (err) {
      if (err) {                                                       // 248
        // TODO(someday): do something with error, for now spinner will just show forever
        return;                                                        // 250
      }                                                                //
      template.reportStatsSaved.set(true);                             // 252
      template.fadeTimeoutId = Meteor.setTimeout(function () {         // 253
        template.fadeCheckmark.set(true);                              // 254
      }, 1000);                                                        //
                                                                       //
      Meteor.call("dismissAdminStatsNotifications", token);            // 257
    });                                                                //
  };                                                                   //
                                                                       //
  Template.adminStats.events({                                         // 261
    'click #regenerateStatsToken': function () {                       // 262
      Meteor.call('regenerateStatsToken');                             // 263
    },                                                                 //
    "change select.package-date": function (ev, template) {            // 265
      template.currentPackageDate.set(ev.currentTarget.value);         // 266
    },                                                                 //
    "change input.enableStatsCollection": function (ev, template) {    // 268
      saveReportStats(ev.target.checked, template);                    // 269
    },                                                                 //
    "click .report-stats-yesno-box>.yes": function (ev, template) {    // 271
      saveReportStats(true, template);                                 // 272
    },                                                                 //
    "click .report-stats-yesno-box>.no": function (ev, template) {     // 274
      saveReportStats(false, template);                                // 275
    }                                                                  //
  });                                                                  //
  Template.adminStats.onCreated(function () {                          // 278
    var state = Iron.controller().state;                               // 279
    var token = state.get("token");                                    // 280
    this.currentPackageDate = new ReactiveVar(null);                   // 281
    this.reportStatsSaved = new ReactiveVar(null);                     // 282
    this.fadeCheckmark = new ReactiveVar(false);                       // 283
    var self = this;                                                   // 284
    this.autorun(function () {                                         // 285
      var stat = ActivityStats.findOne({}, { sort: { timestamp: -1 } });
      if (stat) {                                                      // 287
        self.currentPackageDate.set(stat._id);                         // 288
      }                                                                //
    });                                                                //
    this.subscribe("activityStats", token);                            // 291
    this.subscribe("realTimeStats", token);                            // 292
    this.subscribe("statsTokens", token);                              // 293
    this.subscribe("allPackages", token);                              // 294
  });                                                                  //
  Template.adminStats.helpers({                                        // 296
    setDocumentTitle: function () {                                    // 297
      document.title = "Stats · Admin · " + globalDb.getServerTitle();
    },                                                                 //
    points: function () {                                              // 300
      return ActivityStats.find({}, { sort: { timestamp: -1 } }).map(function (point) {
        return _.extend({                                              // 302
          // Report date of midpoint of sample period.                 //
          day: new Date(point.timestamp.getTime() - 12 * 60 * 60 * 1000).toLocaleDateString()
        }, point);                                                     //
      });                                                              //
    },                                                                 //
    appDates: function () {                                            // 308
      var template = Template.instance();                              // 309
      return ActivityStats.find({}, { sort: { timestamp: -1 }, fields: { timestamp: 1 } }).map(function (point) {
        return _.extend({                                              // 312
          // Report date of midpoint of sample period.                 //
          day: new Date(point.timestamp.getTime() - 12 * 60 * 60 * 1000).toLocaleDateString(),
          selected: point._id === template.currentPackageDate.get()    // 315
        }, point);                                                     //
      });                                                              //
    },                                                                 //
    apps: function () {                                                // 319
      var template = Template.instance();                              // 320
      var stats = ActivityStats.findOne({ _id: template.currentPackageDate.get() });
      if (!stats) {                                                    // 322
        return;                                                        // 323
      }                                                                //
                                                                       //
      var apps = {};                                                   // 326
      var pivotApps = function (time) {                                // 327
        var data = stats[time];                                        // 328
        if (!data) {                                                   // 329
          return;                                                      // 330
        }                                                              //
        data = data.apps;                                              // 332
        for (var appId in babelHelpers.sanitizeForInObject(data)) {    // 333
          var p = data[appId];                                         // 334
          apps[appId] = apps[appId] || {};                             // 335
          apps[appId][time] = p;                                       // 336
        }                                                              //
      };                                                               //
      pivotApps("daily");                                              // 339
      pivotApps("weekly");                                             // 340
      pivotApps("monthly");                                            // 341
      pivotApps("forever");                                            // 342
      return _.chain(apps).map(function (packObj, appId) {             // 343
        packObj.appId = appId;                                         // 345
        // Find the newest version of this app.                        //
        var p = Packages.findOne({ appId: appId, manifest: { $exists: true } }, { sort: { "manifest.appVersion": -1 } });
        if (p) {                                                       // 349
          packObj.appTitle = SandstormDb.appNameFromPackage(p);        // 350
        }                                                              //
        return packObj;                                                // 352
      }).sortBy(function (app) {                                       //
        return -((app.daily || {}).owners || 0);                       // 354
      }).value();                                                      //
    },                                                                 //
    current: function () {                                             // 357
      return RealTimeStats.findOne("now");                             // 358
    },                                                                 //
    today: function () {                                               // 360
      return RealTimeStats.findOne("today");                           // 361
    },                                                                 //
    token: function () {                                               // 363
      return StatsTokens.findOne();                                    // 364
    },                                                                 //
    reportStats: function () {                                         // 366
      var setting = Settings.findOne({ _id: "reportStats" });          // 367
      return setting && setting.value === true;                        // 368
    },                                                                 //
    reportStatsFirstVisit: function () {                               // 370
      var setting = Settings.findOne({ _id: "reportStats" });          // 371
      return !setting || setting.value === "unset";                    // 372
    },                                                                 //
    reportStatsSaving: function () {                                   // 374
      return !Match.test(Template.instance().reportStatsSaved.get(), Match.OneOf(undefined, null));
    },                                                                 //
    reportStatsSaved: function () {                                    // 377
      return Template.instance().reportStatsSaved.get();               // 378
    },                                                                 //
    fadeCheckmark: function () {                                       // 380
      return Template.instance().fadeCheckmark.get();                  // 381
    }                                                                  //
  });                                                                  //
}                                                                      //
/////////////////////////////////////////////////////////////////////////

}).call(this);

//# sourceMappingURL=stats.js.map
